/*         oct2road.c
 *
 *         Author: Enrico Malavasi
 *         Ver. 3.20
 *         Last revised:  August 1991
 *
 *         Main module for the interface Oct-to-Road.
 */
#define  FIRSTDEF 
#include "port.h"
#include "utility.h"
#include "errtrap.h"
#include "options.h"
#include "vov.h"
#include "oct.h"
#include "roadGlobals.h"
#include "roadTechnol.h"
#include "roadMacros.h"
#include "transf.h"
#include "elmacro.h"

#define    INIT_RECTALLOC               500
#define    DEFAULT_INPUT_CELL_VIEW      "unrouted"
#define    MIN_SPACE_AROUND  600       /* Min. Spacing around cells in octCoord */
#define    _EXPAND_ALL   COMMON_FLAG

EXTERN void otrGetCell ARGS((char *cellname));
EXTERN int  printPins(), ShowWirePath();

optionStruct optionList[] = {
    {"b", "N", "Set the external safety boundary to N um"},
    {"c", "",  "Clean up. Delete the old dump files before starting"},
    {"e", "",  "Expand all cells"},
    {"v", "",  "Verbose mode"},
    {"i", "cell_name", "Input cell (def.: <circuit_name>:unrouted)"},
    {"o", "file_name", "Output dump file name (def.: <circuit_name>.dmpX)"},
    {"T", "tech_file", "Technology file name (def.: road.technol)"},
    { OPT_RARG,	"name",	 "Circuit name" },
    { 0, 0, 0 }
};


main(argc,argv)
int argc;
char **argv;
{
    char       option;      /*  Input Option  */
    struct octTransform TT;          /*  Main cell transformation  */
    float      BoundSafe;   /*  Safe boundary in micron   */
    RoadCoord   margin;
    RoadCoord  x;           /*  Safe boundary in Road units */
    int  i;
    int  cleanUp = 0;
    char  *technologyFile;
    char  *circuitName = NULL;
    char  *inputCell = NULL;
    char  *outputFile = NULL;

    errProgramName(argv[0]);
    printf("oct2road: Interface Oct to RoAD");
    printf("\nVersion %s\n", ROAD_VERSION);

    /*  Defaults      */
    technologyFile = util_strsav( TECHNOLFILE );
    BoundSafe = MIN_SPACE_AROUND;
    
    RD_RESET_ALLFLAGS;
    BoundSafe = 30;           /*  Default: 30 lambdas  */
    while ( ( option = optGetOption (argc, argv)) != EOF) {
	switch( option ) {
	  case 'b':      sscanf(optarg, "%f", &BoundSafe);     break;
	  case 'c':      cleanUp = 1;
	  case 'v':      RD_SET_FLAG(RD_VERBOSE_FLAG);         break;
	  case 'e':      RD_SET_FLAG(_EXPAND_ALL);             break;
	  case 'i':      inputCell  = util_strsav( optarg );       break;
	  case 'o':      outputFile = util_strsav( optarg ); break;
	  case 'T':      technologyFile = util_strsav( optarg ); break;
	}
    }

    if(optind != (argc-1)) optUsage();
    margin = MicronToRoad(BoundSafe);
    circuitName = util_strsav( argv[optind] );
    if(inputCell == NULL)   inputCell  = util_strsav( circuitName );
    if(outputFile == NULL)  outputFile = NewDumpFileName( circuitName );

    if(cleanUp) {
	char unixMess[128];
	sprintf(unixMess, "rm -f %s.dmp*", circuitName);
	system(unixMess);
    }

    /*  ------------------------------------------------------------
     *  Initialize the counting variables.
     */
/*    NET_NUM      = 0;        /* No. of rd_net.  */
/*    PIN_NUM      = 0;        /* No. of pins.      */
/*    SUB_CELL_NUM = 0;        /* No. of sub-cells. */
/*    WIRE_NUM     = 0;        /* No. of wires.     */
/*    INST_NUM     = 0;      /* No. of instances  */
    RECT_ALLOC   = INIT_RECTALLOC;

    /*  ------------------------------------------------------------
     *  Allocate a large memory for the needed structures.
     */
    rects =  ALLOC(struct rectangle, RECT_ALLOC);
    cells =  ALLOC(Cell, CELL_MAX);
    insts =  ALLOC(RoadInstance, MAX_INST_NUM);
    pins  =  ALLOC(RoadTerm, PIN_MAX);
    pins[0].rect = NULL;	/* Not still set the rect reference */
    rd_net = ALLOC(RoadNet, NET_MAX);
    rdWire = ALLOC(_rdWire, NODE_MAX*3);  

    VOVbegin( argc, argv );
    VOVinput( VOV_UNIX_FILE, technologyFile );
    read_technol( technologyFile );        /* Read Technology data ............ */
    
    VERBOSE(printf("\nReading %s \n", inputCell ));
    otrGetCell( inputCell, margin );
    
    cells      = REALLOC(Cell,     cells,  SUB_CELL_NUM+1);
    pins       = REALLOC(RoadTerm, pins,   PIN_NUM);
    rd_net     = REALLOC(RoadNet,  rd_net, NET_NUM);
    rdWire     = REALLOC(_rdWire,  rdWire,  NET_NUM*3);

    /*  ------------------------------------------------------------
     *  Create the final rectagonal subdivision.
     */
    initMessage("Creating routing obstacles ", 0);
    fusePins();
    FuseCells();
    initMessage("Creating the grid frame ", 0);
    divide();
    fragment();
    CostruisciIFili();
    SortCells();

    VOVoutput( VOV_UNIX_FILE, outputFile );
    RoadDump( outputFile );     /*  Write RDF file VOV reference inside   */
    /*  ------------------------------------------------------------
     *  Print some data on the circuit
     */
    printf("\n    %5d instances", INST_NUM);
    printf("\n    %5d grid nodes", TOTAL_RECT_NUM);
    printf("\n    %5d nets", NET_NUM);
    printf("\n    %5d wires\n", WIRE_NUM);
    rdExit( 0 );
}

rdExit( s )
int s;
{
    VOVend( s );
}

#include "dummy.h"

