/*       otrGetCell.c
 *
 *       Author: Enrico Malavasi
 *       Ver. 3.10 -    Last revised: July 1991
 *
 *       Purpose: Utilities for the Oct-To-RoAD interface.
 */
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "utility.h"
#include "roadGlobals.h"
#include  "transf.h"

#define   _EXPAND_ALL   COMMON_FLAG
#define   otrReduceTerm(t)    \
    { octObject ttt;      \
	if(octGenFirstContent(t, OCT_TERM_MASK, &ttt) == OCT_OK) \
	  *t = ttt; }

static int
  otrLayerToNum(lyr, first)
octObject *lyr;
int  first;
/*
 *   Given a layer OCT, return the index of the corresponding
 *   layer ROAD.
 */
{
  int i, j;

  for(i=first ; i<LAYER_NUM ; i++) {
    if(strcasecmp(lyr->contents.layer.name, layers[i].name) == 0)
      return(i);

    if(layers[i].masking_layers != NULL)
      for(j=0 ; layers[i].masking_layers[j] != NULL ; j++)
	if(strcasecmp(lyr->contents.layer.name, layers[i].masking_layers[j]) == 0)
	  return(i);
  }

  return(NOTFOUND);
}

static int
otrNewCell(bb, layerNo, instNo, margin)
octBox  *bb;
int      layerNo;
int      instNo;
octCoord margin;
/*
 *  Create a new cell ROAD.
 */
{
  /*  No need for reallocation  */
  SUB_CELL_NUM++;
  if(SUB_CELL_NUM >= CELL_MAX)
    errRaise("OTR", -1, "CELL_MAX is too small ...\n");

  cells[SUB_CELL_NUM].x1 = bb->lowerLeft.x;
  cells[SUB_CELL_NUM].y1 = bb->lowerLeft.y;
  cells[SUB_CELL_NUM].x2 = bb->upperRight.x;
  cells[SUB_CELL_NUM].y2 = bb->upperRight.y;
  cells[SUB_CELL_NUM].layer = layerNo;
  cells[SUB_CELL_NUM].instanceNo = instNo;

  if(SUB_CELL_NUM == 1) {
    cells[0].x1 = cells[SUB_CELL_NUM].x1 - margin;
    cells[0].y1 = cells[SUB_CELL_NUM].y1 - margin;
    cells[0].x2 = cells[SUB_CELL_NUM].x2 + margin;
    cells[0].y2 = cells[SUB_CELL_NUM].y2 + margin;
  } else {
    cells[0].x1 = min(bb->lowerLeft.x-margin,  cells[0].x1);
    cells[0].y1 = min(bb->lowerLeft.y-margin,  cells[0].y1);
    cells[0].x2 = max(bb->upperRight.x+margin, cells[0].x2),
    cells[0].y2 = max(bb->upperRight.y+margin, cells[0].y2);
  }
  return SUB_CELL_NUM;
}

static int
otrNewInstance(inst)
octObject *inst;
/*
 *  Create a new instance ROAD
 */
{
    int  i;

    for(i=0 ; i<INST_NUM ; i++) {
	if(strcmp(insts[i].name, inst->contents.instance.name) == 0) {
	    fprintf("\nTwo instances with same name: %s", insts[i].name);
	    exit(1);
	}
    }
    /*  New instance  */
    insts[INST_NUM].name = ALLOC(char, 128);
    strcpy(insts[INST_NUM].name, inst->contents.instance.name);
    insts[INST_NUM].master = ALLOC(char, 128);
    strcpy(insts[INST_NUM].master, inst->contents.instance.master);
    insts[INST_NUM].view = ALLOC(char, 32);
    strcpy(insts[INST_NUM].view, inst->contents.instance.view);
    insts[INST_NUM].transf = inst->contents.instance.transform.transformType;
    insts[INST_NUM].x      = inst->contents.instance.transform.translation.x;
    insts[INST_NUM].y      = inst->contents.instance.transform.translation.y;
    insts[INST_NUM].noov   = 0;
    return (INST_NUM++);
}

static int
otrNewNet(net)
octObject  *net;
/*
 *  Create a new net ROAD.
 */
{
  static counter = 0;

  /*   No need for reallocation  */
  rd_net[NET_NUM].resistive_weight  = 0;
  rd_net[NET_NUM].capacitive_weight = 0;
  rd_net[NET_NUM].bulk_shld         = 0;
  rd_net[NET_NUM].wwidth            = 0;
  rd_net[NET_NUM].link              = NO;
  rd_net[NET_NUM].name = ALLOC(char, 128);
  if(net->contents.net.name[0] == '\0') {
    sprintf(net->contents.net.name, "rd_net%d_%d", getpid(), counter++);
    OH_ASSERT(octModify(&net));
    VERBOSE( printf("Unnamed net renamed %s\n", net->contents.net.name));
  } else
    strcpy(rd_net[NET_NUM].name, net->contents.net.name);
  return (NET_NUM++);
}

static void
  otrExpandInst(inst, margin)
octObject *inst;
octCoord  margin;
/*
 *  Expand an instance OCT to a set of cell obstacles for ROAD
 */
{
    octObject    master;    /*  Master facet of the instance  */
    octObject    layer;
    octGenerator layerGen;
    octObject    bb;
    int          layerNo;
    int          instNo;
    
    instNo = otrNewInstance(inst);
    
    bb.type = OCT_BOX;
    OH_ASSERT(ohOpenMaster(&master, inst, "interface", "r"));
    OH_ASSERT(octInitGenContents(&master, OCT_LAYER_MASK, &layerGen));
    while(octGenerate(&layerGen, &layer) == OCT_OK)
      for(layerNo=0 ; (layerNo=otrLayerToNum(&layer, layerNo)) != NOTFOUND ; layerNo++) {
	  OH_ASSERT(ohBB(&layer, &(bb.contents.box)));
	  octTransformGeo(&bb, &(inst->contents.instance.transform));
	  otrNewCell(&(bb.contents.box), layerNo, instNo, margin);
      }
    OH_ASSERT(octFreeGenerator(&layerGen));
    OH_ASSERT(octCloseFacet(&master));
}

static void
otrFullExpandInst(inst, margin)
octObject *inst;
octCoord  margin;
/*
 *  Expand an instance OCT to a set of cell obstacles for ROAD
 */
{
    octObject    master;    /*  Master facet of the instance  */
    octObject    layer;
    octObject    geo;
    octGenerator layerGen;
    octGenerator geoGen;
    octObject    bb;
    int          layerNo;
    int          instNo;

    instNo = otrNewInstance(inst);
    
    bb.type = OCT_BOX;
    OH_ASSERT(ohOpenMaster(&master, inst, "interface", "r"));
    OH_ASSERT(octInitGenContents(&master, OCT_LAYER_MASK, &layerGen));
    while(octGenerate(&layerGen, &layer) == OCT_OK)
      for(layerNo=0 ; (layerNo=otrLayerToNum(&layer, layerNo)) != NOTFOUND ; layerNo++) {
	  OH_ASSERT(octInitGenContents(&layer, OCT_GEO_MASK, &geoGen));
	  while(octGenerate(&geoGen, &geo) == OCT_OK) {
	      OH_ASSERT(ohBB(&geo, &(bb.contents.box)));
	      octTransformGeo(&bb, &(inst->contents.instance.transform));
	      otrNewCell(&(bb.contents.box), layerNo, instNo, margin);
	  }
	  octFreeGenerator(&geoGen);
      }
    OH_ASSERT(octFreeGenerator(&layerGen));
    OH_ASSERT(octCloseFacet(&master));
}

static octStatus
  otrTermBB(term, refLayerName, bb)
octObject *term;
char      *refLayerName;
octBox    *bb;
/*
 *  Return the bounding box for the given terminal 
 *  on the given layer.
 */
{
    octObject    aTerm, fTerm;
    octObject    geo;
    octObject    layer;
    octObject    inst;
    octObject    box;
    octGenerator geoGen;

    aTerm = *term;
    otrReduceTerm(&aTerm);
    if(octIdIsNull(aTerm.contents.term.instanceId)) {  /*  Formal terminal */
	OH_ASSERT(octInitGenContents(&aTerm, OCT_GEO_MASK, &geoGen));
	while(octGenerate(&geoGen, &geo) == OCT_OK) {
	    OH_ASSERT(octGenFirstContainer(&geo, OCT_LAYER_MASK, &layer));
	    if(strcmp(layer.contents.layer.name, refLayerName) == 0) {
		/*  Eureka */
		
		OH_ASSERT(ohBB(&geo, bb));
		return OCT_OK;
	    }
	}
	/*  Not found  */
	return  OCT_NOT_FOUND;
    } else {
	OH_ASSERT(octGenFirstContainer(&aTerm, OCT_INSTANCE_MASK, &inst));
	if(ohFindFormal(&fTerm, &aTerm) != OCT_OK) {
	    fprintf(stderr, "WARNING: corrupt terminal %s ignored\n", aTerm.contents.term.name);
	    return  OCT_NOT_FOUND;
	}
	if(otrTermBB(&fTerm, refLayerName, bb) == OCT_NOT_FOUND)
	  return  OCT_NOT_FOUND;
	box.type = OCT_BOX;
	box.contents.box = *bb;
	octTransformGeo(&box, &(inst.contents.instance.transform));
	*bb = box.contents.box;
	return  OCT_OK;
    }
}

static int
  otrInstNo(obj)
octObject *obj;
/*
 *  Given an object, return the instance index corresponding
 *  to it (inst), or to the instance containing it (term).
 */
{
    int instNo;

    switch (obj->type) {
      case OCT_TERM:
	{  octObject  jTerm, inst;
	   jTerm = *obj;
	   otrReduceTerm(&jTerm);
	   if(octGenFirstContainer(&jTerm, OCT_INSTANCE_MASK, &inst) == OCT_OK) {
	       if((instNo=otrInstNo(&inst)) == NOTFOUND)
		 instNo = otrNewInstance(&inst);
	       return instNo;
	   }
       }
	
      case OCT_INSTANCE:
	{  int  i;
	   ForAllInsts(i,0) {
	       if(strcmp(insts[i].name, obj->contents.instance.name) == 0) return i;
	   }
	   return NOTFOUND;
       }
	
      default:
	errRaise("OTR", 1, "Unknown object type in otrInstNo");
    }
}

static void
  otrTermEquivalences(facet)
octObject *facet;
{
    octObject  master, inst;
    octObject  Jbag, bag;
    octObject  term;
    octGenerator jbagGener;
    octGenerator termGener;
    int   instNo;
    int   termNo;
    int   eqNo;

    ForAllInsts(instNo, 1) {
	OH_ASSERT(ohGetByInstName(facet, &inst, insts[instNo].name));
	OH_ASSERT(ohOpenMaster(&master, &inst, "interface", "r"));
	if(ohGetByBagName(&master, &Jbag, "JOINEDTERMS") == OCT_OK) {
	    OH_ASSERT(octInitGenContents(&Jbag, OCT_BAG_MASK, &jbagGener));
	    while(octGenerate(&jbagGener, &bag) == OCT_OK) {
		eqNo = NOTFOUND;
		OH_ASSERT(octInitGenContents(&bag, OCT_TERM_MASK, &termGener));
		while(octGenerate(&termGener, &term) == OCT_OK) {
		    for(termNo=0 ; termNo<PIN_NUM ; termNo++) {
			if(cells[pins[termNo].cellNo].instanceNo == instNo) {
			    if(strcmp(pins[termNo].TermName, term.contents.term.name) == 0) {
				if(eqNo == NOTFOUND)  eqNo = pins[termNo].equivNo;
				else                  pins[termNo].equivNo = eqNo;
			    }
			}
		    }
		}
	    }
	}
	octCloseFacet(&master);
    }
}

static void
  otrTermExtract(term, netNo)
octObject  *term;     /*  Already reduced terminal  */
int         netNo;
/*
 *  Generate the terminals for ROAD corresponding to a terminal OCT
 */
{
    octBox       bb;
    octCoord     pos;
    int          layerNo;
    int          instNo;
    int          cellNo;
    int          eqNo;
    int          i;

    eqNo = PIN_NUM;
    ForAllLayers(layerNo) {
	if(otrTermBB(term, layers[layerNo].name, &bb) == OCT_OK) {
	    instNo = otrInstNo(term);
	    cellNo = otrNewCell(&bb, layerNo, instNo, 0);
	    pos = (bb.upperRight.x + bb.lowerLeft.x) / 2 - cells[cellNo].x1;
	    newRoadPin(netNo, layerNo, NORTH, pos, NULL, cellNo, eqNo, term->contents.term.name);
	    newRoadPin(netNo, layerNo, SOUTH, pos, NULL, cellNo, eqNo, term->contents.term.name);
	    pos = (bb.upperRight.y + bb.lowerLeft.y) / 2 - cells[cellNo].y1;
	    newRoadPin(netNo, layerNo, EAST,  pos, NULL, cellNo, eqNo, term->contents.term.name);
	    newRoadPin(netNo, layerNo, WEST,  pos, NULL, cellNo, eqNo, term->contents.term.name);
	}
    }
}

void
  otrGetCell( cellname, margin )
char      *cellname;     /*  Cell name ................................ */
RoadCoord  margin;
/*
 *  Get an instance and extract its geometry, nets and terminals
 */
{
    octObject    facet;
    octObject    instBag;        /*  Instances bag */
    octObject    inst;           /*  instance  */
    octObject    net;
    octObject    term;           /*  Terminal */
    octGenerator instGen;        /*  Instance generator */
    octGenerator netGen;         /*  Net generator */
    octGenerator termGen;        /*  Term generator */
    PathName     facetName;
    int          netNo;          /*  Net index  */

    sprintf(facetName, "%s", cellname);
    ohUnpackDefaults(&facet, "a", ":unrouted:contents");
    ohUnpackFacetName(&facet, facetName);
    OH_ASSERT(octOpenFacet(&facet));
    
    insts[0].name   = util_strsav( facet.contents.facet.cell );
    insts[0].master = insts[0].name;
    insts[0].view   = util_strsav( facet.contents.facet.view );
    insts[0].transf = OCT_NO_TRANSFORM;
    insts[0].x      = 0;
    insts[0].y      = 0;
    insts[0].noov   = 0;
    INST_NUM = 1;

    /*   Get all the instance cells  */
    if(ohGetOrCreateBag(&facet, &instBag, "INSTANCES") == OCT_OK) {
	OH_ASSERT(octInitGenContents(&instBag, OCT_INSTANCE_MASK, &instGen));
	while(octGenerate(&instGen, &inst) == OCT_OK) {
	    if(RD_CHECK_FLAG(_EXPAND_ALL)) {
		VERBOSE( printf("--> Expanding instance %s\n", inst.contents.instance.name));
		otrFullExpandInst(&inst, margin);
	    } else {
		VERBOSE( printf("--> Reading instance %s\n", inst.contents.instance.name));
		otrExpandInst(&inst, margin);
	    }
	}
	OH_ASSERT(octFreeGenerator(&instGen));
    }

    if(SUB_CELL_NUM < 1) {
	errRaise("ROAD", 1, "No instances found in facet %s", ohFormatName(&facet));
    }

    /*  Get all nets  */
    OH_ASSERT(octInitGenContents(&facet, OCT_NET_MASK, &netGen));
    while(octGenerate(&netGen, &net) == OCT_OK) {
	netNo = otrNewNet(&net);

	/*  Get all the terminals for this net */
	OH_ASSERT(octInitGenContents(&net, OCT_TERM_MASK, &termGen));
	while(octGenerate(&termGen, &term) == OCT_OK) {
	    otrTermExtract(&term, netNo);
	}
	OH_ASSERT(octFreeGenerator(&termGen));
    }
    OH_ASSERT(octFreeGenerator(&netGen));
    if(NET_NUM < 1) {
	errRaise("ROAD", 1, "No nets found in facet %s", ohFormatName(&facet));
    }
    otrTermEquivalences(&facet);   /*  Build the equivalences between terminals */
    otrWireTranslate(&facet);       /*  Extract all wires */
    
    OH_ASSERT(octCloseFacet(&facet));
}
