/*      otrWire.c
 *
 *      Author: Enrico Malavasi
 *      Ver. 3.00
 *      Last revised: January, 1991
 *
 *      Purpose: Get wires from given paths
 */
#include "port.h"
#include "oct.h"
#include "utility.h"
#include "roadGlobals.h"
#include "oh.h"
#include  "transf.h"

#define  MAX_PNT   500
#define  isFormalTerminal(Tr) ((Tr)->contents.term.instanceId==oct_null_id)

typedef struct { RoadCoord  x;
		 RoadCoord  y;
		 int        z;
	       } RoadPoint;
typedef struct {  RoadCoord  wwidth;
		  RoadPoint  pts[MAX_PNT];
		  RoadPoint  via[MAX_PNT];
		  char      *NetName;
		  char      *TermName;
		} preWire;

static int            PWN;       /*  preWire Size */
static preWire       *preWires = NULL;

/*  ============================================================
 *  This function extracts from a net its node number, which is
 *  returned. If the net name does not contain such a number, it
 *  is automatically generated.
 */
static
  otrExtractNetName(Net)
octObject *Net;
{
    if(Net->contents.net.name[0] == '\0') {
	Net->contents.net.name = ALLOC(char, 10);
	sprintf(Net->contents.net.name, "rd_%d", Net->objectId);
	warnMessage("Unnamed net renamed %s", Net->contents.net.name);
    }
}

static int
  otrLayerToNum(Layer, first)
octObject *Layer;
int  first;
/*  Transform a given OCT-Layer into the corresponding RoAD layer index
 *  for obstacles. The diffusions are transformed into  obstacles for POLY. */
{
    int i, j;

    for(i=first ; i<LAYER_NUM ; i++)
      if(strcasecmp(Layer->contents.layer.name, layers[i].name) == 0)
	return i;

    for(i=first ; i<LAYER_NUM ; i++)
      if(layers[i].masking_layers != NULL)
	for(j=0 ; layers[i].masking_layers[j] != NULL ; j++)
	  if(strcasecmp(Layer->contents.layer.name, layers[i].masking_layers[j]) == 0)
	    return i;
    
    return NOTFOUND;
}

static void
  PathCenterEnds(Path, Pts)
octObject       *Path;
struct octPoint *Pts;
{
    static octGenerator Gener;
    octObject  Term1, Term2;
    struct octBox BB1, BB2;

    if((Pts[0].x > Pts[1].x) || (Pts[0].y > Pts[1].y))
      SWAP(struct octPoint, Pts[0], Pts[1]);
    
    octInitGenContents(Path, OCT_TERM_MASK, &Gener);
    octGenerate(&Gener, &Term1);
    octGenerate(&Gener, &Term2);
    ohTerminalBB(&Term1, &BB1);
    ohTerminalBB(&Term2, &BB2);
    if((BB1.lowerLeft.x > BB2.lowerLeft.x) || (BB1.lowerLeft.y > BB2.lowerLeft.y))
      SWAP(struct octBox, BB1, BB2);
    if(Pts[0].x == Pts[1].x) {
	Pts[0].y = (BB1.lowerLeft.y+BB1.upperRight.y)/2;
	Pts[1].y = (BB2.lowerLeft.y+BB2.upperRight.y)/2;
    }
    if(Pts[0].y == Pts[1].y) {
	Pts[0].x = (BB1.lowerLeft.x+BB1.upperRight.x)/2;
	Pts[1].x = (BB2.lowerLeft.x+BB2.upperRight.x)/2;
    }
}

/*  ===============================================================
 *  This function extracts every net of the given facet and creates
 *  the corresponding pre_wires.
 */
otrWireTranslate(Facet)
octObject *Facet;
{
    octObject    Net;           /*  Net     */
    octObject    Layer;         /*  Layer   */
    octObject    Path;          /*  Path    */
    octObject    Conn;          /*  Connector */
    octObject    Term;          /*  Terminal  */
    octObject    Bag;           /*  CONNECTORS bag  */
    octGenerator Gener;         /*  Net generator  */
    octGenerator NGener;        /*  Path generator */
    octGenerator TGener;        /*  Term generator */
    struct octBox  BB;
    struct octPoint Pts[2];     /*  Path terminals */
    int          via[2];
    int          Due;           /*  Dummy int.     */
    int          z, i, ViaNum;
    int          PntNum;        /*  Current pre_wire size. */

    octInitGenContents(Facet, OCT_NET_MASK, &Gener);
    while(octGenerate(&Gener, &Net) == OCT_OK) {
	PntNum = 0;
	preWires = REALLOC(preWire, preWires, (PWN+1));
	otrExtractNetName(&Net);
	preWires[PWN].NetName = util_strsav( Net.contents.net.name );
	preWires[PWN].TermName = NULL;
	octInitGenContents(&Net, OCT_PATH_MASK, &NGener);
	while(octGenerate(&NGener, &Path) == OCT_OK) {
	    preWires[PWN].wwidth = Path.contents.path.width;
	    octGenFirstContainer(&Path, OCT_LAYER_MASK, &Layer);
	    preWires[PWN].pts[PntNum+1].z = 
	      preWires[PWN].pts[PntNum].z = otrLayerToNum(&Layer, 0);
	    octGetPoints(&Path, &Due, Pts);
	    PathCenterEnds(&Path, Pts);
	    preWires[PWN].pts[PntNum].x   = Pts[0].x;
	    preWires[PWN].pts[PntNum++].y = Pts[0].y;
	    preWires[PWN].pts[PntNum].x   = Pts[1].x;
	    preWires[PWN].pts[PntNum++].y = Pts[1].y;
	    if(PntNum > MAX_PNT)
	      errMessage("path_to_wire: MAX_PNT too small", 0);
	    preWires[PWN].pts[PntNum].z = (-1);
	    /*  ------------------------------------------------------------
	     *  Looking for the terminal attachment.
	     */
	    if(preWires[PWN].TermName == NULL) {
		octInitGenContents(&Path, OCT_TERM_MASK, &TGener);
		while(octGenerate(&TGener, &Term) == OCT_OK) {
		    if(isFormalTerminal(&Term)) {
			preWires[PWN].TermName = util_strsav( Term.contents.term.name );
			break;
		    }
		}
		octFreeGenerator(&TGener);
	    }
	}
	octFreeGenerator(&NGener);
	
	if(PntNum > MAX_PNT) {
	    errMessage("OPWireTranslate: MAX_PNT is too small", 0);
	}
	if(PntNum > 0) PWN++;
    }
    octFreeGenerator(&Gener);
    
    /*  ------------------------------------------------------------
     *  Find all vias.
     */
    for(i=0 ; i<PWN ; i++)
      preWires[i].via[0].z = (-1);
    if(ohGetByBagName(Facet, &Bag, "CONNECTORS") > 0) {
	octInitGenContents(&Bag, OCT_INSTANCE_MASK, &Gener);
	while(octGenerate(&Gener, &Conn) == OCT_OK) {
	    via[0] = (-1); Due = 0;
	    ohBB(&Conn, &BB);
	    octGenFirstContent(&Conn, OCT_TERM_MASK, &Term);
	    octInitGenContainers(&Term, OCT_PATH_MASK, &NGener);
	    while(octGenerate(&NGener, &Path) == OCT_OK) {
		octGenFirstContainer(&Path, OCT_LAYER_MASK, &Layer);
		z = otrLayerToNum(&Layer, 0);
		if(via[0] < 0) via[0] = z;
		else if(via[0] != z) {
		    Due = 2;
		    via[1] = max(via[0], z);
		    via[0] = min(via[0], z);
		    break;
		}
	    }
	    octFreeGenerator(&NGener);
	    
	    if(Due == 2) {
		octGenFirstContainer(&Path, OCT_NET_MASK, &Net);
		otrExtractNetName(&Net);
		for(i=0 ; i<PWN ; i++)
		  if(strcasecmp(preWires[i].NetName, Net.contents.net.name) == 0)
		    break;
		for(ViaNum=0 ; preWires[i].via[ViaNum].z>=0 ; ) ViaNum++;
		do {
		    preWires[i].via[ViaNum].x = (BB.lowerLeft.x+BB.upperRight.x)/2;
		    preWires[i].via[ViaNum].y = (BB.lowerLeft.y+BB.upperRight.y)/2;
		    preWires[i].via[ViaNum++].z = via[0]++;
		    preWires[i].via[ViaNum].z   = (-1);
		} while(via[0] < via[1]);
	    }
	}
	octFreeGenerator(&Gener);
    }
}

#define    MAXWIRESIZE  300
CostruisciIFili()
{
  int       i, j, jm1, l;
  int       side;
  int       z1;
  RoadCoord x1,y1,x2,y2;
  Rect      r;

/*  ------------------------------------------------------------
 *  Transform to RoAD coordinates.
 */
  for(i=0 ; i<PWN ; i++) {
    for(j=0 ; preWires[i].pts[j].z>=0 ; j++) {
      preWires[i].pts[j].x = preWires[i].pts[j].x;
      preWires[i].pts[j].y = preWires[i].pts[j].y;
    }
    for(j=0 ; preWires[i].via[j].z>=0 ; j++) {
      preWires[i].via[j].x = preWires[i].via[j].x;
      preWires[i].via[j].y = preWires[i].via[j].y;
    }
  }

  for(i=WIRE_NUM=0 ; i<PWN ; i++) {
      WIRE_NUM++;
      wireNetNo(i) = name_to_net( preWires[i].NetName );
      wireDesc(i)  = DEFWWIDTH;  /*  DEFAULT: fixed width */
      rdWire[i].wWidth = ALLOC(RoadCoord, 1);
      wSegmWidth(i, 0) = preWires[i].wwidth;
      rdWire[i].rect = ALLOC(Rect, 1);
      rdWire[i].side = ALLOC(int,  1);
      wSegmRect(i, 0)           = NULL;
      wireSize(i)               = 0;

    for(j=1,jm1=0 ; preWires[i].pts[jm1].z>=0 ; j+=2,jm1+=2) {
	if((preWires[i].pts[jm1].x > preWires[i].pts[j].x) ||
	   (preWires[i].pts[jm1].y > preWires[i].pts[j].y))
	  SWAP(RoadPoint, preWires[i].pts[jm1], preWires[i].pts[j]);
	x1 = preWires[i].pts[jm1].x + insts[0].x;
	y1 = preWires[i].pts[jm1].y + insts[0].y;
	z1 = preWires[i].pts[jm1].z;
	x2 = preWires[i].pts[j].x   + insts[0].x;
	y2 = preWires[i].pts[j].y   + insts[0].y;
	if(x1 == x2) side = EAST;
	else         side = NORTH;

	ForAllRects(l) {
	    if(rects[l].layer != z1) continue;
	    if(rects[l].x1 >= x2) continue;
	    if(rects[l].x2 <  x1) continue;
	    if(rects[l].y1 >= y2) continue;
	    if(rects[l].y2 <  y1) continue;
	    if((rects[l].x2 == x1) && (rects[l].y2 == y1)) continue;
	    
#define   ACTION(D,side)   { GridSplit(rects+l,D,side,NOTFOUND);l--;continue; }
	    if(rects[l].y2 > y2)	  ACTION(y2-rects[l].y1, EAST);
	    if((rects[l].y1 < y1) &&
	       (y1 < rects[l].y2))    ACTION(y1-rects[l].y1, EAST);
	    if(rects[l].x2 > x2)	  ACTION(x2-rects[l].x1, NORTH);
	    if((rects[l].x1 < x1) &&
	       (x1 < rects[l].x2))    ACTION(x1-rects[l].x1, NORTH);
	    add_to_wire(i, rects+l, side);
	    set_free_neighbor(rects+l, side, 1);
	}
    }
/*  ------------------------------------------------------------
 *  Vias.
 */
      for(j=0 ; preWires[i].via[j].z>=0 ; j++) {
	  x1 = preWires[i].via[j].x + insts[0].x;
	  y1 = preWires[i].via[j].y + insts[0].y;
	  z1 = preWires[i].via[j].z;
	  if((r=find_rect(x1, y1, z1)) != NULL) {
	      add_to_wire(i, r, UP);
	      set_free_neighbor(r, UP, 1);
	  }
      }
  }
  ForAllWires( i ) {
      free_wire(i, 1);
  }
}
