/*        LTG.c
 *
 *        Author: Enrico Malavasi, Univ of Bologna
 *        Ver. 1.00    -    September, 1988
 *
 *        Purpose: Line-To-Ground (Capacity-to-bulk) models.
 *                 Capacity is returned in pFarad.
 *        Bibliography:
 *        T.Sakurai, K.Tamaru, "Simple Formulas for Two- and 
 *                   Three-Dimensional Capacitances", IEEE Trans. on E.D.,
 *                   vol. ED-30, n.2, February 1983, pp.183-185
 *        N.v.d.Meijs, J.T.Fokkema, "VLSI circuit reconstruction from
 *                   mask topology", Integration, vol.2, n.2, 1984, pp.85-119
 *        E. Barke, "Line-to-Ground Capacitance Calculation for
 *                   VLSI: A Comparison", IEEE Trans. on CAD,
 *                   vol.7, n.2, February 1988, pp.295-298
 */
#include <math.h>
#include "elmacro.h"
#include "rd_types.h"
#include "roadTechnol.h"

/*  ---------------------------------------------------------------
 *  This function implements the Sakurai-Tamaru's wire-to-ground
 *  capacitive model
 */
pFarad LTG_SakuraiTamaru(h, t, W, L)
micron    h, t;        /* Oxide and Wire thickness  */
micron    W, L;        /* Wire width and length     */
{
  pFarad x;

  x = TECH_OX_DIELECTRIC *
    (1.15 * W/h + 2.80 * pow(t/h, 0.222));
  x *= L;
  return(x);
}

/*  ---------------------------------------------------------------
 *  This function implements the v.d.Meijs-Fokkema's wire-to-ground
 *  capacitive model
 */
pFarad LTG_MeijsFokkema(h, t, W, L)
micron    h, t;        /* Oxide and Wire thickness  */
micron    W, L;        /* Wire width and length     */
{
  pFarad x;

  x = W/h;
  x = TECH_OX_DIELECTRIC *
    (x + 0.77 + 1.06 * pow(x, 0.25) + 1.06 * sqrt(t/h));
  x *= L;
  return(x);
}

/*  ---------------------------------------------------------------
 *  This function implements the parallel-plates wire-to-ground
 *  capacitive model
 */
pFarad LTG_XxxModel(h, t, W, L)
micron    h, t;        /* Oxide and Wire thickness  */
micron    W, L;        /* Wire width and length     */
{
  pFarad x;

  x = TECH_OX_DIELECTRIC * W/h;
  x *= L;
  return(x);
}
