/*      appendix.c
 *
 *	Author: Enrico Malavasi, Univ. of Bologna
 *      Ver. 1.00
 * 	Last revised: April, 1991
 *
 *	Purpose: Investigate on a wire to find the presence of eventual
 *               floating appendixes. The given wire must be FREED(0)
 */
#include "roadGlobals.h"

/*  ============================================================
 *  Appendixes on UP or DOWN sides. The given rectangle is the one
 *  on whose NE corner the upward via is located. The function
 *  returns 1 if the segment is floating and can be deleted.
 */
static
  delete_appendixes_c(wn, r)
int  wn;
Rect r;
{
    int  i;
    Rect s1, s2;

    /*  ------------------------------------------------------------
     *  Look at the north and east sides of <r>.
     */
    for(i=0 ; i<wireSize(wn) ; i++) {
	if((wSegmRect(wn,i) == r) && (wSegmSide(wn,i) != UP))      break;
	s1 = neighbor(r, EAST);
	if((wSegmRect(wn,i) == s1) && (wSegmSide(wn,i) == NORTH))  break;
	s1 = neighbor(r, NORTH);
	if((wSegmRect(wn,i) == s1) && (wSegmSide(wn,i) == EAST))   break;
	s1 = neighbor(r, DOWN);
	if((wSegmRect(wn,i) == s1) && (wSegmSide(wn,i) == UP))     break;
    }
    if(i >= wireSize(wn))    return YES;
    
    r = neighbor(r, UP);
    for(i=0 ; i<wireSize(wn) ; i++) {
	if(wSegmRect(wn,i) == r)                                   return NO;
	s1 = neighbor(r, EAST);
	if((wSegmRect(wn,i) == s1) && (wSegmSide(wn,i) == NORTH))  return NO;
	s1 = neighbor(r, NORTH);
	if((wSegmRect(wn,i) == s1) && (wSegmSide(wn,i) == EAST))   return NO;
    }
    return YES;
}

/*  ============================================================
 *  This function investigates on a given wire segment, to
 *  determine whether it is a floating appendix. It returns
 *  YES if it is a floating appendix, else it returns NO.
 */
delete_appendixes(wn, r, side)
int  wn;            /* Wire index  */
Rect r;            /* Given rectangle ..................... */
int side;           /* Side occupied on the given rectangle  */
{
    int dir1, i, j;
    Rect a,
         s0,s1,s2;
    
    if(r == NULL) return NO;
    if(side == UP) return(delete_appendixes_c(wn, r));
    
    dir1 = AbsDualDir(side);
    s0 = free_neighbor(r,side);
    s1 = free_neighbor(r,dir1);
    s2 = neighbor(r,DOWN);
    for(i=0 ; i<wireSize(wn) ; i++) {
	a = wSegmRect(wn, i);
	j = wSegmSide(wn, i);
	
	if(((a == r) && (j != side))	 ||
	   ((a == s0) && (j == dir1))	 ||
	   ((a == s1) && (j == side))	 ||
	   ((a == s2) && (j == UP)))
	  break;
    }
    if(i == wireSize(wn))
      return YES; /* This segment is floating and must be removed . */
    
    /*      Lower side.  */
    s0 = free_neighbor(r, -dir1);
    if(s0 == NULL)  return(YES);   /* Obstacle: This is a floating segment  */
    s1 = free_neighbor(s0, side);
    s2 = neighbor(s0, DOWN);
    
    for(i=0 ; i<wireSize(wn) ; i++) {
	a = wSegmRect(wn, i);
	j = wSegmSide(wn, i);

	if((a == s0) ||
	   ((a == s1) && (j == dir1))                         ||
	   ((a == s2) && (j == UP)))
	  return NO; /* This segment is not floating ................ */
    }
    return YES;  /*  This segment is floating and it must be removed */
}


#define    RD_ADJ_NORTH    (0x0001<<1)
#define    RD_ADJ_DNORTH   (0x0001<<2)
#define    RD_ADJ_UP       (0x0001<<3)
#define    RD_ADJ_DUP      (0x0001<<4)
#define    RD_ADJ_EAST     (0x0001<<5)
#define    RD_ADJ_DEAST    (0x0001<<6)
#define    RD_RECTCLEAR_ADJ(r) ((r)->flag &= 0xff81)

int
  rdCheckWireDMark( wn, pos )
int  wn;
int  pos;
{
    switch (wSegmSide(wn,pos)) {
      case NORTH:
	return (RD_CHECKRECTFLAG(wSegmRect(wn,pos), RD_ADJ_NORTH) &&
		RD_CHECKRECTFLAG(wSegmRect(wn,pos), RD_ADJ_DNORTH));
      case EAST:
	return (RD_CHECKRECTFLAG(wSegmRect(wn,pos), RD_ADJ_EAST) &&
		RD_CHECKRECTFLAG(wSegmRect(wn,pos), RD_ADJ_DEAST));
      case UP:
	return (RD_CHECKRECTFLAG(wSegmRect(wn,pos), RD_ADJ_UP) &&
		RD_CHECKRECTFLAG(wSegmRect(wn,pos), RD_ADJ_UP));
    }
    return 1;
}

void
  rdSetWireMark( wn )
int  wn;
{
    int  i;
    Rect s, s2, r;

    for(i=0 ; i<wireSize(wn) ; i++) {
	RD_RECTCLEAR_ADJ(wSegmRect(wn,i));
    }

    for(i=0 ; i<wireSize(wn) ; i++) {
	r = wSegmRect(wn,i);
	switch ( wSegmSide(wn, i) ) {
	  case NORTH:
	    RD_SETRECTFLAG(r, RD_ADJ_DEAST);
	    RD_SETRECTFLAG(r, RD_ADJ_UP);
	    if((s=free_neighbor(r,WEST)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_DNORTH);
		RD_SETRECTFLAG(s, RD_ADJ_UP);
		RD_SETRECTFLAG(s, RD_ADJ_DEAST);
		if((s2=free_neighbor(s, NORTH)) != NULL) {
		    RD_SETRECTFLAG(s2, RD_ADJ_EAST);
		}
		if((s2=neighbor(s, DOWN)) != NULL) {
		    RD_SETRECTFLAG(s2, RD_ADJ_DUP);
		}
	    }
	    if((s=free_neighbor(r,NORTH)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_EAST);
	    }
	    if((s=free_neighbor(r,EAST)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_NORTH);
	    }
	    if((s=neighbor(r,DOWN)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_DUP);
	    }
	    break;

	  case EAST:
	    RD_SETRECTFLAG(r, RD_ADJ_DNORTH);
	    RD_SETRECTFLAG(r, RD_ADJ_UP);
	    if((s=free_neighbor(r, SOUTH)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_DEAST);
		RD_SETRECTFLAG(s, RD_ADJ_UP);
		RD_SETRECTFLAG(s, RD_ADJ_DNORTH);
		if((s2=free_neighbor(s, EAST)) != NULL) {
		    RD_SETRECTFLAG(s2, RD_ADJ_NORTH);
		}
		if((s2=neighbor(s, DOWN)) != NULL) {
		    RD_SETRECTFLAG(s2, RD_ADJ_DUP);
		}
	    }
	    if((s=free_neighbor(r,EAST)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_NORTH);
	    }
	    if((s=free_neighbor(r,NORTH)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_EAST);
	    }
	    if((s=neighbor(r,DOWN)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_DUP);
	    }
	    break;

	  case UP:
	    RD_SETRECTFLAG(r, RD_ADJ_DNORTH);
	    RD_SETRECTFLAG(r, RD_ADJ_DEAST);
	    if((s=free_neighbor(r, EAST)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_NORTH);
	    }
	    if((s=free_neighbor(r,NORTH)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_EAST);
	    }
	    if((s=neighbor(r,DOWN)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_DUP);
	    }
	    if((s=free_neighbor(r, UP)) != NULL) {
		RD_SETRECTFLAG(s, RD_ADJ_DNORTH);
		RD_SETRECTFLAG(s, RD_ADJ_UP);
		RD_SETRECTFLAG(s, RD_ADJ_DEAST);
		if((s2=free_neighbor(s, EAST)) != NULL) {
		    RD_SETRECTFLAG(s2, RD_ADJ_NORTH);
		}
		if((s2=neighbor(s, NORTH)) != NULL) {
		    RD_SETRECTFLAG(s2, RD_ADJ_EAST);
		}
	    }
	    break;
	}
    }
}
