/*        capcoup.c
 *
 *        Author: Enrico Malavasi
 *        Ver. 2.10
 *        Last revised:  October, 1990
 *
 *        Purpose: Compute the amount of capacitive coupling between
 *                 two given nets.
 */
#include "roadGlobals.h"
#include "roadTechnol.h"
#include "models.h"
#include "transf.h"
#define  WireWidth(wn,r,side)      layers[r->layer].wire_width

/*  ============================================================
 *  This function returns the capacitance to bulk for a given net
 */
static pFarad
  capToBulk(wireNo)
int  wireNo;
{
    int       i;         /*  Loop counter  */
    RoadCoord len;       /*  Wire segment length in RoadCoord */
    pFarad    cap;       /*  Returned capacitance  */
    micron    Mw, Ml;

    for(i=0,cap=0.0 ; i<wireSize(wireNo) ; i++) {
	if(wSegmSide(wireNo, i) != UP) {
	    len = ((wSegmSide(wireNo, i) == NORTH) ?
		   rwidth(wSegmRect(wireNo, i)) : rheight(wSegmRect(wireNo, i)));
	    Ml  = RoadToMicron(len);
	    len = WireWidth(wireNo, wSegmRect(wireNo, i), wSegmSide(wireNo, i));
	    Mw  = RoadToMicron(len);
	    cap += LTG_MODEL((micron) layers[wSegmLayer(wireNo, i)].bulk_dist,
			     (micron) layers[wSegmLayer(wireNo, i)].wire_thickness,Mw,Ml);
	}
    }
    return cap;
}

/*  ============================================================
 *  This function returns an estimate of the capacitive coupling
 *  between the wires implementing two given nodes.
 */
pFarad
capacitive_coupling(net1, net2)
int net1, net2;     /* Nets to be controlled */
{
  int   wn1, wn2,          /* Wire indexes .................... */
        i, j,              /* Loop counters ................... */
        side, mside;
  RoadCoord x, y, lw[6];
  pFarad COUP = 0.0;       /* Capacitive coupling: return par.  */
  Rect  r,
        s, s1, s2;
  micron Ml, Mw, Mh;

  if(net1 == net2)  return(COUP);
/*  ------------------------------------------------------------
 *  Check: the two wires must be already routed
 */
  if((wn1=net_to_wire(net1,0)) == NOTFOUND)  return(COUP);
  if((wn2=net_to_wire(net2,0)) == NOTFOUND)  return(COUP);

  for(i=0 ; i<LAYER_NUM ; i++)
    lw[i] = layers[i].wire_width + layers[i].wire_dist;

/*  ------------------------------------------------------------
 *  1. Horizontal coupling between parallel wires
 */
  for(i=0 ; i<wireSize(wn1) ; i++) {
      r    = wSegmRect(wn1, i);
      side = wSegmSide(wn1, i);
      if(side == UP)  continue; /* We look only for horizontal elements */

    mside = (- side);
    s1 = follow_FreeNeighbor(neighbor(r, side), side);
    s2 = follow_FreeNeighbor(r, mside);

    for(j=0,s=s1 ; j<2 ; j++,s=s2)
      if(wSegmHere(wn2, s, side)) {
	if(side == EAST) {
	  x = s->x2 - r->x2;
	  y = rheight(r);
	}
	else {
	  x = s->y2 - r->y2;
	  y = rwidth(r);
	}
	if(x < lw[r->layer]) x = lw[r->layer];
	x -= WireWidth(wn2, r, side);
	COUP += PLL_MODEL((micron) layers[rLayer(r)].wire_thickness,
			  (micron) RoadToMicron(x),
			  (micron) RoadToMicron(y));
      }
  }

/*  ------------------------------------------------------------
 *  2. Vertical coupling between parallel wires on different layers,
 *     running one above the other
 */
  for(i=0 ; i<wireSize(wn1) ; i++) {
    r    = wSegmRect(wn1, i);
    side = wSegmSide(wn1, i);
    if(side == UP)   continue;  /* We look only for horizontal elements  */

    s1 = neighbor(r, UP);
    s2 = neighbor(r, DOWN);

    if(s1 != NULL)
      if(wSegmHere(wn2, s1, side)) {
/*  ------------------------------------------------------------
 *  <wn1> is running under <wn2>
 */
	x = WireWidth(wn1, r, side);                  Ml = RoadToMicron(x);
	x = WireWidth(wn2, s1, side);                 Mw = RoadToMicron(x);
	x = (side == EAST) ? rheight(r) : rwidth(r); Mh = RoadToMicron(x);
	COUP += OVL_MODEL(layer_distance(rLayer(r), rLayer(s1)), Ml, Mw, Mh);
	continue;
      }

    if(s2 != NULL)
      if(wSegmHere(wn2, s2, side)) {
/*  ------------------------------------------------------------
 *  <wn2> is running under <wn1>
 */
	x = WireWidth(wn2, s2, side);                 Ml = RoadToMicron(x);
	x = WireWidth(wn1, r, side);                  Mw = RoadToMicron(x);
	x = (side == EAST) ? rheight(r) : rwidth(r); Mh = RoadToMicron(x);
	COUP += OVL_MODEL(layer_distance(rLayer(s2), rLayer(r)), Ml, Mw, Mh);
	continue;
      }
  }

/*  ------------------------------------------------------------
 *  3.) Wire Crossing
 */
  for(i=0 ; (r=wSegmRect(wn1, i))!=NULL ; i++) {
      if((side=wSegmSide(wn1, i)) == UP)  continue;  /* Look only for horiz. elements */
      mside = AbsDualDir(side);

      s1 = neighbor(r, UP);
      s2 = neighbor(r, DOWN);

    if(s1 != NULL)
      if(wSegmHere(wn2, s1, mside)) {
/*  ------------------------------------------------------------
 *  There is a crossing, with <wn1> under <wn2>
 */
	x = WireWidth(wn1, r, side);
	Ml = RoadToMicron(x);
	x = WireWidth(wn2, s1, side);
	Mw = RoadToMicron(x);
	COUP += CRS_MODEL(layer_distance(rLayer(r), rLayer(s1)), Ml, Mw);
	continue;
      }

    if(s2 != NULL)
      if(wSegmHere(wn2, s2, mside)) {
/*  ------------------------------------------------------------
 *  There is a crossing, with <wn2> under <wn1>
 */
	x = WireWidth(wn2, s2, side);
	Ml = RoadToMicron(x);
	x = WireWidth(wn1, r, side);
	Mw = RoadToMicron(x);
	COUP += CRS_MODEL(layer_distance(rLayer(s2), rLayer(r)), Ml, Mw);
	continue;
      }
  }

  if(net1 == status.BulkNet)       COUP += capToBulk(wn2);
  else if(net2 == status.BulkNet)  COUP += capToBulk(wn1);
  return(COUP);
}
