/*      critlist.c
 *
 *      Author: Enrico Malavasi
 *      Ver. 3.00
 *      Last revised: January, 1991
 *
 *      Purpose: Management of the critic coupling array.
 */
#include "port.h"
#include "utility.h"
#include "roadGlobals.h"
#include "roadTechnol.h"
#include "transf.h"
#include "models.h"

typedef struct _critPair {
    int        net1;       /* critic nets.                      */
    int        net2;
    RoadWeight coupling;    /* Coupling weight        ........  */
    pFarad     maxcap;      /* Max. capacitive coupling (in pF) */
    int        shield;      /* Shield net         ............. */
} _critPair;

static  _critPair *critPair = NULL;   /* Critical signal coupling array .... */
static  int   critPairNum   = 0;
static  int   critPairAlloc = 0;  /*  Array size  */

#define  DEFCoupling     0.5       /*  Default coupling weight  */
#define  DEFMaxcap       (-1.0)    /*  Default max.allowed cap. */
#define  DEFShield       NOTFOUND  /*  Default shield net       */

getCriticCapPair(netNo, coupling, maxcap, shldNet)
int         netNo;
RoadWeight *coupling;      /*  Coupling weight */
pFarad     *maxcap;        /*  Max.Allowed Capacitance  */
int        *shldNet;       /*  Shield net     */
/*  Get a given net and return its coupled net,
 *  if any, filling the given parameters with their corresponding
 *  values. If no coupling is specified for the given net,
 *  NOTFOUND is returned.
 */
{
  int i;

  for(i=0 ; i<critPairNum ; i++) {
    if(critPair[i].net1 == netNo) {
      (*coupling) = critPair[i].coupling;
      (*maxcap)   = critPair[i].maxcap;
      (*shldNet)  = critPair[i].shield;
      return(critPair[i].net2);
    }
    if(critPair[i].net2 == netNo) {
      (*coupling) = critPair[i].coupling;
      (*maxcap)   = critPair[i].maxcap;
      (*shldNet)  = critPair[i].shield;
      return(critPair[i].net1);
    }
  }
  return(NOTFOUND);
}

modifyCriticCap(net1, net2, coupling, maxcap, shldNet, mask)
int        net1, net2;
RoadWeight coupling;
pFarad     maxcap;
int        shldNet;
{
    int   i;

    SORT(int, net1, net2);
    for(i=0 ; i<critPairNum ; i++)
      if((critPair[i].net1 == net1) && (critPair[i].net2 == net2)) {
	  if(mask & CRIT_MAXCAP_MASK)   critPair[i].maxcap   = maxcap;
	  if(mask & CRIT_COUPLING_MASK) critPair[i].coupling = coupling;
	  if(mask & CRIT_SHIELD_MASK)   critPair[i].shield   = shldNet;
	  return;
      }

    if(i >= critPairAlloc) {
	critPairAlloc += 10;
	critPair= REALLOC(_critPair, critPair, critPairAlloc);
    }
    critPair[i].net1    = net1;
    critPair[i].net2    = net2;
    critPair[i].coupling = (mask & CRIT_COUPLING_MASK) ? coupling : DEFCoupling;
    critPair[i].maxcap   = (mask & CRIT_MAXCAP_MASK)   ? maxcap   : DEFMaxcap;
    critPair[i].shield   = (mask & CRIT_SHIELD_MASK)   ? shldNet  : DEFShield;
    critPairNum++;
}

/*  ==============================================================
 *  This function returns the coupling weight of the pair (n1, n2)
 */
RoadWeight
InquireCC_coupling(net1, net2, maxcap)
int      net1, net2;
pFarad  *maxcap;
{
    int i;

    SORT(int, net1, net2);
    for(i=0 ; i<critPairNum ; i++) {
	if((critPair[i].net1 == net1) && (critPair[i].net2 == net2)) {
	    (*maxcap) = critPair[i].maxcap;
	    return critPair[i].coupling;
	}
    }
    return((RoadWeight) 0.0);
}
