/*	fragment.c
 *
 *	Author: Enrico Malavasi
 *      Ver. 3.10
 * 	Last revised: April, 1991
 *
 *	Purpose: Grid fragmentation.
 */
#include "roadGlobals.h"
#include "transf.h"

#define  regNeededSpacing(l)  (layers[l].wire_width/2.0 + layers[l].wire_dist)
#define   MAX_RECT_SIZE  20    /* Max size = 1/20 of the circuit min dimension */

/*  ============================================================
 *  This function performs a grid fragmentation, in order to:
 *  - Allocate every pin to a different rectangle for each net;
 *  - Allow not too big grid nodes.
 */
fragment()
{
    int i,j;        /* General purpose variables ... */
    RoadCoord        x0, y0, MaxW;
    Rect         r;            /* Current rectangle ........... */
    Pins         p, pj;


    /* Max. allowed grid node size. */
    x0   = cRightX(cells) - cLeftX(cells);
    y0   = cUpperY(cells) - cLowerY(cells);
    if(x0 < y0)  y0 = x0;
    MaxW  =  y0 / MAX_RECT_SIZE;
    /*  Set the <rect> field for every pin and delete the useless pins.  */
    ForAllPins(i) {
	p = pins+i;
	p->rect = NULL;    /*  Don't delete this line  ! */
	if((p->rect=find_rect(pin_x(p), pin_y(p), rdLayer(p))) == NULL) {
	    delete_pin(i--);
	    continue;
	}
	(p->rect)->flag |= RECTFLAG_TERM;

	switch(p->side) {
	  case NORTH:
	  case SOUTH:
	    r = follow_Neighbor(p->rect, EAST);
	    x0 = rRightX(r) - pin_x(p);
	    if(GetSymmetryFlag())
	      if(rRightX(r) == cRightX(cells))
		x0 = 1000;
	    break;

	  default:
	    r = follow_Neighbor(p->rect, NORTH);
	    x0 = rUpperY(r) - pin_y(p);
	    break;
	}
	if(x0 < regNeededSpacing(p->layer)) {
	    delete_pin(i--);
	    continue;
	}
	switch(p->side) {
	  case NORTH:
	  case SOUTH:
	    r = follow_Neighbor(p->rect, WEST);
	    x0 = pin_x(p) - rLeftX(r);
	    break;

	  default:
	    r = follow_Neighbor(p->rect, SOUTH);
	    x0 = pin_y(p) - rLowerY(r);
	    break;
	}
	if(x0 < regNeededSpacing(p->layer)) {
	    delete_pin(i--);
	    continue;
	}
    }

    /*  Introduce a fragmentation corresponding to the pins position.  */
    ForAllPins(i) {
	p = pins+i;
	r  = p->rect;
	if north_or_south(p->side) GridSplit(r,pin_x(p)-rLeftX(r),NORTH,NOTFOUND);
	else                       GridSplit(r,pin_y(p)-rLowerY(r),EAST,NOTFOUND);
    }

    /*  ------------------------------------------------------------
     *  Now find the rectangles where pins of the same net affer
     *  and split them.
     */
    ForAllPins(i)
      for(j=i+1,p=pins+i ; j<PIN_NUM ; j++)  {
	  pj = pins+j;
	  /*  ------------------------------------------------------------
	   *  If two pins belong to the same net
	   *  and affer to the same rectangle, then ...
	   */
	  if((p->netNo == pj->netNo)  &&
	     ((r=p->rect) == pj->rect))  {
	      /*  ------------------------------------------------------------
	       *  ... split the rectangle between them.
	       */
	      x0 = pin_x(p) - pin_x(pj);
	      y0 = pin_y(p) - pin_y(pj);
	      
	      if(abs(x0) > abs(y0))
		GridSplit(r, (pin_x(p)+pin_x(pj))/2-rLeftX(r), NORTH, NOTFOUND);
	      else
		GridSplit(r, (pin_y(p)+pin_y(pj))/2-rLowerY(r), EAST, NOTFOUND);
	  }
      }

    ForAllPins(i) {
	(pins[i].rect)->flag |= RECTFLAG_TERM; /* Could be lost in the previous step */
    }
    /*  -------------------------------------------------------------------
     *  Find the rectangles whose size is greater than a minimum fixed size
     *  and split them.
     */
    for(i=0 ; i<TOTAL_RECT_NUM ; ) {
	r = rects+i;
	if(rwidth(r) > MaxW) {
	    GridSplit(r, MaxW, NORTH, NOTFOUND);
	    continue;
	}
	if(rheight(r) > MaxW) {
	    GridSplit(r, MaxW, EAST, NOTFOUND);
	    continue;
	}
	i++;
    }
}

/*  ============================================================
 *  This function returns the pointer to the rectangle where the
 *  point (x,y) is.
 */
Rect find_rect(x, y, l)
RoadCoord x,y;	/* Coordinates .......... */
char l;		/* Level ................ */
{
    int  i;
    Rect r;

    for(i=0,r=rects ; i<TOTAL_RECT_NUM ; i++,r++)
      if((x >  rLeftX(r))    &&
	 (x <= rRightX(r))   &&
	 (y >  rLowerY(r))   &&
	 (y <= rUpperY(r))   &&
	 ((r->layer & 0x0F) == l))
	return(r);
    return(NULL);    /*  Rect not found  */
}

/*  ============================================================
 *  Fuse the overlapped pins
 */
fusePins()
{
    int i, k;

    for(i=0 ; i<PIN_NUM ; i++)
      for(k=i+1 ; k<PIN_NUM ; k++)
	if((pins[i].layer == pins[k].layer) && (pins[i].netNo == pins[k].netNo))
	  if((pin_x(pins+i) == pin_x(pins+k)) && (pin_y(pins+i) == pin_y(pins+k))) {
	      join_equivalence(pins+k, pins+i);
	      delete_pin(k--);
	  }
}
