/*	inpnet.c
 *
 *	Author: Enrico Malavasi
 *      Ver. 3.00
 * 	Last revised:  January, 1991
 *
 * 	Purpose: Input the set of net constraints from the NDF file.
 */
#include "port.h"
#include "utility.h"
#include "roadGlobals.h"
#include "roadTechnol.h"
#include "transf.h"
#include "models.h"

#define   MAXCOMMNUM      30     /* Max number of commands on a line */
#define   MAXCOMMLEN      80     /* Max length of a command in chars */

#define   RANGE0    0
#define   RANGE1    1            /* Max range for the given weights  */
#define   CheckWeight(w)        ((w <= RANGE1) && (w >= RANGE0))

#define   GETNETNO(z,x)    if((z=name_to_net(x)) == NOTFOUND) \
          errMessage("NDF: Net %s does not exist in the circuit", x)
#define   GETCLASSNETS(x,F)     \
  { int k,classNo; if((classNo=name_to_class(x)) != NOTFOUND) {   \
                     for(k=0 ; NetClass[classNo].net[k]>=0 ; k++) {   \
		       strcpy(x, rd_net[NetClass[classNo].net[k]].name); \
		       F(argv, argc);  }  \
		     return; }}

typedef struct _NetClass {
    int  *net;                /* List of nets in this class */
    char *name;               /* Class name */
} _NetClass;
static _NetClass *NetClass = NULL;      /* Array of net classes */
static int        NetClassNum = 0;      /* Size of the previous array */

static int
  name_to_class(name)
char *name;
{
    int i;
    for(i=0 ; i<NetClassNum ; i++)
      if(strcasecmp(name, NetClass[i].name) == 0)
	return i;
    return(NOTFOUND);
}

static void
  readNETCLASS(argv, argc)
char **argv;
int    argc;
/*   Read a net class 
 *   Syntax:    CLASS   classname  {nets}*
 */
{
    int k;
    char name[40];

    if((name_to_net(argv[1]) != NOTFOUND) || (name_to_class(argv[1]) != NOTFOUND))
      errMessage("Class %s defined twice", argv[1]);
    NetClass = REALLOC(_NetClass, NetClass, NetClassNum+1);
    NetClass[NetClassNum].name = util_strsav(argv[1]);
    NetClass[NetClassNum].net  = ALLOC( int, argc-1 );
    for(k=2 ; k<argc ; k++) {
	GETNETNO(NetClass[NetClassNum].net[k-2], argv[k]);
    }
    NetClass[NetClassNum++].net[argc-2] = NOTFOUND;
}

static void
  readDEFAULT(argv, argc)
char **argv;
int    argc;
/*   Read the default parameters  */
{
    int k, i;
    float  weight;

    for(k=1 ; k<argc ; k++) {
	if(strcasecmp(argv[k], "RESISTIVE") == 0) {
	    sscanf(argv[++k], "%f", &weight);
	    ForAllNets( i )
	      rd_net[i].resistive_weight = weight;
	    continue;
	}
	if(strcasecmp(argv[k], "CAPACITIVE") == 0) {
	    sscanf(argv[++k], "%f", &weight);
	    ForAllNets( i )
	      rd_net[i].capacitive_weight = weight;
	    continue;
	}
    }
}
static void
  readSTRATEGY(argv, argc)
char **argv;
int    argc;
{
    int    k, l, dir;
    float  prw;

    preferWeight = 2.0;    /*   Default  */
    for(k=1 ; k<argc ; k++) {
	if(strcasecmp(argv[k], "PREFER") == 0) {
	    sscanf(argv[++k], "%f", &prw);
	    if(prw <= 1) {
		warnMessage("PREFER weight <= 1. Command will be ignored", 0);
		prw = 0.0;
	    }
	    preferWeight = prw;
	    continue;
	}
	ForAllLayers( l )
	  if(strcasecmp(argv[k], layers[l].name) == 0) {
	      sscanf(argv[++k], "%d", &dir);
	      layers[l].preferDir = abs(dir);
	      break;
	  }
    }
}

static void
  renumberNet(oldNet, NetNo)
int  oldNet, NetNo;
{
    int i, wn;

    for(wn=0 ; (wn=net_to_wire(oldNet,0)) != NOTFOUND ; wn++)
      wireNetNo(wn) = NetNo;
    if(oldNet != NetNo) {
	ForAllPins(i)
	  if(pins[i].netNo == oldNet)
	    pins[i].netNo = NetNo;
    }
    NET_NUM--;
    for(wn=0 ; (wn=net_to_wire(NET_NUM,0)) != NOTFOUND ; wn++)

      wireNetNo(wn) = oldNet;
    rd_net[oldNet] = rd_net[NET_NUM];
}

static void
  readNET(argv, argc)
char **argv;
int    argc;
{
    int  netNo, oldNet=NOTFOUND;
    int  k;
    float  weight, width;
    
    GETCLASSNETS(argv[1], readNET);
    GETNETNO(netNo, argv[1]);

    oldNet = NOTFOUND;
    for(k=2 ; k<argc ; k++) {
	if(strcasecmp(argv[k], "RESISTIVE") == 0) {
	    sscanf(argv[++k], "%f", &weight);
	    if(netNo != NOTFOUND)
	      rd_net[netNo].resistive_weight  = weight;
	    unless(CheckWeight(weight))
	      errMessage("A weight is out of range", 0);
	    continue;
	}
	if(strcasecmp(argv[k], "CAPACITIVE") == 0) {
	    sscanf(argv[++k], "%f", &weight);
	    if(netNo != NOTFOUND)
	      rd_net[netNo].capacitive_weight = weight;
	    unless(CheckWeight(weight))
	      errMessage("A weight is out of range", 0);
	    continue;
	}
	if(strcasecmp(argv[k], "WIDTH") == 0) {
	    sscanf(argv[++k], "%f", &width);
	    if(netNo != NOTFOUND)    rd_net[netNo].wwidth = MicronToRoad(width);
	    continue;
	}
	if(strcasecmp(argv[k], "BULK") == 0) {
	    k++;
	    if(netNo != NOTFOUND)    status.BulkNet = netNo;
	    continue;
	}
    }
    if(oldNet != NOTFOUND) renumberNet(oldNet, netNo);
}

static void
  readCRITIC(argv, argc)
char **argv;
int    argc;
{
    int   net1, net2,
    shldNet, k;
    float weight, cap;
    char  *nName;
    
    
    GETCLASSNETS(argv[1], readCRITIC);
    GETNETNO(net1, argv[1]);
    GETCLASSNETS(argv[2], readCRITIC);
    GETNETNO(net2, argv[2]);
    sscanf(argv[3], "%f", &weight);
    
    SORT(int, net1, net2);
    
    cap = (-1);            /*  Default  */
    shldNet = NOTFOUND;    /*  Default  */
    
    for(k=4 ; k<argc ; k++) {
	if(strcasecmp(argv[k], "SHIELD") == 0) {
	    nName = argv[++k];
	    GETNETNO(shldNet, nName);
	    continue;
	}
	if(strcasecmp(argv[k], "MAXCAP") == 0) {
	    sscanf(argv[++k], "%f", &cap);
	    cap /= 1000.0;
	    continue;
	}
    }
    modifyCriticCap(net1, net2,
		    (RoadWeight) weight, (pFarad) cap, shldNet,
		    CRIT_ALL_MASK);
}

static void
  readAutoSymm(argv, argc)
char **argv;
int    argc;
{
    int i, k, netNo;
    int classNo;
    
    if((argc == 1) || (strcasecmp(argv[1], "ALL") == 0)) {
	for(i=0 ; i<NET_NUM ; i++)
	  if(InquireSymmetryList(i) == NOTFOUND)
	    AddToSymmetryList(i, i);
    } else
      for(i=1 ; i<argc ; i++) {
	  if((classNo=name_to_class(argv[i])) != NOTFOUND) {
	      for(k=0 ; NetClass[classNo].net[k]>=0 ; k++)
		AddToSymmetryList(NetClass[classNo].net[k], NetClass[classNo].net[k]);
	  } else {
	      GETNETNO(netNo, argv[i]);
	      AddToSymmetryList(netNo, netNo);
	  }
      }
    SetSymmetryFlag(YES);
}

static void
  readSPLIT(argv, argc)
char **argv;
int    argc;
{
    int i, k, netNo;
    int classNo;
    
    if((argc == 1) || (strcasecmp(argv[1], "ALL") == 0))
      for(i=0 ; i<NET_NUM ; i++)
	rd_net[i].link = YES;
    else
      for(i=1 ; i<argc ; i++) {
	  if((classNo=name_to_class(argv[i])) != NOTFOUND) {
	      for(k=0 ; NetClass[classNo].net[k]>=0 ; k++)
		rd_net[NetClass[classNo].net[k]].link = YES;
	  } else {
	      GETNETNO(netNo, argv[i]);
	      rd_net[netNo].link = YES;
	  }
      }
}

static void
  readNOOVERLAP(argv, argc)
char **argv;
int    argc;
{
    int i, j;
    
    if((argc == 1) || (strcasecmp(argv[1], "ALL") == 0))
      for(i=1 ; i<INST_NUM ; i++)
	insts[i].noov = YES;
    else
      for(i=1 ; i<argc ; i++) {
	  for(j=1 ; j<INST_NUM ; j++)
	    if((strcasecmp(argv[i], insts[j].name)) == 0) {
		insts[j].noov = YES;
		break;
	    }
	  if(j == INST_NUM)
	    errMessage("Instance %s not found", argv[i]);
      }
}

static void
  readSYMMETRY(argv, argc)
char **argv;
int    argc;
{
    int net1, net2;
    
    GETNETNO(net1, argv[1]);
    GETNETNO(net2, argv[2]);
    AddToSymmetryList(net1, net2);
    SetSymmetryFlag(YES);
}

input_netlist( ndfFile )
char *ndfFile;          /*  Name of the net-descriptor file  */
{
    FILE      *fp;   /*  Pointer to the net-descriptor file  */
    int        i, wn1, wn2;
    int        retValue = 0;
    int        argc, k;
    RoadWeight wght;
    pFarad     maxcap;
    char     **argv;

    /*   Preset size and last-element of the constraint structures.   */
    SetSymmetryFlag(NO);
    status.BulkNet = NOTFOUND;

    /*  If the NDF is read-enabled AND has been defined,
     *  now it is read.
     */
    if(RD_CHECK_FLAG( NET_READ ) && ((fp = fopen(ndfFile,"r")) != NULL)) {

	retValue = 1;

	argv = ALLOC(char *, MAXCOMMNUM);
	for(i=0 ; i<MAXCOMMNUM ; i++)
	  argv[i] = ALLOC(char, MAXCOMMLEN);
	
	loop {
	    inputCommandLine(fp, &argc, argv);
	    if(argc <= 0) break;
	    
	    if(argv[0][0] == '#')   continue;               /*  Comment  */
	    
	    if(argc > MAXCOMMNUM) {
		fprintf(stderr, "Too many arguments for command %s in file %s\n", argv[0], ndfFile);
		exit( 1 );
	    }
	    
	    if(strcasecmp(argv[0], "DEFAULT") == 0) {
		readDEFAULT(argv, argc);
		continue;
	    }
	    if(strcasecmp(argv[0], "STRATEGY") == 0) {
		readSTRATEGY(argv, argc);
		continue;
	    }
	    
	    if(strcasecmp(argv[0], "CLASS") == 0) {
		readNETCLASS(argv, argc);
		continue;
	    }
	    
	    /*  ------------------------------------------------------------
	     *  Terminals
	     */
	    if((strcasecmp(argv[0], "PIN") == 0) ||
	       (strcasecmp(argv[0], "TERM") == 0)) {
		fprintf(stderr, "Command %s is obsolete. Please use padplace", argv[0]);
		exit(1);
	    }
	    
	    /*  ------------------------------------------------------------
	     *  Cells
	     */
	    if(strcasecmp(argv[0], "NOOVERLAP") == 0) {
		readNOOVERLAP(argv, argc);
		continue;
	    }
	    
	    /*  ------------------------------------------------------------
	     *  NETS.
	     */
	    if(strcasecmp(argv[0], "NET") == 0) {
		readNET(argv, argc);
		continue;
	    }
	    
	    /*  ------------------------------------------------------------
	     *  Nets to be split
	     */
	    if(strcasecmp(argv[0], "SPLIT") == 0) {
		readSPLIT(argv, argc);
		continue;
	    }
	    
	    /*  ------------------------------------------------------------
	     *  Critic coupling relations.
	     */
	    if(strcasecmp(argv[0], "CRITIC") == 0) {
		readCRITIC(argv, argc);
		continue;
	    }
	    /*  ------------------------------------------------------------
	     *  Symmetries.
	     */
	    if(strcasecmp(argv[0], "SYMMETRY") == 0) {
		readSYMMETRY(argv, argc);
		continue;
	    }
	    if((strcasecmp(argv[0], "AUTOSYMM") == 0) ||
	       (strcasecmp(argv[0], "SELFSYMM") == 0)) {
		readAutoSymm(argv, argc);
		continue;
	    }
	}
	
	fclose( fp );
	
	for(i=0 ; i<MAXCOMMNUM ; i++)
	  free(argv[i]);
	free(argv);
    }
   
    /*   Reset the link field of the nets which have not to be split */
    ForAllNets( i )
      if(rd_net[i].link == YES) {
	  ForAllPins(k)
	    if((pins[k].netNo == i) && (layers[pins[k].layer].type == TYPE_POLY))
	      break;
	  if(k == PIN_NUM) {      /*  No pins on poly  */
	      rd_net[i].link = NO;
	      continue;
	  }
	  
	  if(rd_net[i].resistive_weight > 0) {
	      ForAllPins(k)
		if((pins[k].netNo == i) && (layers[pins[k].layer].type != TYPE_POLY))
		  break;
	      if(k == PIN_NUM) {   /*  All pins on poly  */
		  rd_net[i].resistive_weight = 0;
		  rd_net[i].link = NO;
		  continue;
	      }
	  }
      }
    
    /*  ============================================================
     *  Add the capacity weight to the nets critically coupled with VSS
     */
    if(status.BulkNet == NOTFOUND)
      status.BulkNet = name_to_net("VSS");
    if(status.BulkNet == NOTFOUND) {
	warnMessage("No net specified for substrate", 0);
	fflush(stderr);
    } else {
	ForAllNets( i )
	  if((wght=InquireCC_coupling(i, k, &maxcap)) > 0)
	    rd_net[i].capacitive_weight = max(rd_net[i].capacitive_weight, wght);
	initMessage("Assuming substrate biased with %s", rd_net[status.BulkNet].name);
    }

    return retValue;
}
