/*    mirror.c
 *
 *    Purpose: Mirroring procedure for symmetric circuits.
 *             The exported function is <RoadMirror>.
 *    Author:       Enrico Malavasi
 *    Ver. 3.20
 *    Last revised:  August 1991
 */
#include "port.h"
#include "utility.h"
#include "roadGlobals.h"
#define    Symm(r)      ((r == NULL)? NULL : (r + TOTAL_RECT_NUM))

static void
  pre_mirror()
/*  Rebuild the right half of the circuit in case of symmetric layout.
 *  Only the frame, obstacles and pins are rebuilt. The left side is
 *  left unchanged, so that the wires allocated there will be
 *  retrieved and mirrored.
 */
{
    int             i,j;         /* Loop counters ............... */
    int             netNo;
    int             wn;
    Rect            mr, r;

    /*  CONTROL. If any wire is partly located on the symmetry axis AND
     *           it is not self-symmetric, it must be shifted leftward. */
    ForAllWires( wn ) {
	if(((netNo=InquireSymmetryList(wireNetNo(wn))) != NOTFOUND) && (netNo != wireNetNo(wn))) {
	    free_wire(wn, 0);
	    for(j=0 ; (r=wSegmRect(wn, j))!=NULL ; j++) {
		if((rRightX(r) == SymmetryAxis) && (wSegmSide(wn, j) == EAST)) {
		    if(wire_SplitAndShift(wn, j, WEST) == SUCCESS) {
			wn--;   break;
		    }
		}
	    }
	    free_wire(wn, 1);
	}
    }

    /*  Reallocate memory for the 2nd half of the circuit.   */
    amplia_rect(TOTAL_RECT_NUM);

    /*  Rebuild obstacles.  */
    ReadMirrData();

    /*  Mirror the frame rects.  */
    for(i=0, j=TOTAL_RECT_NUM, r=rects ; i<TOTAL_RECT_NUM; i++, j++, r++) {
	mr = rects + j;
	
	if(rLeftX(r) == SymmetryAxis) {
	    rLeftX(r)  = cRightX(cells) + 1;
	    rLeftX(mr) = rRightX(r) = rLeftX(r) + 1;
	    rRightX(mr) = rLeftX(mr) + 1;
	    neighbor(r,WEST) = NULL;
	} else {
	    rLeftX(mr)  = cLeftX(cells)+cRightX(cells)-min(rRightX(r),SymmetryAxis);
	    rRightX(mr) = cLeftX(cells)+cRightX(cells) - rLeftX(r);
	}
	
	rLowerY(mr) = rLowerY(r);
	rUpperY(mr) = rUpperY(r);
	mr->layer = r->layer;
	mr->free  = 0;
	mr->flag  = 0;
	
	mr->adjacent[0] = Symm(r->adjacent[0]);
	mr->adjacent[1] = Symm(r->adjacent[1]);
	mr->adjacent[2] = Symm(r->adjacent[3]);
	mr->adjacent[3] = Symm(r->adjacent[2]);
	mr->adjacent[4] = Symm(r->adjacent[4]);
	mr->adjacent[5] = Symm(r->adjacent[5]);
	
	if(rRightX(r) == SymmetryAxis)
	  set_neighb(mr, r,  EAST);
    }
    DOUBLE(TOTAL_RECT_NUM);
    
    CreateNetList();      /* Build the full list of the unrouted nets */
    CreatePinSheetList(); /* Re-build the pin sheet image              */
}

/*  ===================================================================
 *  This function mirrors a given left-side wire into its corresponding
 *  right-side part.
 */
static void
  mirror(wn, netNo)
int      wn;             /* Index of the wire to be duplicated ........ */
int      netNo;          /* Net of the new wire ............. */
{
    int   j,               /* Loop counters ............................ */
          old_rect_num;    /* Half the current TOTAL_RECT_NUM .......... */

    /*  Allocate the new wire.    */
    rdWire[WIRE_NUM].rect = ALLOC(Rect, wireSize(wn)+1);
    rdWire[WIRE_NUM].side = ALLOC(int,  wireSize(wn)+1);

    /*  Create the specular copy.    */
    old_rect_num = TOTAL_RECT_NUM/2;
    for(j=0 ; j<wireSize(wn) ; j++)  {
	wSegmSide(WIRE_NUM, j) = wSegmSide(wn, j);
	wSegmRect(WIRE_NUM, j) = wSegmRect(wn, j) + old_rect_num;
	if(wSegmSide(wn, j) != NORTH)
	  wSegmRect(WIRE_NUM, j) = neighbor(wSegmRect(WIRE_NUM, j), WEST);
    }

    wireSize(WIRE_NUM)  = wireSize(wn);
    wSegmRect(WIRE_NUM, wireSize(WIRE_NUM)) = NULL;
    wireNetNo(WIRE_NUM) = netNo;
    wireFree(WIRE_NUM)  = 0; free_wire(WIRE_NUM, 1);
    WIRE_NUM++;
}

RoadMirror()
/*  This function looks whether it is necessary to duplicate the
 *  left half of the circuit which has already been built. In order
 *  to do it, a control over the accomplishment of the routing of
 *  all the symmetric nets is carried out. If such routing results
 *  complete, the duplication of all the structures is performed
 *  and every pair of half-wires produced by this process are linked and
 *  completed in a proper way.
 */
{
    int       i, j, k;        /*  Loop counters. */
    int       OldWireNum;
    int       netNo;
    int       pi, pj;

    /*  If the symmetry flag is active AND all the symmetric wires have been
     *  connected, the duplication process can be carried on.
     */
    unless(GetSymmetryFlag())                            return;
    if((i=SuggestNextNet()) != NOTFOUND)
      if(InquireSymmetryList(i) != NOTFOUND)  return;

    /*  All the existing pins have already been connected, but the symmetry
     *  flag is still active. Therefore the duplication process must be
     *  accomplished. SYMM_FLAG is reset and it will not be set again.
     */
    SetSymmetryFlag(NO);

    /*  Duplicate frame and cells.  */
    initMessage("Mirroring the half circuit ... ", NOTFOUND);
    pre_mirror();

    /*  DUPLICATE WIRES.    */
    OldWireNum = WIRE_NUM;
    for(i=0 ; i<OldWireNum ; i++)
      if((netNo=InquireSymmetryList(wireNetNo(i))) != NOTFOUND)
	if(wSegmRect(i, 0) != NULL) {     /*  Not a Dummy wire  */
	    mirror(i, netNo);
	}
    ForAllWires(i) {
	if(wSegmRect(i, 0) == NULL)    /*   Dummy wire   */
	  delete_wire(i--);
    }
    if( INTERACTIVE_MODE )    { RoadZoom(0);show_all(-1); }
    
    /*  Fragmentation is necessary because of the new pins.   */
    fragment();
    CreatePinSheetList();
    CreateCongestionAreas();

    /*  Delete the unnecessary half-wires.  */
    ForAllWires(k) {
	if(wSegmRect(k, 0) == NULL) continue;   /* Empty wire  */
	ForAllPins(i)
	  if(pins[i].netNo == wireNetNo(k)) {
	      for(j=0 ; wSegmRect(k, j)!=NULL ; j++)
		if(pins[i].rect == wSegmRect(k, j))    break;  /* It is necessary */
	      if(wSegmRect(k, j) != NULL)              break;  /* ............... */
	  }
	if(i == PIN_NUM)
	  delete_wire(k--);
    }
    
    ForAllWires(k) {
	PruneWire(k);    /* At the end all wires are 1-freed */
	AddNetListItem(wireNetNo(k));
    }
    
    /*  Link the half wires if they are in touch */
    ForAllWires(i) {
	for(j=i+1 ; j<WIRE_NUM ; j++) {
	    if(wireNetNo(i) == wireNetNo(j)) {
		for(pi=0 ; pi<wireSize(i) ; pi++) {
		    for(pj=0 ; pj<wireSize(j) ; pj++) {
			if(wSegmRect(i, pi) == wSegmRect(j, pj))   break;
		    }
		    if(pj < wireSize(j)) {
			for(pj=0 ; pj<wireSize(j) ; pj++) {
			    add_to_wire(i, wSegmRect(j, pj), wSegmSide(j, pj));
			}
			if(j < (--WIRE_NUM)) {
			    SWAP(_rdWire, rdWire[j], rdWire[WIRE_NUM]);
			}
			FREE(rdWire[WIRE_NUM].rect);
			FREE(rdWire[WIRE_NUM].side);
			j--;
			break;
		    }
		}
	    }
	}
    }
}
