/*      netlist.c
 *
 *      Author: Enrico Malavasi
 *      Ver. 3.00
 *      Last revised: January, 1991
 *
 *      Purpose: Management of the priority list.
 */
#include "port.h"
#include "utility.h"
#include "roadGlobals.h"

typedef  int  RoadPriority;

#define    WeightResistance(x)       (x)*10
#define    WeightCapacitance(x)      (x)*10

#define    PRIO_SYMMETRY   100.0
#define    PRIO_SUPPLY      50.0
#define    PRIO_COUPLING    10.0
#define    PRIO_RESISTIV    10.0
#define    PRIO_CAPACITI    10.0

static     int   *NetList = NULL;        /*  List of rd_net indexes  */
static     int    NetListSize = 0;

/*  ============================================================
 *  Compute the priority of the given net
 */
static RoadPriority
netPriority(netNo)
int  netNo;
{
    RoadWeight  m;
    RoadPriority  prio = 0;
    pFarad        maxcap;
    int           i;

    if(isSupply(netNo))   prio += PRIO_SUPPLY;
    if(GetSymmetryFlag())
      if(InquireSymmetryList(netNo) != NOTFOUND)
	prio += PRIO_SYMMETRY;
    prio += (rd_net[netNo].resistive_weight * PRIO_RESISTIV);
    prio += (rd_net[netNo].capacitive_weight* PRIO_CAPACITI);
    for(i=0 ; i<NET_NUM ; i++) {
	if(i != netNo) {
	    if((m=InquireCC_coupling(netNo, i, &maxcap)) > 0) {
		prio += (m * PRIO_COUPLING);
	    }
	}
    }
    return(prio);
}

/*  ============================================================
 *  Creation or updating of the NetList.
 */
void
CreateNetList()
{
    int i;

    NetList = REALLOC(int, NetList, NET_NUM);
    NetListSize = 0;
    for(i=0 ; i<NET_NUM ; i++)
      AddNetListItem(i);
}

/*  ============================================================
 *  Add a new item to the NetList, after checking that
 *  it is not there yet.
 */
AddNetListItem(NetNo)
int NetNo;         /*  Net index  */
{
    PATH  pth;
    int   i;

    if(InquireNetList(NetNo) != NOTFOUND)   return;    /* It is already there */
    if((i=net_to_wire(NetNo, 0)) != NOTFOUND) {
	pth = BuildPathFromWire( i );
	if(IsPathComplete(pth, NetNo)) {
	    FREE(pth);
	    return;            /* Already fully routed */
	}
	FREE(pth);
    }
    /*  Insert the new item and sort the list.   */
    NetList[NetListSize++] = NetNo;
    for(i=NetListSize-2 ; i>=0 ; i--) {
	if(netPriority(NetList[i]) > netPriority(NetNo))    break;
	NetList[i+1] = NetList[i];
	NetList[i]   = NetNo;
    }
}
/*  ============================================================
 *  Return the position of a net in the NetList
 */
InquireNetList(NetNo)
int NetNo;        /*  Net index  */
{
  int i;

  if(NetNo == NOTFOUND)  return(NOTFOUND);
  for(i=0 ; i<NetListSize ; i++)
    if(NetList[i] == NetNo)  return(i);
  return(NOTFOUND);
}
/*  ============================================================
 *  Return the net at the head of the NetList
 */
getFirstNetListItem()
{
  int i;

  if(NetListSize == 0) return(NOTFOUND);
  for(i=1 ; i<NetListSize ; i++) {
    if(netPriority(NetList[i]) < netPriority(NetList[0]))
      break;
    if(netCongestion(NetList[i]) > netCongestion(NetList[0])) {
      SWAP(int, NetList[i], NetList[0]);
    } else {
      if(netCongestion(NetList[i]) == netCongestion(NetList[0])) {
	if(netLength(NetList[i]) < netLength(NetList[0]))
	  SWAP(int, NetList[i], NetList[0]);
      }
    }
  }
  return(NetList[0]);
}
/*  ============================================================
 *  Delete the given net from the Netlist
 */
deleteNetFromNetList(NetNo)
int NetNo;
{
  int pos;   /*  Position of the net in the NetList  */

  if((pos=InquireNetList(NetNo)) == NOTFOUND)
    return;
  NetListSize--;
  while(pos < NetListSize) {
    NetList[pos] = NetList[pos+1];
    pos++;
  }
}
GetNetListSize()
{   return(NetListSize);  }
