/*     pathutil.c
 *
 *     Purpose: Utilities for the path processing.
 *     Author: Enrico Malavasi
 *     Ver. 3.20
 *     Last revised: August 1991
 */
#include "port.h"
#include "utility.h"
#include "roadGlobals.h"

int
  path_length(p)
PATH p;
/*  Return the length of a null-terminated path  */
{
    int i;
    for(i=0 ; p[i]!=NULL ; ) i++;
    return i;
}

PATH
  rdPathSave( new, pth )
PATH new;
PATH pth;
{
    int   i, k;

    k = path_length( pth );
    new = REALLOC(Rect, new, k+1);
    for(i=0 ; i<k ; i++) {
	new[i] = pth[i];
    }
    new[k] = NULL;
    return new;
}

static void
  CatPaths(a, b)
PATH a, b;
/*  Catenate two paths. The resulting path is stored in <a>, which is supposed
 *  to be already allocated. The size of <a> is returned.  */
{
    int i, l;
    
    l = path_length(a);
    for(i=0 ; b[i]!=NULL ; i++) {
	a[l++] = b[i];
    }
    a[l] = NULL;
}
PATH
  FullCatPaths(a, b)
PATH a, b;
{
    a = REALLOC(Rect, a, (path_length(a)+path_length(b)+1));
    CatPaths(a, b);
    return a;
}

/*  ============================================================
 *  This function deletes all the repeated elements from a given
 *  path. The new size of the cleaned path is returned.
 */
int
  CleanPath(p)
PATH p;
{
    int i, j, l;

    l = path_length(p);
    for(i=0 ; i<l ; i++)
      for(j=i+1 ; j<l ; j++)
	if(p[i] == p[j])      p[j--] = p[--l];
    p[l] = NULL;
    return l;
}

RoadStatus
  IntersectingPaths(p1, p2)
PATH  p1, p2;
/*  Check whether the given paths intersect with each other.  */
{
    int i, j;
    
    for(i=0 ; p1[i]!=NULL ; i++)
      for(j=0 ; p2[j]!=NULL ; j++)
	if(p1[i] == p2[j])   return YES;
    return NO;
}
RoadStatus
  OverlappingPaths(p1, p2)
PATH  p1, p2;
{
    int i, j;
  
    for(i=0 ; p1[i]!=NULL ; i++)
      for(j=0 ; p2[j]!=NULL ; j++)
	if((p1[i] == neighbor(p2[j], UP)) || (p1[i] == neighbor(p2[j], DOWN)))   return YES;
    return NO;
}

RoadTerm
  *PinIsIn(r, netNo)
Rect     r;
int      netNo;
/*  Return the pin, attached to the net, contained in the rect.
 *  If no pins are in that rect, the NULL pointer is returned. */
{
    int i;         /* Loop counter  ............. */
    RoadTerm *p;   /* Term found in the given rect */

    if(r == NULL)                            return NULL;
    if(! RD_CHECKRECTFLAG(r,RECTFLAG_TERM))  return NULL;
    if( netNo == NOTFOUND )                  return NULL;
    ForAllPins( i ) {
	p = pins+i;
	if(p->rect == r)
	  if(p->netNo == netNo)   return p;
    }
    return NULL;
}

PATH
  BuildPathFromWire(wn)
int   wn;               /*  Given wire (input) ..  */
/*  Return the corresponding cleaned path. The path should be already allocated. */
{
    PATH  pth;
    int  i;     /*  Loop counter          */
    int  dim = 0;   /*  Path size             */
    Rect r;     /*  Wire and Path element */

    /*   Build the path using  the given wire.    */
    pth = ALLOC(Rect, wireSize(wn)*3+1);
    for(i=0 ; i<wireSize(wn) ; i++) {
	r = wSegmRect(wn, i);
	pth[dim++] = r;
	if((wSegmSide(wn, i) == EAST) && (neighbor(r, SOUTH) != NULL))
	  pth[dim++] = neighbor(r, SOUTH);
	if up_or_down(wSegmSide(wn, i))
	  pth[dim++] = neighbor(r, wSegmSide(wn, i));
    }
    pth[dim] = NULL;
    dim = CleanPath( pth );
    pth = REALLOC(Rect, pth, dim+1);
    return  pth;
}

static RoadStatus
  ElectricallyEquivalent(r1, r2, NetNo)
Rect  r1, r2;
int   NetNo;
/*  Check whether the two path elements are electrically equivalent.  */
{
    RoadTerm *p1, *p2;

    if(r1 == r2) return YES;                     /* The same path element. */
    if(((p1=PinIsIn(r1, NetNo)) != NULL) &&
       ((p2=PinIsIn(r2, NetNo)) != NULL) &&
       (p1->equivNo == p2->equivNo))
      return YES;                                /* Equiv. pins in them */
    return NO;                                   /* Not equiv. */
}
static RoadStatus
  PinInPathEquivalent(p, pt, NetNo)
RoadTerm  *p;                 /*  Given pin          */
PATH       pt;                /*  Path to be checked */
int        NetNo;
/*  Check whether the pin or one of its equivalents is touched by the path */
{
    int i;
    
    for(i=0 ; pt[i]!=NULL ; i++)
      if(ElectricallyEquivalent(p->rect, pt[i], NetNo))  return YES;
    return NO;
}
RoadStatus
  IsPathComplete(pt, NetNo)
PATH pt;             /*  Path to be checked */
int  NetNo;
/*  Chack whether the path is complete for the given net */
{
    int i;
    
    ForAllPins(i)
      if(pins[i].netNo == NetNo)
	unless(PinInPathEquivalent(pins+i, pt, NetNo))
	  return NO;
    return YES;
}

int
  rectInPath(r, p)
Rect r, *p;
{
    int  i;
    
    for(i=0 ; p[i]!=NULL ; i++)
      if(p[i] == r)
	return  i;
    return NOTFOUND;
}

delFromPath(r, p)
Rect r, *p;
{
    int  i, k;
    
    for(k=0 ; p[k]!=NULL ; k++) ;
    for(i=0 ; i<k ; i++)
      if(p[i] == r)
	p[i--] = p[--k];
    p[k] = NULL;
}
