/*     propagate.c
 *
 *     Purpose: Algorithm for the depth-search maze routing.
 *         Here the A* algorithm is implemented. Array
 *         myPath contains all the grid frames explored in the
 *         search algorithm. Array WaveFront contains the wave fronts
 *         to be explored for the best element to propagate.
 *         Each element of myPath contains a pointer <father> pointing
 *         to its generator in the previous wave front.
 *     Author: Enrico Malavasi
 *     Ver. 3.20
 *     Last revised: August 1991
 */
#include "port.h"
#include "utility.h"
#include "road.h"
#include "roadGlobals.h"
#include "roadMacros.h"

extern RoadCost    *rdQuota;

typedef struct _rdPath {
    Rect       rect;      /*  Grid element  */
    RoadCoord  hhh;       /*  Estimated distance to the closest target */
    int        dad;       /*  Generator path element */
    char       father;    /*  Direction to the former element (generator)  */
} _rdPath;

static _rdPath  *rdPath    = NULL;
static int      *WaveFront = NULL;
static PATH      target    = NULL;
static RoadCost *tgtCost   = NULL;
static int    rdTargetNum;            /*  No. of target elements      */
static int    rdWaveNum;              /*  No. of wave front elements  */
static int    rdPathNum;              /*  No. of path elements        */
static int    rdWaveMaxSize = 0;      /* No. of allocated wave front elem.  */
static int    rdPathMaxSize = 0;      /* No. of allocated path elements     */

#define ForAllTargets(i)       for(i=0 ; i<rdTargetNum ; i++)
#define  QuotaOf(r)            rdQuota[(r)-rects]
#define  QuotaNo(n)            rdQuota[n]

#define viaCost(l,m)  (((l)==(m)) ? 0 : layers[min((l),(m))].via_cost)

static
  isFather(a, b)
int a, b;
{
    if(a != NOTFOUND)
      while((a=rdPath[a].dad) != NOTFOUND)
	if(a == b)  return YES;
    return NO;
}
static void
  lowerFollowers(pathNo, diff)
int      pathNo;
RoadCost diff;
{
    int  i;

    for(i=pathNo+1 ; i<rdPathNum ; i++)
      if(rdPath[i].dad == pathNo) {
	  QuotaOf(rdPath[pathNo].rect) -= diff;
	  lowerFollowers(i, diff);
      }
}

static void
  modifyMyPath(pathNo, rec, father, dad, hhh, fff)
int       pathNo;
Rect      rec;
int       father;
int       dad;
RoadCoord hhh, fff;
{
    rdPath[pathNo].rect        = rec;
    rdPath[pathNo].father      = father;
    rdPath[pathNo].dad         = dad;
    rdPath[pathNo].hhh         = hhh;
    QuotaOf(rec)               = fff;
}

static int
  addToRdPath(r, father, dadNo, hhh, fff)
Rect      r;
int       father;
int       dadNo;
RoadCoord hhh;
RoadCoord fff;
/*  Add an item to the rdPath list  */
{
    while(rdPathNum >= rdPathMaxSize) {
	rdPathMaxSize += 1000;
	rdPath = REALLOC(_rdPath, rdPath, rdPathMaxSize);
    }
    modifyMyPath(rdPathNum++, r, father, dadNo, hhh, fff);
    return  (rdPathNum-1);
}
static void
  addToWaveFront(pathNo)
int   pathNo;
{
    while(rdWaveNum >= rdWaveMaxSize) {
	rdWaveMaxSize += 1000;
	WaveFront = REALLOC(int, WaveFront, rdWaveMaxSize);
    }
    WaveFront[rdWaveNum++] = pathNo;
}

static void
  delFromWaveFront(wfn)
int  wfn;
{
    rdWaveNum--;
    if(wfn != rdWaveNum)
      WaveFront[wfn] = WaveFront[rdWaveNum];
}
static int
  findPathNo(r)
Rect r;
{
    int i;

    for(i=0 ; i<rdPathNum ; i++)
      if(rdPath[i].rect == r)
	return i;
    return NOTFOUND;
}

/*  ============================================================
 *  This function returns the estimate h(x) of the distance between
 *  a given grid node and the closest target, weighted by layer cost.
 *  h(x) = m(x,t) + tgt(t) + v(x)
 *  where m(x,t)  is the Manhattan distance between x and r
 *        tgt(x)  is the "target cost"
 *        v(x)    is the cost of a via if x is a via, 0 otherwise.
 */
static RoadCoord
  targetDist(r)
Rect r;
{
  int i, l;
  RoadCoord  theDist,
             bestDist;

  l = rLayer(r);
  ForAllTargets(i) {
    theDist = DistToCost(rect_dist(target[i], r), l) + tgtCost[i] +
      viaCost(l, rLayer(target[i]));
    if(i == 0) bestDist = theDist;
    else       bestDist = min(bestDist, theDist);
  }
  return(bestDist);
}

/*  ============================================================
 *  This function selects, among the existing front elements,
 *  the best one to propagate on.
 */
static int
  selBestFrontElem()
{
    RoadCost     q, qi, hhh;
    register int i;
    int          theBest=0;

    if(rdWaveNum <= 0)   return NOTFOUND;
    q    = QuotaOf(rdPath[WaveFront[theBest]].rect);
    hhh  = rdPath[WaveFront[theBest]].hhh;
    for(i=1 ; i<rdWaveNum ; i++) {
	if(((qi=QuotaOf(rdPath[WaveFront[i]].rect)) < q) ||
	   ((qi == q) && (rdPath[WaveFront[i]].hhh < hhh))) {
	    theBest = i;
	    q    = qi;
	    hhh  = rdPath[WaveFront[theBest]].hhh;
	} 
    }
  
    return(theBest);
}

/*  ============================================================
 *  This is the key function of the propagation process. It
 *  assigns quotas to the grid nodes.
 *        fff = ggg + hhh,
 *  where  ggg = distance of the rect from the source node
 *         hhh = estimate of the distance to the target.
 *  This function propagates one wave front element, deletes it from
 *  the WaveFront list and adds the generated nodes to array myPath.
 */
#define    ACCELERATOR     
static int
  propagateFromBest(wfn)
int  wfn;
{
    int        side, oldDir, pathNo, dadNo;
    RoadCoord  hhh, ggg, fff, diff;
    Rect       dad, son;

    dadNo  =  WaveFront[wfn];
    dad    =  rdPath[dadNo].rect;
    oldDir =  (- rdPath[dadNo].father);
    ggg    = (QuotaOf(dad) - rdPath[dadNo].hhh) ACCELERATOR;

    if(oldDir != NODIRECTION)
      if((son=free_neighbor(dad, oldDir)) != NULL) {
	  side = oldDir;
	  if(QuotaOf(son) == 0) {
	      hhh = targetDist(son);
	      if((fff=connection_cost(dad, side)) < MAX_COST) {
		  addToWaveFront( addToRdPath(son, -oldDir, dadNo, hhh, fff+hhh+ggg) );
		  if(QuotaOf(son) <= QuotaOf(dad))  return;
	      }
	  } else if(QuotaOf(son) > QuotaOf(dad)) {
	      pathNo  = findPathNo(son);
	      hhh = rdPath[pathNo].hhh;
	      if((fff=connection_cost(dad, side)) < MAX_COST) {
		  fff += (hhh + ggg);
		  if(fff < QuotaOf(son))
		    if(isFather(pathNo, dadNo) == NO) {
			diff = QuotaOf(son) - fff;
			modifyMyPath(pathNo, son, (-side), dadNo, hhh, fff);
			lowerFollowers(pathNo, diff);
			if(QuotaOf(son) <= QuotaOf(dad))  return;
		    }
	      }
	  }
      }
    
    delFromWaveFront(wfn);
    six_sides(side)
/*    if((side != oldDir) && (side != (-oldDir)))   */
      if(side != oldDir)
	if((son=free_neighbor(dad, side)) != NULL) {
	    if(QuotaOf(son) == 0) {
		hhh = targetDist(son);
		if(rLayer(son) == rLayer(dad))
		  hhh++;           /* Weight for a corner */
		if((fff=connection_cost(dad, side)) < MAX_COST) {
		    addToWaveFront( addToRdPath(son, -side, dadNo, hhh, fff+hhh+ggg) );
		}
	    } else if(QuotaOf(son) > QuotaOf(dad)) {
		pathNo  = findPathNo(son);
		hhh = rdPath[pathNo].hhh;
		if(rLayer(son) == rLayer(dad))
		  hhh++;           /* Weight for a corner */
		if((fff=connection_cost(dad, side)) < MAX_COST) {
		    fff += (hhh + ggg);
		    if(fff < QuotaOf(son))
		      if(isFather(pathNo, dadNo) == NO) {
			  diff = QuotaOf(son) - fff;
			  modifyMyPath(pathNo, son, (-side), dadNo, hhh, fff);
			  lowerFollowers(pathNo, diff);
		      }
		}
	    }
	}
}

/*  ------------------------------------------------------------
 *  This routine correctly allocates the structures to be used
 *  in the propagation process. It also prepares the first front wave
 *  to be used to start the propagation procedure.
 */
define(NetNo, wList)
int   NetNo;
DTERM *wList;    /* List of partially-routed sub-nets. .......... */
{
    int   i, j, k;
    Pins  p;

    ForAllRects(i) {
	QuotaNo(i) = 0;
    }
    rdPathNum   = 0;
    rdTargetNum = 0;
    rdWaveNum   = 0;

    /*  1.) Define the propagating wave */
    /*  Bring the smallest distributed terminal to the first position  */
    j = path_length(wList[0].rect);
    for(i=1 ; wList[i].rect[0]!=NULL ; i++) {
	if(j == 1) break;
	if((k=path_length(wList[i].rect)) < j) {
	    j = k;
	    SWAP(DTERM, wList[i], wList[0]);
	}
    }
    for(i=0 ; wList[0].rect[i]!=NULL ; i++) {
	addToRdPath(wList[0].rect[i], NODIRECTION, NOTFOUND, 0, 0);
    }

    /*  2.) Define the set of targets and their costs */
    target = REALLOC(Rect, target, 1);
    target[0] = NULL;
    for(j=1 ; wList[j].rect[0]!=NULL ; j++) {
	if(wList[j].pin->equivNo == wList[0].pin->equivNo) {
	    for(i=0 ; wList[j].rect[i]!=NULL ; i++) {
		addToRdPath(wList[j].rect[i], NODIRECTION, NOTFOUND, 0, 0);
	    }
	}
	else {
	    target = FullCatPaths(target, wList[j].rect);
	    rdTargetNum = path_length(target);
	}
    }
    if(rdTargetNum == 0)    return FAILURE;
    
    tgtCost = REALLOC(RoadCost, tgtCost, rdTargetNum);
    for(i=0 ; i<rdTargetNum ; i++) {
	tgtCost[i] = 0;
	if((p=PinIsIn(target[i], NetNo)) != NULL)
	  tgtCost[i] = p->exitCost;
    }
    
    /*  3.) Allocate the wave fronts */
    for(i=0 ; i<rdPathNum ; i++) {
	QuotaOf(rdPath[i].rect) = rdPath[i].hhh  = targetDist(rdPath[i].rect);
	four_sides(j)
	  if(neighbor(rdPath[i].rect, j) == NULL) {
	      rdPath[i].father = j;
	      break;
	  }
	addToWaveFront(i);
    }
    
    return SUCCESS;
}

/*  ============================================================
 *  Propagation routine.
 *  When this function is invoked, it performs a propagation cycle.
 *  Starting from the initial wave front, prepared by "define", the
 *  propagation is continued until one of the targets is reached. Such
 *  target is then returned.
 */
Rect propagate()
{
    int wfn,   /*  Next wave front element to be propagated  */
        i;         /*  loop counter  */

/*  ------------------------------------------------------------
 *  Among the front elements, select the best one to propagate
 */
    loop {
	if((wfn=selBestFrontElem()) == NOTFOUND) return(NULL);
	if(RD_CHECK_FLAG(GRAPH_VERB)) { show_rect(rdPath[WaveFront[wfn]].rect); }
	ForAllTargets(i)
	  if(rdPath[WaveFront[wfn]].rect == target[i])
	    return(target[i]);
	propagateFromBest(wfn);
    }
}

/*  ============================================================
 *  Backtrace routine. Array rdPath is backtraced starting from its
 *  first element to the one corresponding to the source, whose
 *  <father> pointer points to NULL.
 */
full_backtrace(retPath)
PATH retPath;
{
  int  i,
       psize = 0;
  Rect r;

  for(r=retPath[0] ; r!=NULL ; )
    for(i=0 ; i<rdPathNum ; i++)
      if(rdPath[i].rect == r) {
	retPath[psize++] = r;
	if(rdPath[i].dad == NOTFOUND)    r = NULL;
	else            r = rdPath[rdPath[i].dad].rect;
	break;
      }
  retPath[psize] = NULL;
}
