/*    roadbox.c
 *
 *    Author:       Enrico Malavasi
 *    Ver.:         2.10
 *    Last revised: September, 1990
 *
 *    Purpose:   Define and transform protection frame boxes
 */
#include "port.h"
#include "utility.h"
#include "roadGlobals.h"
#include "space.h"

HalfBox(BB, dir)
RoadBox *BB;            /*   Box  */
int      dir;           /*   Shift direction  */
{
   switch(dir) {
   case NORTH:
     BB->y1 = (BB->y1+BB->y2)/2-1;  break;
   case SOUTH:
     BB->y2 = (BB->y1+BB->y2)/2+1;  break;
   case EAST :
     BB->x1 = (BB->x1+BB->x2)/2-1;  break;
   case WEST :
     BB->x2 = (BB->x1+BB->x2)/2+1;  break;
   }
}

ShiftBox(BB, dir, amount)
RoadBox *BB;            /*   Box  */
int      dir;           /*   Shift direction  */
RoadCoord    amount;        /*   Shift amount  */
{
  switch(dir) {
  case NORTH:
    BB->y1 += amount; BB->y2 += amount;  break;
  case SOUTH:
    BB->y1 -= amount; BB->y2 -= amount;  break;
  case EAST:
    BB->x1 += amount; BB->x2 += amount;  break;
  case WEST:
    BB->x1 -= amount; BB->x2 -= amount;  break;
  }
}

sliceBox(r, dir, BB)
Rect  r;
int      dir;           /*   Shift direction  */
RoadBox *BB;            /*   Box  */
{
  Rect s;

  while((s=free_neighbor(r, WEST)) != NULL) {
    if(rRightX(s) <= BB->x1) break;
    r = s;
  }
  if(rLeftX(r) > BB->x1)   BB->x1 = rLeftX(r);

  for(s=r ; (s!=NULL)&&(rLeftX(s)<BB->x2) ; s=neighbor(s,EAST)) {
    r = follow_FreeNeighbor(s, SOUTH);
    if(rLowerY(r) > BB->y1)     BB->y1 = rLowerY(r);

    r = follow_FreeNeighbor(s, NORTH);
    if(rUpperY(r) < BB->y2)     BB->y2 = rUpperY(r);
  }
  if(rRightX(r) < BB->x2)     BB->x2 = rRightX(r);
}

/*  ============================================================
 *  This function computes the bounding boxes of the safety frame
 *  for a wire segment.
 */
void
WireSegmentToBox(r, side, room, BB, bUp, HalfDir, SliceDir)
Rect     r;             /*  Segment rect ................... */
int      side;          /*  Segment side ................... */
RoadCoord    room;      /*  Required room .................. */
RoadBox *BB, *bUp;      /*  Bounding box(es) ............... */
int      HalfDir;       /*  Halving direction  ............. */
int      SliceDir;      /*  Slicing direction .............. */
{
  if(room <= 0)  room = spacingBetween(rLayer(r), side, rLayer(r), NORTH);
  BB->x1 = ((side == NORTH) ? r->x1 : r->x2) - room;
  BB->y1 = ((side == EAST)  ? r->y1 : r->y2) - room;
  BB->x2 = r->x2 + room;
  BB->y2 = r->y2 + room;
  BB->z   = r->layer;
  if(bUp == NULL)    return;
  if(HalfDir != NODIRECTION)     HalfBox(BB, HalfDir);
  if(SliceDir != NODIRECTION)    sliceBox(r, SliceDir, BB);
  if(side == UP) {
    r = neighbor(r, UP);
    BB->z++;
    if(SliceDir != NODIRECTION)  sliceBox(r, SliceDir, BB);
    (*bUp) = (*BB);
    BB->z--;
  }
}

/*  ============================================================
 *  The following two functions retrieve an intersection between
 *  an item (cell and wires respectively) and a given box.
 */
CellInBox(BB, begin)
RoadBox *BB;               /*  Safe bounding box  */
int      begin;
{
  int  i;

  if(begin < 0)  begin = 0;
  if(begin == 0) {
    if((cells[0].x1 > BB->x1) || (cells[0].x2 < BB->x2) ||
       (cells[0].y1 > BB->y1) || (cells[0].y2 < BB->y2))
/*      unless((GetSymmetryFlag()) && (cells[0].x2 < BB->x2)) */
	return(0);
    begin = 1;
  }
  for(i=begin ; i<=SUB_CELL_NUM ; i++)
    if(cells[i].layer == BB->z)
      if((cells[i].x1 < BB->x2) && (cells[i].x2 > BB->x1) &&
	 (cells[i].y1 < BB->y2) && (cells[i].y2 > BB->y1))
	return(i);
  return(NOTFOUND);
}

static int
RectToWpos(r, side, wp)
Rect           r;
int            side;
_wirePos *wp;
{
  (wp->wn) = (-1);
  while((wp->wn=net_to_wire(rNet(r), wp->wn+1)) != NOTFOUND)
    if((wp->pos=wire_segment_position(wp->wn, r, side, NOTFOUND)) != NOTFOUND)
      break;
}

/*  ============================================================
 *  Assumptions:  r != NULL
 *                if r is out of the box, it is possible to get there
 *                  without meeting obstacles.
 *                wn != NOTFOUND
 */
WireInBox(wn, BB, exclude, r, wp)
int      wn;
RoadBox *BB;
int      exclude;
Rect     r;
_wirePos *wp;
{
    Rect          s;
    int           side;

    while(rRightX(r) <= BB->x1)  r = neighbor(r, EAST);
    while(rUpperY(r) <= BB->y1)  r = neighbor(r, NORTH);
    while((s=neighbor(r,WEST)) != NULL) {
	if(rRightX(s) <= BB->x1) break;
	r = s;
    }
    while((s=neighbor(r,SOUTH)) != NULL) {
	if(rUpperY(s) <= BB->y1) break;
	r = s;
    }
    
    for(side=NODIRECTION, s=r ; (s!=NULL)&&(rLeftX(s)<BB->x2) ; ) {
	
	if(exclude != UP)
	  if((rUpperY(s) < BB->y2) && (rRightX(s) < BB->x2)) {
	      if(ViaUp(s)) {
		  side = UP; break;
	      }
	      if(ViaDown(s)) {
		  s = neighbor(s, DOWN);
		  side = UP; break;
	      }
	  }
	if(exclude != NORTH)
	  if(rUpperY(s) < BB->y2)
	    if(free_neighbor(s, NORTH) == NULL)
	      if(neighbor(s, NORTH) != NULL) {
		  side = NORTH; break;
	      }
	if(exclude != EAST)
	  if(rRightX(s) < BB->x2)
	    if(free_neighbor(s, EAST) == NULL)
	      if(neighbor(s, EAST) != NULL) {
		  side = EAST; break;
	      }
	
	s = neighbor(s, NORTH);
	if((s == NULL) || (rLowerY(s) >= BB->y2)) {
	    while((neighbor(r, EAST) == NULL) &&
		  (rUpperY(r) < BB->y2))
	      if((r=neighbor(r, NORTH)) == NULL)  break;
	    if(r != NULL) r = neighbor(r, EAST);
	    s = r;
	}
    }

    if(side == NODIRECTION)   return NOTFOUND;
    RectToWpos(s, side, wp);
    common = wp->pos;
    return wp->wn;
}

static RoadBox *WB = NULL;
#define  SAFEGUARD  100
updateWireBoxes(only)
int only;
{
  int wn, pos;

  WB = REALLOC(RoadBox, WB, WIRE_NUM);
  ForAllWires( wn )
    if((only == NOTFOUND) || (only == wn)) {      
      WB[wn].x1 = cRightX(cells);
      WB[wn].y1 = cUpperY(cells);
      WB[wn].x2 = cLeftX(cells);
      WB[wn].y2 = cLowerY(cells);
      for(pos=0 ; pos<wireSize(wn) ; pos++) {
	if(rLeftX(wSegmRect(wn, pos)) < WB[wn].x1)
	  WB[wn].x1 = rLeftX(wSegmRect(wn, pos));
	if(rRightX(wSegmRect(wn, pos)) > WB[wn].x2)
	  WB[wn].x2 = rRightX(wSegmRect(wn, pos));
	if(rLowerY(wSegmRect(wn, pos)) < WB[wn].y1)
	  WB[wn].y1 = rLowerY(wSegmRect(wn, pos));
	if(rUpperY(wSegmRect(wn, pos)) > WB[wn].y2)
	  WB[wn].y2 = rUpperY(wSegmRect(wn, pos));
      }
      WB[wn].x1 -= SAFEGUARD;
      WB[wn].y1 -= SAFEGUARD;
      WB[wn].x2 += SAFEGUARD;
      WB[wn].y2 += SAFEGUARD;
    }
}

checkWireBox(wn, BB)
int wn;
RoadBox *BB;
{
  if(wn < WIRE_NUM) {
    if(BB->x2 <= WB[wn].x1)  return NO;
    if(BB->x1 >= WB[wn].x2)  return NO;
    if(BB->y2 <= WB[wn].y1)  return NO;
    if(BB->y1 >= WB[wn].y2)  return NO;
  }
  return YES;
}
