/*      route.c
 *
 *	Author: Enrico Malavasi
 *      Ver. 3.0
 *      Last revised: January, 1991
 *
 *	Purpose: Perform the whole routing procedure for a single net
 *		 with the perimetral method.
 */
#include "port.h"
#include "utility.h"
#include "roadGlobals.h"
#include "roadTechnol.h"

int         Kwave;

/*  =============================================================
 *  This function creates an array of all the pins of the current
 *  net. The pin array must have already been allocated. The
 *  returned value is the size of the pin array.
 */
static int
  MakePinArray(PinArray, netNo)
RoadTerm **PinArray;          /*  Pin array   */
int        netNo;
{
    int PinNo = 0;       /* Pin number (returned)   */
    int i;               /* Loop counter            */

    ForAllPins( i )
      if(pins[i].netNo == netNo)
	PinArray[PinNo++] = pins+i;
    PinArray[PinNo] = NULL;
    
    return(PinNo);
}

static void
  SetKwave()
{
  int i;

  for(i=0,Kwave=(-1) ; i<LAYER_NUM ; i++)
    if(layers[i].cost >= 500)
      Kwave = i;
}

/*  ============================================================
 *  The following functions save and restore the set of 
 *  equivalences of the whole set of pins.
 */
static int *eeqq = NULL;
static void
  SaveEquivalences()
{
    int i;
    eeqq = REALLOC(int, eeqq, PIN_NUM);
    ForAllPins(i)
      eeqq[i] = pins[i].equivNo;
}
static void
  RestoreEquivalences()
{
    int i;
    ForAllPins(i)
      pins[i].equivNo = eeqq[i];
    FREE(eeqq);
    eeqq = NULL;
}

/*  ============================================================
 *  This function "cleans" the <layer> field of every element of
 *  each wire.
 */
static void
  CleanWires()
{
    int  i, j;
    
    ForAllWires( i )
      for(j=0 ; wSegmRect(i, j)!=NULL ; j++)
	wSegmLayer(i, j)  &= 0x0F;
}

/*  =================================================================
 *  This function takes two equivalence sets and links them, creating
 *  a new one. All the pins references are consequently changed.
 */
join_equivalence(eq1, eq2)
RoadTerm *eq1, *eq2;    /* Are two terms whose equival. sets are to be linked */
{
  int  eq;
  int  i;

/*  ------------------------------------------------------------
 *  Check whether they are already equivalent,
 */
  if(eq1->equivNo == eq2->equivNo)	return;

/*  ------------------------------------------------------------
 *  Set the whole set of pins references.
 */
  eq = eq1->equivNo;
  ForAllPins(i)
    if(pins[i].equivNo == eq)
      pins[i].equivNo = eq2->equivNo;
}

/*  ============================================================
 *  The following function changes the netNo- field of all
 *  pins on the net linked to the specified net
 */
static void
  DivideLinkedPins(netNo)
int netNo;
{
    int  i, j, nNet;
    int  wn;
    Pins p;

    nNet = (-1 - netNo);
    /*  Mark all the pins on POLY */
    if(rd_net[netNo].link == YES) {
	ForAllPins( i ) {
	    if((pins[i].netNo == netNo) && (layers[pins[i].layer].type == TYPE_POLY)) {
		pins[i].netNo = nNet;
	    }
	}
    }
    /*  Unmark the pins already connected by a wire */
    wn = (-1);
    while((wn=net_to_wire(netNo, wn+1)) != NOTFOUND) {
	for(i=0 ; i<wireSize(wn) ; i++) {
	    if((p=PinIsIn(wSegmRect(wn, i), nNet)) != NULL) {
		p->netNo = netNo;
	    }
	}
    }
}

static void
  JoinLinkedPins(netNo)
int netNo;
{
    int i;

    ForAllPins( i )
      if(pins[i].netNo < 0)
	pins[i].netNo = (-1 - pins[i].netNo);
}

static int
  RoutePartitionedNet(netNo)
int netNo;         /* Index of the linked net, already routed.  */
{
    int        i, wn,          /*  Wire index  */
               dim;
    int        newWire;
    RoadNet    myNet;
    RoadWeight wght[LAYER_MAX];

    initMessage("*Net partitioning*: first sub-net \n", 0);
    
    set_cost(netNo);
    ForAllLayers( i )
      wght[i] = layers[i].cost;

    DivideLinkedPins(netNo);
    newWire = RouteManager(netNo);
    if( newWire ) {
	if((wn=net_to_wire(netNo, 0)) == NOTFOUND)      return;  /* Strange */
	if(wSegmRect(wn, 0) != NULL) fullSpaceWire(wn);
    }

    JoinLinkedPins(netNo);

    initMessage("*Net partitioning*: Second sub-net \n", 0);
    
    myNet = rd_net[netNo];
    rd_net[netNo].resistive_weight = 0;
    set_cost(netNo);
    newWire += RouteManager(netNo);
    rd_net[netNo] = myNet;
    
    ForAllLayers( i )
      layers[i].cost = wght[i];
    return newWire;
}

/*
 *  When a new net has been selected for routing, this function
 *  cares for its whole running: it executes the necessary checks,
 *  sets the cost function and calls the route-manager.
 */
RouteSelectedNet(netNo)
int netNo;
{
    int   oldwn, wn = (-1);
    int   newWire;

    if(netNo == NOTFOUND)   return;

    /*  Has it at least two non-equivalent pins ?    */
    if(isNetRouted( netNo )) {
	deleteNetFromNetList(netNo);
	return;
    }
    
    initMessage("Routing %s \n", rd_net[netNo].name);
    /*  If the net has to be partitioned, route it properly   */
    if((rd_net[netNo].link == YES) && (rd_net[netNo].resistive_weight > 0)) {
	newWire = RoutePartitionedNet(netNo);
    } else {
	set_cost(netNo);
	newWire = RouteManager(netNo);
    }
    if( newWire ) {
	if((wn=net_to_wire(netNo, 0)) == NOTFOUND)      return;  /* Strange */
	if(wSegmRect(wn, 0) != NULL)    fullSpaceWire(wn);
    }
    
    RoadMirror();        /* Check if mirroring is necessary */
    CleanWires();
    UpdateSheetRects(netNo);
    RestorePinSheetList(netNo);
}

int
  RouteManager(netNo)
int netNo;
/*  Carry out the whole routing procedure for a  single net netNo. */
{
    int          oldwn;
    int          wn = (-1);
    int          i, j, l, k;      /* Loop counters   */
    int          wList_created;  /* No. of wList created */
    DTERM       *wList   = NULL;
    int          wListNum = 0;
    PATH         btPath;         /* Back-trace path  */
    Rect         propagate();
    Pins        *pinList;
    int          pinNum;

    pinList = ALLOC(Pins, PIN_NUM+1);
    pinNum  = MakePinArray(pinList, netNo);
    SaveEquivalences();
    
    /*  Allocate the wList structure array.  */
    wList = ALLOC(DTERM, pinNum);
    wListNum = 0;

    wn = (-1);
    while((wn=net_to_wire(netNo, wn+1)) != NOTFOUND) {
	wList[wListNum].rect = BuildPathFromWire(wn);
	if(IsPathComplete(wList[wListNum].rect, netNo)) {
	    oldwn = wn;
	    while((wn=net_to_wire(netNo, oldwn+1)) != NOTFOUND)
	      fuseWires(oldwn, wn);
	    deleteNetFromNetList(netNo);
	    FREE( pinList );
	    for(i=0 ; i<=wListNum ; i++)
	      FREE(wList[i].rect);
	    FREE( wList );
	    return 0;
	}
	wList[wListNum].pin = NULL;
	/*  Delete the pins already connected by the pre-defined paths. */
	for(k=0 ; k<pinNum ; k++)
	  for(l=0 ; wList[wListNum].rect[l]!=NULL ; l++)
	    if(pinList[k]->rect == wList[wListNum].rect[l]) {
		if(wList[wListNum].pin == NULL)
		  wList[wListNum].pin = pinList[k];
		join_equivalence(wList[wListNum].pin, pinList[k]);
		pinList[k--] = pinList[--pinNum];
		break;
	    }

	wListNum++;
	delete_wire(wn);
	AddNetListItem(netNo);
    }

    for(i=0 ; i<wListNum ; i++)
      if(wList[i].rect[0] == NULL)
	if(i < (--wListNum))
	  SWAP(DTERM, wList[i], wList[wListNum]);
    /*  ------------------------------------------------------------
     *  Transform the pins into one-element paths.
     */
    wList_created = wListNum  + pinNum + 1;
    wList = REALLOC(DTERM, wList, wList_created);
    for(i=0 ; i<pinNum ; i++) {
	wList[wListNum].rect = ALLOC(Rect, 2);
	wList[wListNum].rect[0] = pinList[i]->rect;
	wList[wListNum].rect[1] = NULL;
	wList[wListNum++].pin   = pinList[i];
    }
    wList[wListNum].rect    = ALLOC(Rect, 1);
    wList[wListNum].rect[0] = NULL;

    /*  ------------------------------------------------------------
     *  Link every pair of equivalent paths.
     */
    for(i=0 ; i<wListNum ; i++)
      for(j=i+1 ; j<wListNum ; j++)
	if(are_equivalent(wList[i].pin, wList[j].pin)) {
	    wList[i].rect = FullCatPaths(wList[i].rect, wList[j].rect);
	    wListNum--;
	    SWAP(DTERM, wList[j], wList[wListNum]);
	    j--;
	    CleanPath(wList[i].rect);
	}
    wList[wListNum].rect[0] = NULL;
    
    /*  ============================================================
     *  MAIN ROUTING LOOP.
     */
    btPath = ALLOC(Rect, TOTAL_RECT_NUM);
    while(wListNum > 1) {
	if(define(netNo, wList) == FAILURE)  break;
	wList[wListNum].rect[0] = NULL;
	SetKwave();
	if((btPath[0]=propagate()) == NULL) {
	    if(Kwave < 0) {
		warnMessage("%s is unroutable \n", rd_net[netNo].name);
		break;
	    }
	    Kwave = (-1);
	    continue;
	}
	
	full_backtrace( btPath );
    
	/*  ------------------------------------------------------------
	 *  Look for a link between paths.
	 */
	for(l=0,k=(-1) ; l<wListNum ; l++)
	  if((IntersectingPaths(btPath, wList[l].rect) == YES) ||
	     (OverlappingPaths(btPath, wList[l].rect) == YES)) {
	      if(k < 0) {
		  wList[l].rect = FullCatPaths(wList[l].rect, btPath);
		  k = l;
	      }
	      if(k < l) {
		  wList[k].rect = FullCatPaths(wList[k].rect, wList[l].rect);
		  wListNum--;
		  SWAP(DTERM, wList[l], wList[wListNum]);
		  wList[wListNum].rect[0] = NULL;
		  l--;
	      }
	  }
    }
    /*  ===================== END OF THE MAIN LOOP ================= */
    /*  ------------------------------------------------------------
     *  Delete the repeated rectangles of the path.
     */
    CleanPath(wList[0].rect);
    
    RestoreEquivalences();

    /*   Physically build the wire.   */
    MakePinArray(pinList, netNo);
    build_wire(netNo, wList[0].rect, pinList);
    
    FREE( pinList );
    FREE( btPath  );

    for(i=0 ; i<wList_created ; i++)
      FREE(wList[i].rect);
    FREE(wList);

    return 1;
}
