/*	setcost.c
 *
 *	Author: Enrico Malavasi
 *      Ver. 2.10
 *	Last revised: August, 1990
 *
 *	Purpose: Set the connection costs of all the rectangles.
 */
#include "port.h"
#include "st.h"
#include "roadGlobals.h"

RoadCost    *rdQuota  = NULL;
#define  QuotaOf(r)            rdQuota[(r)-rects]
#define  QuotaNo(n)            rdQuota[n]

#define PIN_ADDED_COST   1
#define  RESISTANCE_TO_COST(x)      ((RoadCost) ((x)*10))
#define  BULKDIST_TO_COST(x)        ((RoadCost) (1.0/(x)))

static RoadCoord lw[LAYER_MAX];

/*  ============================================================
 *  Raise of a given amount the cost on every side of a given
 *  rectangle.
 */
static void setMaxCost(r, cst)
Rect     r;
RoadCost cst;
{
  int dir;

  six_sides(dir)
    if(free_neighbor(r, dir) != NULL)
      set_connection_cost(r, dir, cst);
}

/*  ============================================================
 *  This function returns YES if the given grid element is movable
 *  in the given direction
 */
static MovableItem(r, side)
Rect r;
int  side;
{
  int i;
  Pins p;

  if(neighbor(r, side) == NULL)       return NO;
  if(free_neighbor(r, side) != NULL)  return YES;
  if((rNet(r) != NOTFOUND) && (net_to_wire(rNet(r), 0) != NOTFOUND)) {
      if(r->flag & RECTFLAG_TERM) {
	  if((p=PinIsIn(r, rNet(r))) != NULL) {
	      if(abs(p->side) != abs(side)) {
		  return NO;
	      }
	  }
      }
  }
  return YES;
}

#define  CONGESTION_AMPLIF    10000.0
#define  CONGESTION_HIGH       8000.0
static void
  setRectCongest(r, theDir, wdt)
Rect      r;
int       theDir;
RoadCoord wdt;
{
    Rect      s1, s2, p;
    RoadCoord room, R;

    s1 = s2 = r;
    theDir = abs(theDir);
    while((p=neighbor(s1, theDir)) != NULL) {
	if(MovableItem(s1, theDir) == NO) break;
	s1 = p;
    }
    CHGSGN(theDir);
    while((p=neighbor(s2, theDir)) != NULL) {
	s2 = p;
	if(MovableItem(s2, theDir) == NO) break;
    }
    room = rect_dist(s1, s2) - lw[rLayer(r)];
    if(room < wdt) room = wdt-1;
    R = (CONGESTION_AMPLIF*wdt)/((float) room);
    
    for(p=s1 ; ; p=neighbor(p,theDir)) {
	QuotaOf(p) += R;
	if(p == s2)  break;
    }
}

static void
congestEvaluate()
{
    int       wn, pos, i;
    int       dir;
    RoadCoord wdt;
    RoadCost  HIGH_COST;
    Rect      r;

    for(i=0 ; i<LAYER_NUM ; i++) {
	lw[i] = (layers[i].wire_width + layers[i].via_width)/2.0 +
	  layers[i].wire_dist;
    }

    rdQuota = REALLOC(RoadCost, rdQuota, TOTAL_RECT_NUM);
    ForAllRects(i)
      QuotaNo(i) = 0;
    
    ForAllWires(wn)
      for(pos=0 ; pos<wireSize(wn) ; pos++) {
	  if((dir=wSegmSide(wn, pos)) == UP)  dir = NORTH;
	  r    = wSegmRect(wn, pos);
	  wdt  = wire_segment_width(wn, pos, r->layer);
	  setRectCongest(r, dir, wdt);
	  if(wSegmSide(wn, pos) == UP) {
	      setRectCongest(r, EAST, wdt);
	      r    = neighbor(r, UP);
	      wdt  = wire_segment_width(wn, pos, r->layer);
	      setRectCongest(r, NORTH, wdt);
	      setRectCongest(r, EAST, wdt);
	  }
      }

    HIGH_COST = 10000;  /*  !!!!!!!!!!!!!!!!!!  */
    ForAllRects(i) {
	if(QuotaNo(i) >= CONGESTION_AMPLIF)
	  setMaxCost(rects+i, MAX_COST);
	else if(QuotaNo(i) >= CONGESTION_HIGH)
	  setMaxCost(rects+i, HIGH_COST);
    }
}

#define   RES_LENGTH(R,ll)  ((R)*layers[ll].wire_width/layers[ll].square_res)
#define   CAP_LENGTH(C,ll)  ((C)/layers[ll].wire_width/layers[ll].unity_cap)
/*  ============================================================
 *  Set the cost for layers and vias, corresponding to the given
 *  net number. For a given wire, let
 */
LayersCost(netNo)
int  netNo;
{
  int   i;         /*   Loop counter............................. */
  int   minResLayer, minCapLayer; /* Reference layers */
  float  Leq,      /*   Reference length   */
         L,        /*   Layer equivalent length      */
         Rcost,    /*   Resistive cost  */
         Ccost,    /*   Capacitive cost  */
         RCcost;   /*   Effective layer cost  */

/*  ------------------------------------------------------------
 *  1. Determine the less expensive layer for Res, Cap and RC
 */
  for(i=1,minResLayer=minCapLayer=0 ; i<LAYER_NUM ; i++) {
    if(layers[i].square_res < layers[minResLayer].square_res) minResLayer = i;
    if(layers[i].unity_cap < layers[minCapLayer].unity_cap)   minCapLayer = i;
  }

/*  ------------------------------------------------------------
 *  2. Set layer costs.
 */
  for(i=0 ; i<LAYER_NUM ; i++) {
    if(i == minResLayer)     Rcost = 0;
    else {
      Leq = RES_LENGTH(1.0, minResLayer);
      L   = RES_LENGTH(1.0, i);
      Rcost = ((L < Leq) ? rd_net[netNo].resistive_weight*(Leq-L)/L : 0);
    }
    if(i == minCapLayer)     Ccost = 0;
    else {
      Leq = CAP_LENGTH(1.0, minCapLayer);
      L   = CAP_LENGTH(1.0, i);
      Ccost = ((L < Leq) ? rd_net[netNo].capacitive_weight*(Leq-L)/L : 0);
    }
    RCcost = 1+(Rcost + Ccost);
    layers[i].cost = RCcost;
  }

/*  ------------------------------------------------------------
 *  3. Set the via costs
 */
#define  RELIABILITY   500
  for(i=0 ; i<LAYER_NUM ; i++) {
    if(layers[i].via_resistance == 0) continue;
    if(layers[i].via_capacitance == 0) continue;

    Leq = RES_LENGTH(layers[i].via_resistance, minResLayer);
    L   = 1.0;
    Rcost = ((L < Leq) ? rd_net[netNo].resistive_weight*(Leq-L)/L : 0);

    Leq = CAP_LENGTH(layers[i].via_capacitance, minCapLayer);
    Ccost = ((L < Leq) ? rd_net[netNo].capacitive_weight*(Leq-L)/L : 0);

    RCcost = 1+(Rcost + Ccost);
    RCcost = max(RCcost, RELIABILITY);
    layers[i].via_cost = RCcost;
  }
}

/*  ============================================================
 *  This function sets the costs of all rectangles. This is
 *  due before routing any wire.
 */
set_cost(NetNo)
int  NetNo;
{
  int i, j;    /* Loop counters .................. */
  Rect r, s;

  LayersCost(NetNo);

/*  ------------------------------------------------------------
 *  Set the basic connection cost for every rectangle.
 */
  ForAllRects(i) {
    r = rects+i;
    if(neighbor(r, UP) != NULL)
      rects[i].cost[1] = layers[rLayer(rects+i)].via_cost;
    if((s=neighbor(r, NORTH)) != NULL) {
      if(rwidth(rects+i) < 2)
	rects[i].cost[0] = rects[i].cost[1] = MAX_COST;
     else
	r->cost[0] = DistToCost(rUpperY(s) - rUpperY(r), rLayer(rects+i));
    }
    if((s=neighbor(r, EAST)) != NULL) {
      if(rheight(rects+i) < 2)
	rects[i].cost[2] = rects[i].cost[1] = MAX_COST;
     else
	rects[i].cost[2] = DistToCost(rRightX(s) - rRightX(r), rLayer(rects+i));
    }
  }
  if(preferWeight > 1)
    for(j=0 ; j<LAYER_NUM ; j++) {
      if(layers[j].preferDir == NORTH) {
	ForAllRects(i)
	  if(rLayer(rects+i) == j)
	    rects[i].cost[2]  *= preferWeight;
      } else if(layers[j].preferDir == EAST) {
	ForAllRects(i)
	  if(rLayer(rects+i) == j)
	    rects[i].cost[0] *= preferWeight;
      }
    }

/*  ------------------------------------------------------------
 *  Add the wire-dependent costs.
 */
  ForAllWires(i)
    costwr(i, NetNo);

/*  ------------------------------------------------------------
 *  Add the congestion costs.
 */
  CongestionEstimate(NetNo);
  congestEvaluate();
  if(GetSymmetryFlag())   SetSymmetryCost(NetNo);

/*  ------------------------------------------------------------
 *  Add the pin-(lower side) costs.
 */
  ForAllPins(i) {
    pins[i].exitCost = 0;
    if(pins[i].netNo == NetNo)
      if(((pins[i].side > 0) && (pins[i].cellNo > 0)) ||
	 ((pins[i].side < 0) && (pins[i].cellNo == 0)))
	pins[i].exitCost = rect_size(pins[i].rect, pins[i].side) *
	  layers[pins[i].rect->layer].cost;
  }
}

SetSymmetryCost(NetNo)
int NetNo;
{
  int i;

  ForAllRects(i)
    if(rRightX(rects+i) == SymmetryAxis)
      rects[i].cost[2] = MAX_COST+1;
}
