/*      shield.c
 *
 *      Author: Enrico Malavasi
 *      Ver. 2.10
 *      Last revised: September, 1990
 *
 *      Purpose: Build a shield between two critical wires.
 */
#include "port.h"
#include "utility.h"
#include "roadGlobals.h"
#include "roadTechnol.h"
#include "transf.h"
#include "models.h"

#define SHIELD_SIZE          300
#define MIN_SHIELD_DIST       25     /* in microns  */

static RoadStatus
  pinToBeDeleted(pinNo)
int pinNo;
{
    if(pins[pinNo].netNo == NOTFOUND)  return(YES);
    if(wire_segment_position(WIRE_NUM-1, pins[pinNo].rect, EAST, NOTFOUND) != NOTFOUND)
      return(delete_appendixes(WIRE_NUM-1, pins[pinNo].rect, EAST));
    if(wire_segment_position(WIRE_NUM-1, pins[pinNo].rect, NORTH, NOTFOUND) != NOTFOUND)
      return(delete_appendixes(WIRE_NUM-1, pins[pinNo].rect, NORTH));
    return(YES);
}

/*  =====================================================================
 *  This function is the front-end for the all-shield building procedure.
 */
make_shield()
{
    int        i;
    int        netNo;
    RoadWeight coupling;
    pFarad     maxcap;
    int        shldNet;
    
    ForAllNets( i ) {
	if((netNo=getCriticCapPair(i, &coupling, &maxcap, &shldNet)) != NOTFOUND) {
	    if(shldNet == NOTFOUND)                   continue;
	    if(netNo < i)   continue;
	    if(shield(i, netNo, shldNet) == SUCCESS) {
		fullSpaceWire(net_to_wire(shldNet, 0));
		show_all(-1);
	    }
	    modifyCriticCap(i, netNo, coupling, maxcap, shldNet,
			    CRIT_SHIELD_MASK);
	}
    }
}

/*  ============================================================
 *  The following function physically builds the required shield
 *  between a pair of specified nets.
 */
RoadStatus
  shield(aNet, bbNet, shldNet)
int      aNet, bbNet,     /*  Electric nets to be shielded   */
         shldNet;         /*  Net to be used for the shield  */
{
    int  wn1;
    Rect r, s[2], rt;
    struct {    int rectNo;
		int side;
	    } dp[SHIELD_SIZE];            /* Dummy pins locations ............ */
    RoadCoord  x1, y1;
    int     k, wb;
    int     i,j;                          /* Loop counter.                     */
    int     side, mside, sd;              /* Side element of the wire segment. */
    int     OldPinNum;                    /* Old PIN_NUM value ............... */
    static  RoadCoord   dist_min = 0;
    
    if(dist_min == 0)  dist_min = MicronToRoad(MIN_SHIELD_DIST);
    if(aNet    == NOTFOUND)  return(FAILURE);
    if(bbNet   == NOTFOUND)  return(FAILURE);
    if(shldNet == NOTFOUND)  return(FAILURE);
    /*  ------------------------------------------------------------
     *  Transform the electric node <b> into its corresp. wire index.
     */
    wb = net_to_wire(bbNet, 0);
    
    /*  ------------------------------------------------------------
     *  Find the places where the shield is necessary.
     */
    wn1 = net_to_wire(aNet, 0);
    for(i=k=0 ; (r=wSegmRect(wn1, i))!=NULL ; i++) {
	if((side=wSegmSide(wn1, i)) == UP)  continue; /* Looking only for horiz. elem. */
	
	mside = (- side);
	s[0] = follow_FreeNeighbor(neighbor(r, side), side);
	s[1] = follow_FreeNeighbor(r, mside);
	
	for(j=0, sd=side ; j<2 ; j++, sd=mside)
	  if((rect_dist(r, s[j])>dist_min) || (neighbor(s[j], sd)==NULL))
	    s[j] = NULL;
	
	/*  ------------------------------------------------------------
	 *  Place the shield in the middle.
	 */
	if(s[1] != NULL) s[1] = neighbor(s[1], mside);
	for(j=0 ; j<2 ; j++)
	  if(wSegmHere(wb, s[j], side)) {
	      x1 = (rRightX(r) + rRightX(s[j]))*0.5;
	      y1 = (rUpperY(r) + rUpperY(s[j]))*0.5;
	      if((rt=find_rect(x1, y1, r->layer)) != NULL)
		if((neighbor(rt, EAST)  != NULL) &&
		   (neighbor(rt, NORTH) != NULL) &&
		   (neighbor(rt, WEST)  != NULL) &&
		   (neighbor(rt, SOUTH) != NULL)) {
		    dp[k].rectNo = rt - rects;
		    GridSplit(rects+dp[k].rectNo,
			      ((side == NORTH) ? y1-rLowerY(rt) : x1-rLeftX(rt)),
			      AbsDualDir(side), NOTFOUND);
		    dp[k++].side = side;
		}
	  }
	dp[k].rectNo = NOTFOUND;
    }
    
    /*  ------------------------------------------------------------
     *  Control over the SHIELD_SIZE parameter.
     */
    if(k == 0) return(FAILURE);        /* No shields are needed .......... */
    if(k >= SHIELD_SIZE)
      errMessage("shield: SHIELD_SIZE too small", 0);
    
    /*  ------------------------------------------------------------
     *  Now build up the shield. 
     *  1. Create the dummy pins along the path of the shield.
     */
    OldPinNum = PIN_NUM;
    pins = REALLOC(RoadTerm, pins, (PIN_NUM+k));
    for(i=0 ; i<k ; i++) {
	newRoadPin(shldNet, rects[dp[i].rectNo].layer, dp[i].side,
		   (RoadCoord) 0, rects+dp[i].rectNo, NOTFOUND,NOTFOUND,"");
    }
    
    /*  ------------------------------------------------------------
     *  Control over the effective room for the new shield wire.
     */
  for(i=0 ; i<k ; i++) {
      
      if(dp[i].side != UP) {
	  mside = AbsDualDir(dp[i].side);
	  if(free_neighbor(rects+dp[i].rectNo, -mside) == NULL) {
	      if(rect_size(rects+dp[i].rectNo, mside) < 2) {
		  pins[OldPinNum+i].netNo = NOTFOUND;
		  continue;
	      }
	      else {
		  GridSplit(rects+dp[i].rectNo,
			    rect_size(rects+dp[i].rectNo, mside)/2, dp[i].side,
			    NOTFOUND);
		  rt = rects+dp[i].rectNo;
		  rt = neighbor(rt, mside);
		  dp[i].rectNo = rt-rects;
	      }
	  }
	  if(free_neighbor(rects+dp[i].rectNo, mside) == NULL) {
	      if(rect_size(rects+dp[i].rectNo, mside) < 2) {
		  pins[OldPinNum+i].netNo = NOTFOUND;
		  continue;
	      }
	      else
		GridSplit(rects+dp[i].rectNo,
			  rect_size(rects+dp[i].rectNo, mside)/2, dp[i].side,
			  NOTFOUND);
	  }
      }
      
      if(free_neighbor(rects+dp[i].rectNo, dp[i].side) == NULL) {
	  
	  if(free_neighbor(rects+dp[i].rectNo, (- dp[i].side)) != NULL) {
	      /* Best solution: ... */
	      rt = rects+dp[i].rectNo;
	      rt = neighbor(rt, (- dp[i].side)); /* There's room  */
	      dp[i].rectNo = rt - rects;
	      pins[OldPinNum+i].rect = rt;          /* at the left side. */
	      continue;
	  }

	  if(((dp[i].side == EAST)  && (rwidth(rects+dp[i].rectNo)  < 2)) ||
	     ((dp[i].side == NORTH) && (rheight(rects+dp[i].rectNo) < 2))) {
	      /* There is not room enough !! */
	      pins[OldPinNum+i].netNo = NOTFOUND;
	      continue;
	  }  else
	    GridSplit(rects+dp[i].rectNo,
		      rect_size(rects+dp[i].rectNo, dp[i].side)/2,
		      DualDir(dp[i].side), NOTFOUND);
      }
  }
    
    for(i=0 ; i<k ; i++)
      if(pins[OldPinNum+i].netNo == NOTFOUND)
	dp[i].rectNo = NOTFOUND;
    for(i=0 ; i<k ; i++)
      if(dp[i].rectNo == NOTFOUND)
	dp[i--].rectNo = dp[--k].rectNo;
    
    if(k == 0) {
	PIN_NUM = OldPinNum;
	return(FAILURE);
    }
    for(i=OldPinNum ; i<PIN_NUM ; i++)
      pins[i].side = (pins[i].side==EAST) ? SOUTH : WEST;
    
    set_cost(shldNet);
    RouteManager(shldNet);
    if INTERACTIVE_MODE      show_all(-1);
    
    /*  ------------------------------------------------------------
     *  Now delete the unnecessary new pins
     */
    for(i=OldPinNum ; i<PIN_NUM ; i++) {
	if(pinToBeDeleted(i) == YES)
	  delete_pin(i--);
	else {
	    pins[i].side = 0;
	    pins[i].type = TERM_SHIELD_TYPE;
	}
    }
    
    pins = REALLOC(RoadTerm, pins, PIN_NUM);
    return SUCCESS;      /* The shield has been built ................. */
}
