/*	show.c
 *
 *	Author: Enrico Malavasi, Univ. of Bologna
 *      Ver. 3.1
 * 	Last revised: April, 1991
 *
 * 	Purpose: Utility functions for the graphic representation of 
 *		 the circuit items.
 */
#include "graph.h"

show_rect(r)  /* Highlight a rectangle of the frame */
Rect r;
{
    set_color(color_set.layer_color[r->layer]);
    box(rLeftX(r), rLowerY(r), rRightX(r), rUpperY(r));
    RoadSync(1);
}

static void show_via(x, y, l)
RoadCoord x,y;
{
    RoadCoord wdt;
    int       layerNo;

    layerNo = (l >= 0) ? l : (-1-l);
    wdt = layers[layerNo].via_width / 2.0;
    filled_box(x-wdt,y-wdt, x+wdt, y+wdt,
	       (l<0) ? color_set.HighLight : color_set.layer_color[layerNo]);
    wdt = layers[layerNo+1].via_width / 2.0;
    filled_box(x-wdt,y-wdt, x+wdt, y+wdt,
	       (l<0) ? color_set.HighLight : color_set.layer_color[layerNo+1]);
    if(l >= 0) {
	wdt = layers[l+1].wire_width / 2.0;
	filled_box(x-wdt,y-wdt, x+wdt, y+wdt, color_set.Obstacles);
	set_layer_color(l);
	set_line_width(layers[l].wire_width);
    }
}
static void show_wire_segment(wn, pos, hl)
    int  wn, pos;
    int  hl;           /*   if hl==0 -> real colors, else HighLight  */
{
    Rect r;
    
    if((r=wSegmRect(wn, pos)) == NULL)  return;
    if(hl) set_color(color_set.HighLight);
    else   set_layer_color(r->layer);
    set_line_width(wire_segment_width(wn, pos, r->layer));
    switch(wSegmSide(wn, pos))   {
      case NORTH : segment(r->x1,r->y2,r->x2,r->y2);
	break;
      case EAST  : segment(r->x2,r->y1,r->x2,r->y2);
	break;
      case UP    : show_via(r->x2,r->y2,((hl)?(-1-r->layer):r->layer));
    }
}

/*  ============================================================
 *  The following functions show a given item, as described by
 *  their names.
 */
show_cell(cellNo)  /*  Show a given obstacle */
int cellNo;
{
    set_layer_color(-1);
    set_line_width(0);
    box(cells[cellNo].x1, cells[cellNo].y1, cells[cellNo].x2, cells[cellNo].y2);
}
show_wire(wn, hl)  /*  Show a given wire  */
int  wn;
int  hl;
{
    int pos;
    
    for(pos=0 ; pos<wireSize(wn) ; pos++)
      show_wire_segment(wn, pos, hl);
    RoadSync(1);
}

#define  LETTERSIZE  6
/*  ============================================================
 *  This function draws a given text in a given position.
 */
static void RoadLabel(x, y, side, s)
RoadCoord  x,y;    /*  Position  */
int    side;   /*  Side of the cell  */
char  *s;      /*  Text  */
{
  int l;

  if((l=strlen(s)) == 0) return;

  x = Xcorrect(x);
  y = Ycorrect(y) + 2;

  switch(side) {
  case EAST:
    x -= (LETTERSIZE*l);
    break;
  case NORTH:
    x -= (l / 2.0)*LETTERSIZE;
    y += 4;
    break;
  case SOUTH:
    x -= (l / 2.0)*LETTERSIZE;
    y -= 9;
  default:
    break;
  }
  set_layer_color(-1);
  set_line_width(0);
  testo(x, y, s, MAINWindow);
}

static void show_pin(pn) /* Show pin number and its internal connections */
int  pn;
{
    Pins p;
    int   i, wn;
    RoadCoord x,y;
    char  *s;
    
    p = pins+pn;
    if(p->cellNo == NOTFOUND) return;      /* It is a dummy pin .......... */
    
    if((wn=net_to_wire(p->netNo, 0)) < 0)   return; /* Wire not routed */
    
    /*  ------------------------------------------------------------
     *  Internal connections.
     */
    set_layer_color(p->layer);
    set_line_width(layers[p->layer].wire_width);
    for(i=0 ; i<PIN_NUM ; i++)
      if((pins[i].netNo    == p->netNo) &&
	 (pins[i].equivNo  == p->equivNo) &&
	 (pins[i].cellNo   != 0) &&
	 (pins[i].cellNo   == p->cellNo) &&
	 (pins[i].side     == (- p->side)) &&
	 (pins[i].position == p->position)) {
	  dashed_segment(pin_x(p), pin_y(p), pin_x(pins+i), pin_y(pins+i));
	  break;
      }
    
    for(i=0 ; i<wireSize(wn) ; i++) {
	if(p->rect == wSegmRect(wn, i)) {
	    x = pin_x(p);
	    y = pin_y(p);
	    s = rd_net[p->netNo].name;
	    RoadLabel(x, y, p->side, s);
	    break;
	}
    }
}

/*  ============================================================
 *  This function shows the whole routed circuit.
 *  The graphic initialization should have already been done.
 */
show_all(k)
int k;		/*	>  0 --> Don't clear the screen
		 *	==-1 --> Shows everything after cls */
{
    int j,l, wn;

    if(RD_CHECK_FLAG( BATCH_MODE ))    return;

    if(k != 0)
      filled_box(Xuncorrect(0), Yuncorrect(0),
		 Xuncorrect(ZX1), Yuncorrect(ZY1), color_set.BackGround);
    
    for (j=0 ; j<=SUB_CELL_NUM ; j++)
      show_cell(j);
    for (j=0 ; j<INST_NUM ; j++)
      RoadLabel((RoadCoord) insts[j].x, (RoadCoord) insts[j].y, SOUTH, insts[j].name);
    
    ForAllWires( wn ) {
	show_wire(wn, 0);
	
	/*  Show the net name near the pin. */
	for(j=0 ; j<PIN_NUM ; j++)
	  if(pins[j].netNo == wireNetNo(wn))
	    for(l=0 ; wSegmRect(wn, l)!=NULL ; l++)
	      if(pins[j].rect == wSegmRect(wn, l))
		show_pin(j);
    }
    
    ShowParamWindow();
    RoadSync(1);
}
