/*     wireshift.c
 *
 *     Author: Enrico Malavasi, Univ. of Bologna
 *     Ver. 2.10
 *     Last revised: September, 1990
 *
 *     Purpose: Shift wire lines to adjacent positions.
 *              Used for weak modifications.
 */
#include "roadGlobals.h"
#include "space.h"

/*  ============================================================
 *  This function tries to shift a wire element of a given amount.
 *  It returns the effective shift size.
 */
RoadCoord
wire_shift_N(wn, pos, dir, room)
int  wn;          /* Index of the wire to be shifted .......... */
int  pos;         /* Position of the segment on the given wire. */
int  dir;         /* Direction of the shift ................... */
RoadCoord room;   /* Shift size ............................... */
{
    Rect r0,     /*  Old proposal */
         r,      /*  Proposal  */
         s;      /*  Span  */
    int  cnt=0, nr,
         side, mdir, wi;
    RoadCoord ddd, ss;
    RoadBox   BB, bUp;

    if(room <= 0)   return(0);
    mdir = (- dir);
    r = r0 = wSegmRect(wn, pos);
    side   = wSegmSide(wn, pos);
    ss = spacingBetween(rLayer(r0), side, rLayer(r0), NORTH);
    WireSegmentToBox(r0, side, ss, &BB, &bUp, dir, dir);

    s = (dir > 0) ? neighbor(r0, dir) : r0;
    if(s == NULL)    return((RoadCoord) 0);

    /*  ============================================================
     *  1.) Shift as much as feasible.
     */
    loop {
	if((ss=rect_size(s, dir)) > room) break;
	r = neighbor(r0, dir);
	unless(XXXavailable(wn, r, side)) break;
	r0 = r;
	ShiftBox(&BB, dir, ss);
	if(side == UP) ShiftBox(&bUp, dir, ss);
	cnt++;
	room -= ss;
	s = neighbor(s, dir);
    }

/*  ============================================================
 *  Now if there was enough room available, (r0==r)&&(room==0),
 *  otherwise (room>0)&&(r0!=r)
 */
  if(room > 0) {
    cnt++;
    r = neighbor(r0, dir);
    if(XXXavailable(wn, r, side)) {
      ShiftBox(&BB, dir, ss);
      if(side == UP) ShiftBox(&bUp, dir, ss);
      wi = newAvailableBox(wn, &BB, &bUp, (side==UP) ? NODIRECTION : AbsDualDir(side), NODIRECTION, r, side);
      if(wi == NOTFOUND)    r0 = r;
      else {
	ShiftBox(&BB, mdir, ss);
	if(side == UP) ShiftBox(&bUp, mdir, ss);
      }
    }
    if(r0 != r) {
      if(dir > 0) {
	r0 = neighbor(r0, dir);
        ddd = room;
      } else
	ddd = rect_size(r0, dir)-room;
      if((ddd <= 0) || (ddd >= rect_size(r0, dir))) {
	cnt--;
	if(dir > 0) r0 = neighbor(r0, mdir);
      } else {
        ShiftBox(&BB, dir, room);
        if(side == UP) ShiftBox(&bUp, dir, room);
	if(newAvailableBox(wn, &BB, &bUp, NODIRECTION, NODIRECTION, r0, side) == NOTFOUND) {
	  nr = r0-rects;
	  GridSplit(r0, ddd, AbsDualDir(dir), wn);
	  r0 = rects+nr;
	} else {
          ShiftBox(&BB, mdir, room);
          if(side == UP) ShiftBox(&bUp, mdir, room);
          cnt--;
          if(dir > 0) r0 = neighbor(r0, mdir);
	}
      }
    }
  }
  if(cnt == 0)     return(0);

/*  ============================================================
 *  If the position is feasible, but illegal, shift back.
 */
  loop {
    if(newAvailableBox(wn, &BB, &bUp, NODIRECTION, NODIRECTION, r0, side) != NOTFOUND) {
      if((--cnt) == 0) return(0);
      if(mdir < 0) s = r0;
      r0 = neighbor(r0, mdir);
      if(mdir > 0) s = r0;
      ss = rect_size(s, mdir);
      ShiftBox(&BB, mdir, ss);
      if(side == UP) ShiftBox(&bUp, mdir, ss);
    } else
      break;
  }

/*  ============================================================
 *  Carry out the required shift
 */
  side = wSegmSide(wn, pos);
  for(room=0, r=wSegmRect(wn, pos) ; cnt>0 ; cnt--) {
    s = (dir > 0) ? neighbor(r, dir) : r;
    room += rect_size(s, dir);
    nr = r-rects;
    wire_segment_shift(wn, pos, dir);
    r = rects+nr;
    if(cnt > 1) {
      r = neighbor(r, dir);
      pos = wire_segment_position(wn, r, side, -1);
/*      if(pos == NOTFOUND)
 *	fprintf("This is for the debugger"); */
    }
  }
  return(room);
}

/*  ============================================================
 *  This function performs the shift of a single element of the
 *  given wire. No controls over the correctness of the parameters
 *  are performed here. The wire free-status must be FREE.
 */
wire_segment_shift(wn, pos, dir)
int  wn;          /* Index of the wire to be shifted ........ */
int  pos;         /* Position of the segment on the given wire. */
int  dir;         /* Direction of the shift ................. */
{
  Rect s;         /* Adjacent position ...................... */
  Rect r;         /* Rect field of the segment to be shifted. */
  int  side,      /* Side of the segment to be shifted.       */
       mside,     /* Complementary side ..................... */
       j;         /* Loop counter .............. */

  r    = wSegmRect(wn, pos);
  side = wSegmSide(wn, pos);

/*  ------------------------------------------------------------
 *  Shift the given wire segment.
 */
  s = neighbor(r, dir);
  if(wire_segment_position(wn, s, side, NOTFOUND) == NOTFOUND)
    wSegmRect(wn, pos) = s;
  else
    sub_from_wire(wn, wSegmRect(wn, pos), side);
  if(wn < WIRE_NUM) set_free_neighbor(r, side, 0);
  if((dir == WEST) || (dir == SOUTH))   s = r;
  mside = AbsDualDir(dir);

/*  ------------------------------------------------------------
 *  Control its connections.
 */
  if(wire_segment_position(wn, s, mside, pos) == NOTFOUND)
    add_to_wire(wn, s, mside);
  j = (side == UP) ? UP : (- mside);

  if(wire_segment_position(wn, neighbor(s, j), mside,pos) == NOTFOUND)
    add_to_wire(wn, neighbor(s, j), mside);
/*  -----------------------------------------------------------------
 *  Now the lateral connections are set. Delete the unnecessary ones.
 */
  if((delete_appendixes(wn, s, mside)) &&
     (PinIsIn(s, wireNetNo(wn)) == NULL))
    sub_from_wire(wn, s, mside);
  if((delete_appendixes(wn, neighbor(s, j),mside)) &&
     (PinIsIn(neighbor(s, j), wireNetNo(wn)) == NULL))
    sub_from_wire(wn, neighbor(s, j), mside);
}

/*  ============================================================
 *  Propagate the wire migration along the segment, if it is
 *  possible.
 *  The wire free-status must be FREE
 */
static RoadCoord
wire_shift_follow(wn, wPos, shiftDir, follDir, room)
int   wn;             /*  Wire index ........... */
int   wPos;           /*  Wire segment position  */
int   shiftDir;       /*  Shift direction ...... */
int   follDir;        /*  Shift foll.direction   */
RoadCoord room;       /*  Shift size ........... */
{
  Rect    r;
  Pins    p;
  int     side, rn;
  RoadCoord retRoom, h;
  RoadCoord z;

  for(retRoom=0 ; ; ) {
    r    = wSegmRect(wn, wPos);
    if((side=wSegmSide(wn, wPos)) != UP)
      if(((p=PinIsIn(r, wireNetNo(wn)))!=NULL) &&
	 (abs(p->side)!=abs(side)))
	break;

    if(shiftDir < 0)
      z = (shiftDir == SOUTH) ? rUpperY(r) : rRightX(r);
    rn = r-rects;
    h = wire_shift_N(wn, wPos, shiftDir, room);
    if(retRoom == 0) {
      if(h == 0) return(0);
      else       retRoom = h;
    }
    r = rects+rn;
    room = h;

/*  ------------------------------------------------------------
 *  Select and move the next wire element
 */
    if(side == UP) {
      if(follDir > 0) r = neighbor(r, follDir);
      if(r == NULL) break;  /*  End of the foll. procedure  */
      if((wPos=wire_segment_position(wn, r, abs(shiftDir), -1))
	  != NOTFOUND) {
	rn = r-rects;
	wire_shift_follow(wn, wPos, shiftDir, follDir, room);
	r = rects+rn;
      }
      r = neighbor(r, UP);
      if((wPos=wire_segment_position(wn, r, abs(shiftDir), -1))
	  == NOTFOUND)
	break;
    } else {
      if(follDir < 0) r = neighbor(r, follDir);
      if(r == NULL) break;  /*  End of the foll. procedure  */
      if((wPos=wire_segment_position(wn, r, UP, -1)) != NOTFOUND)
	continue;

      if(((neighbor(r, DOWN) != NULL) &&
	  ((wPos=wire_segment_position(wn, neighbor(r, DOWN), UP, -1)))
	  != NOTFOUND))
	continue;
/*  ------------------------------------------------------------
 *  No vias here !
 */
      if(follDir > 0) r = neighbor(r, follDir);
      if(r == NULL) break;  /*  End of the foll. procedure  */
      if(((shiftDir == SOUTH) && (z != rUpperY(r))) ||
	 ((shiftDir == WEST)  && (z != rRightX(r))))
	r = neighbor(r, (-shiftDir));
      if((wPos=wire_segment_position(wn, r, side, -1)) == NOTFOUND)
	break;
    }
  }

  return(retRoom);
}

/*  ============================================================
 *  This function applies the wire-shift to a given wire element,
 *  then it propagates the same shift along the wire, in order
 *  to minimize the number of corners.
 *  The free-status of the wire must be FREE
 */
RoadCoord
wire_shift(wn, wPos, dir, room)
int   wn;          /* Index of the wire to be shifted ........ */
int   wPos;         /* Position of the segment on the given wire. */
int   dir;         /* Direction of the shift ................. */
RoadCoord room;
{
  int  dualDir,             /* Dual dir.  */
       TheSide;             /* The wire side to be shifted */
  int  nr;
  RoadCoord z, retRoom = 0;
  Rect r;

  r              = wSegmRect(wn, wPos);
  TheSide        = wSegmSide(wn, wPos);
  dualDir        = AbsDualDir(dir);

  z = (dualDir == EAST) ? rUpperY(r) : rRightX(r);
  nr = r-rects;
  if((retRoom=wire_shift_follow(wn, wPos, dir, dualDir,room)) == 0)
    return(0);
  r = rects+nr;

  if(dualDir == EAST) {
    while(rUpperY(r) < z)
      r = neighbor(r, (-dir));
  } else {
    while(rRightX(r) < z)
      r = neighbor(r, (-dir));
  }
  CHGSGN(dualDir);
  if(neighbor(r, dualDir) == NULL)   return(retRoom);

  if(TheSide == UP) {
    if((wPos=wire_segment_position(wn, neighbor(r, UP),abs(dir),-1))
       != NOTFOUND) {
      nr = r-rects;
      wire_shift_follow(wn, wPos, dir, dualDir,room);
      r = rects+nr;
    }
  } else {
    r = neighbor(r, dualDir);
    if((wPos=wire_segment_position(wn, r, UP, -1)) != NOTFOUND) {
      wire_shift_follow(wn, wPos, dir, dualDir,room);
      return(retRoom);
    }
    if((wPos=wire_segment_position(wn, neighbor(r, DOWN), UP, -1))
       != NOTFOUND) {
      wire_shift_follow(wn, wPos, dir, dualDir,room);
      return(retRoom);
    }
  }

  if((wPos=wire_segment_position(wn, r, abs(dir), -1)) != NOTFOUND)
    wire_shift_follow(wn, wPos, dir, dualDir,room);

  return(retRoom);
}

/*  ============================================================
 *  This function tries to shift a wire segment. If it is not
 *  possible, the grid is split and the shift is performed.
 *  The free-status of the wire must be FREE.
 */
RoadStatus
wire_SplitAndShift(wn, pos, dir)
int  wn;          /* Index of the wire to be shifted ........ */
int  pos;         /* Position of the segment on the given wire. */
int  dir;         /* Direction of the shift ................. */
{
  Rect  s;        /*  Grid node to be eventually splitten  */
  RoadCoord h;    /*  Size of s .........................  */

/*  ------------------------------------------------------------
 *  If possible, shift the wire and return ...
 */
  if(wire_shift(wn, pos, dir, 0) > 0)                  return(SUCCESS);
  s = (dir > 0) ? neighbor(wSegmRect(wn, pos), dir) : wSegmRect(wn, pos);
  if(s == NULL)                                        return(FAILURE);
  if((h=rect_size(s, dir)) < MIN_NODE_SIZE)            return(FAILURE);
  GridSplit(s, (RoadCoord) h/2, AbsDualDir(dir), wn);
  return((wire_shift(wn, pos, dir, 0) > 0) ? SUCCESS : FAILURE);
}
