/*	wireutil.c
 *
 *	Author: Enrico Malavasi
 *      Ver. 2.10
 *      Last updated: September 1990
 *
 *	Purpose: Utility functions about wires: wire segment extraction,
 *               pruning and search. Wire updating.
 */
#include "port.h"
#include "utility.h"
#include "roadGlobals.h"
#include "space.h"

static int
best_layer(l1, l2)
int l1, l2;	     /* Layers between which to choose . */
{
  if(layers[l1].cost == layers[l2].cost)  return(max(l1, l2));
  return((layers[l1].cost > layers[l2].cost) ? l2 : l1);
}

/*  ============================================================
 *  Return the position of a given segment in the wire.
 *  If the segment is not in the wire, (NOTFOUND) is returned.
 *  An <exclude> position can be supplied, which is not allowed
 *  to be the returned position.
 */
int
wire_segment_position(wn, r, side, exclude)
int  wn;         /* Wire index ......................... */
Rect r;          /* segment  rect ...................... */
int  side;       /* Segment side ....................... */
int exclude;     /* Excluded position .................. */
{
    int i;           /* loop counter */

    if(r == NULL)      return NOTFOUND;
    for(i=0 ; wSegmRect(wn, i)!=NULL ; i++)
      if((r == wSegmRect(wn, i)) && (side == wSegmSide(wn, i)))
	if(i != exclude)    return i;
    return NOTFOUND;
}

/*  ============================================================
 *  Return the first wire element on side <side>
 */
int
wireSegmOnSide(wn, side, begin)
int  wn,        /*  Wire index  */
     side,      /*  Side */
     begin;     /*  Starting element  */
{
  int  i;
  for(i=begin ; i<wireSize(wn) ; i++)
    if(wSegmSide(wn, i) == side)
      return i;
  return NOTFOUND;
}

/*  ============================================================
 *  Return the wire width for a given element
 */
RoadCoord
  wire_segment_width(wn, pos, layer)
int  wn,
     pos;
int  layer;
{
    switch(wireDesc(wn)) {
      case DEFWWIDTH :
	if(wSegmSide(wn,pos) != UP) {
	    if(wSegmLayer(wn,pos) == layer)
	      return wSegmWidth(wn, 0);
	} else {
	    if((wSegmLayer(wn,pos) == layer) ||
	       (wSegmLayer(wn,pos) == (layer-1)))
	      return max(layers[layer].via_width, wSegmWidth(wn, 0));
	}
	return ((RoadCoord) 0);
      case VARWWIDTH :
	if(wSegmSide(wn,pos) != UP) {
	    if(wSegmLayer(wn,pos) == layer)
	      return wSegmWidth(wn, pos);
	} else {
	    if((wSegmLayer(wn,pos) == layer) || (wSegmLayer(wn,pos) == (layer-1)))
	      return max(layers[layer].via_width, wSegmWidth(wn, pos));
	    return ((RoadCoord) 0);
	}
      case MINWWIDTH:
	default :
	  if(wSegmSide(wn,pos) != UP) {
	      if(wSegmLayer(wn,pos) == layer)
		return(layers[layer].wire_width);
	  } else {
	      if((wSegmLayer(wn,pos) == layer) || (wSegmLayer(wn,pos) == (layer-1)))
		return(layers[layer].via_width);
	  }
	return((RoadCoord) 0);
    }
}

#define Eureka(x)     (rNet(x) == wireNetNo(wn))
#define WIREPOS(x,d)  (Eureka(x) ?                 \
		       wire_segment_position(wn,x,d,NOTFOUND) : NOTFOUND)
int
  wire_obstacle(wn, r, side, theDir)
int  wn;     /*  wire index */
Rect r;      /*  rect  */
int  side;   /*  wire element side  */
int  theDir; /*  direction  */
/*  Look for a wire element of wire <wn> hampering the creation of a wire element
 *  on a given position. Return NOTFOUND if such element was not found, otherwise
 *  the element index is returned.  */
{
    Rect s;
    int  aDir;

    if(theDir < 0) {
	if((r=neighbor(r, theDir)) == NULL)   return(NOTFOUND);
	CHGSGN(theDir);
    }
    aDir = AbsDualDir(theDir);
    
    if(ViaUp(r))                           return WIREPOS(r, UP);
    if(ViaDown(r))                         return WIREPOS(neighbor(r,DOWN), UP);
    if(free_neighbor(r, theDir) == NULL)   return WIREPOS(r, theDir);
    if((s=neighbor(r, aDir)) != NULL)
      if(free_neighbor(s,theDir) == NULL)  return WIREPOS(s, theDir);
    if(side == UP) {
	if((r=neighbor(r, UP)) == NULL)      return NOTFOUND;
	if(free_neighbor(r, theDir) == NULL) return WIREPOS(r, theDir);
	if((s=neighbor(r, aDir)) != NULL)
	  if(free_neighbor(s, theDir) == NULL) return WIREPOS(s, theDir);
	return NOTFOUND;
    }
    CHGSGN(aDir);
    if((r=neighbor(r, aDir)) != NULL) {
	if(ViaUp(r))                         return WIREPOS(r, UP);
	if(ViaDown(r))                       return WIREPOS(neighbor(r, DOWN), UP);
	if(free_neighbor(r, theDir) == NULL) return WIREPOS(r, theDir);
    }
    return NOTFOUND;
}

/*  ============================================================
 *  The following function returns YES if the wire segment
 *  is connected to other wire segments on at least one side.
 */
RoadStatus
isConnected(wn, pos)
int  wn;
int  pos;
{
  Rect r, s;
  int  side, adir, madir;

  r    = wSegmRect(wn, pos);
  side = wSegmSide(wn, pos);
  adir = AbsDualDir(side);
  madir = (-adir);

  if(ViaUp(r))
    if(Eureka(r))                    return YES;
  if(ViaDown(r))
    if(Eureka(neighbor(r, DOWN)))    return YES;
  if(wSegmHere(wn, r, adir))         return YES;
  if((s=neighbor(r, adir)) != NULL)
    if(wSegmHere(wn, s, side))     return YES;
  if((s=neighbor(r, side)) != NULL) {
    if(wSegmHere(wn, s, adir))       return YES;
    if((s=neighbor(s, madir)) != NULL)
      if(wSegmHere(wn, s, adir))     return YES;
  }
  if((s=neighbor(r, madir)) != NULL) {
    if(ViaUp(s))
      if(Eureka(s))                  return YES;
    if(ViaDown(s))
      if(Eureka(neighbor(s, DOWN)))  return YES;
    if(wSegmHere(wn, s, adir))       return YES;
    if(wSegmHere(wn, s, side))     return YES;
    if((s=neighbor(s, side)) != NULL)
      if(wSegmHere(wn, s, adir))     return YES;
  }
  return NO;
}

static int
  DummyNet(netNo)
int netNo;              /*  Index of the net to be checked  */
{
    int      i, j;        /*  Loop counters        */
    int      lnkNet;
  
    lnkNet = (-1 -netNo);
    ForAllPins(i)
      if((pins[i].netNo == netNo) || (pins[i].netNo == lnkNet)) {
	  for(j=i+1 ; j<PIN_NUM ; j++)
	    if((pins[j].netNo == netNo) || (pins[i].netNo == lnkNet))
	      if(pins[j].equivNo != pins[i].equivNo)   return 0;
	  break;
      }
    
    return 1;
}
int
  isNetRouted( netNo )
int netNo;
/*  Check whether the given net has been routed or it is dummy, i.e. it
 *  has less than two independent pins */
{
    int   wn;           /*  Wire index  */
    int   isRouted;     /*  Returned value  */
    PATH  pth;          /*  Wire path  */

    wn = net_to_wire(netNo, 0);
    if(wn == NOTFOUND)    return DummyNet( netNo );
    pth = BuildPathFromWire(wn);
    isRouted = IsPathComplete( pth, netNo );
    FREE( pth );
    return isRouted;
}

void
  WireDeleteDuplicates(wn)
int  wn;
/*  Clean wire wn of all its duplicates.  */
{
    int  i, j,    /*  Loop counters .......................... */
         ws;      /*  Previous wire size ..................... */

    for(i=0,ws=wireSize(wn) ; i<ws ; i++)
      for(j=i+1 ; j<ws ; j++)
	if((wSegmRect(wn, j) == wSegmRect(wn, i)) && (wSegmSide(wn, j) == wSegmSide(wn, i))) {
	    ws--;
	    wSegmRect(wn, j) = wSegmRect(wn, ws);
	    wSegmSide(wn, j) = wSegmSide(wn, ws);
	    j--;
	}
    wireSize(wn) = ws;
    wSegmRect(wn, ws) = NULL;
}

/*  =================================================================
 *  In this function the given wire is searched for overlapping segments.
 *  Whenever two such segments are found, one is deleted, depending on
 *  layer characteristics. The segment maintained is the one
 *  selected by function <best_layer>.
 *  Wire size is kept updated.
 */
void
WireDeleteOverlap(wn)
int   wn;       /* wire index .............................. */
{
  int       i, k,                  /* Loop counters ................... */
            pos;                   /* wire position index ............. */
  RoadCoord wdt;                   /* wire spacing .................... */
  Rect      UpperRect, LowerRect;  /* rects of the two overlapped segms */
  RoadBox   BB;                    /* bounding box .................... */

  for(i=0 ; i<wireSize(wn) ; i++)	{
    if(wSegmSide(wn, i) == UP)  continue;

    if((UpperRect=neighbor(wSegmRect(wn, i), UP)) != NULL)
      if((pos=wire_segment_position(wn, UpperRect,wSegmSide(wn, i),i)) != NOTFOUND) {

/*  ------------------------------------------------------------
 *  Check if there is enough room for vias
 */
	LowerRect = wSegmRect(wn, i);
	wdt = spacingBetween(rLayer(LowerRect), UP, rLayer(LowerRect), NORTH);
	WireSegmentToBox(LowerRect, wSegmSide(wn, i), wdt, &BB,
			 NULL, NODIRECTION, NODIRECTION);
	if(newAvailableBox(wn,&BB, NULL, NODIRECTION, NODIRECTION,
			   LowerRect, wSegmSide(wn, i)))
	  continue;

	wdt = spacingBetween(rLayer(UpperRect), UP, rLayer(UpperRect), NORTH);
	WireSegmentToBox(UpperRect, wSegmSide(wn, i), wdt, &BB,
                            NULL, NODIRECTION, NODIRECTION);
	if(newAvailableBox(wn,&BB, NULL, NODIRECTION, NODIRECTION,
			   UpperRect, wSegmSide(wn, i)))
	   continue;
/*  ------------------------------------------------------------
 *  Add two vias if there are not
 */
	k = AbsDualDir(wSegmSide(wn, i));
	unless(wSegmHere(wn, LowerRect, UP)) {
	  add_to_wire(wn, LowerRect, UP);
	  if(wn < WIRE_NUM) set_free_neighbor(LowerRect, UP, 1);
	}
	if((LowerRect=neighbor(LowerRect, (-k))) != NULL)
	  unless(wSegmHere(wn, LowerRect, UP)) {
	    add_to_wire(wn, LowerRect, UP);
	    if(wn < WIRE_NUM) set_free_neighbor(LowerRect, UP, 1);
	  }

/*  ---------------------------------------------------------------
 *  Delete one of the two overlapped segments, choosing between the
 *  available layers
 */
	k = pos;
	if(best_layer(rLayer(wSegmRect(wn, i)),rLayer(UpperRect))==rLayer(UpperRect))
	  k = i--;

/*  ------------------------------------------------------------
 *  Delete the worse segment
 */
	if(wn < WIRE_NUM) set_free_neighbor(wSegmRect(wn, k), wSegmSide(wn, k), 0);
	sub_from_wire(wn, wSegmRect(wn, k), wSegmSide(wn, k));
      }
  }
}

/*  ============================================================
 *  This function adds a new element to a given wire.
 *  The wire memory is reallocated and wire size is updated.
 */
void
  add_to_wire(wn, r, side)
int  wn;             /*  Wire index                 */
Rect r;              /*  Pair to be added: rect ... */
int  side;           /*  ... and side               */
{
    if(r != NULL) {
	if( wireSize(wn) == 0) {
	    rdWire[wn].rect = NULL;
	    rdWire[wn].side = NULL;
	}
	rdWire[wn].rect = REALLOC(Rect, rdWire[wn].rect, (wireSize(wn)+2));
	rdWire[wn].side = REALLOC(int,  rdWire[wn].side, (wireSize(wn)+2));
	wSegmRect(wn, wireSize(wn))     = r;
	wSegmSide(wn, wireSize(wn))     = side;
	(wireSize(wn))++;
	wSegmRect(wn, wireSize(wn))     = NULL;
    }
}

/*  ============================================================
 *  This function subtracts an element from a given wire.
 *  Wire size is updated.
 */
void
sub_from_wire(wn, r, side)
int  wn;
Rect r;              /* Rect ...                       */
int  side;           /* ... and wire to be subtracted. */
{
  int i;             /* Loop counter */

  if((i=wire_segment_position(wn, r, side, NOTFOUND)) != NOTFOUND) {
      (wireSize(wn))--;
      wSegmRect(wn, i) = wSegmRect(wn, wireSize(wn));
      wSegmSide(wn, i) = wSegmSide(wn, wireSize(wn));
      wSegmRect(wn, wireSize(wn))  = NULL;
  }
}

void
rdSubFromWire(wn, pos)
int  wn;
int  pos;
/*  This function subtracts an element from a given wire.
 *  Wire size is updated. */
{
    (wireSize(wn))--;
    wSegmRect(wn, pos) = wSegmRect(wn, wireSize(wn));
    wSegmSide(wn, pos) = wSegmSide(wn, wireSize(wn));
    wSegmRect(wn, wireSize(wn))  = NULL;
}

void
fuseWires(wn1, wn2)
int  wn1, wn2;
{
    int  i;

    free_wire(wn2, 0);
    rdWire[wn1].rect = REALLOC(Rect, rdWire[wn1].rect, (wireSize(wn1)+wireSize(wn2)+2));
    rdWire[wn1].side = REALLOC(int,  rdWire[wn1].side, (wireSize(wn1)+wireSize(wn2)+2));
    for(i=0 ; i<wireSize(wn2) ; i++) {
	wSegmRect(wn1, wireSize(wn1)+i)   = wSegmRect(wn2, i);
	wSegmSide(wn1, wireSize(wn1)+i)   = wSegmSide(wn2, i);
    }
    wireSize(wn1) += wireSize(wn2);
    wSegmRect(wn1, wireSize(wn1))   = NULL;
    wireFree(wn1) = 0; free_wire(wn1, 1);
    delete_wire(wn2);
    deleteNetFromNetList(wireNetNo(wn1));
}
