/*       rtoSaveAll.C
 *
 *       Author: Enrico Malavasi
 *       Ver. 3.10
 *       Last revised:  April, 1991
 *
 *       Purpose: Utilities for the RoAD To Oct interface.
 */
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "utility.h"
#include "vov.h"
#include "region.h"
#include "symlib.h"
#include "roadGlobals.h"

#define  _CLEANUP  COMMON_FLAG
#define RoadToTransf(t)  ((enum octTransformType) t)

octObject         CONNECTORSBag;

void
rtoSaveAll( outputCell )
char *outputCell;
/*
 *  This function is the interface manager: it transforms the
 *  whole circuit into an Oct cell.
 */
{
    octObject    facet,  /*  Cell facet             */
                 origFacet,     /*  Unrouted circuit facet */
                 aTerm,
                 fTerm,         /*  Formal terminal        */
                 Net,
                 Path,
                 Inst,          /*  Instance item.         */
                 theProp;       /* Property      */
    octGenerator netGener,      /*  Net generator */
                 instGener,     /*  Instance generator  */
                 pathGener;     /*  Path generator  */
    char        *name;
    PathName     def;
    int          i;             /* Loop counters. */

    /*  ------------------------------------------------------------
     *  Copy the original facet and open the working facet
     */
    VERBOSE ( printf("-> Copy original facet into %s\n", outputCell) );
    name = util_strsav( outputCell );
    ohUnpackDefaults(&facet, "a", ":routed:contents");
    ohUnpackFacetName(&facet, name);
    VOVoutputFacet( &facet );
    
    sprintf(def, ":%s:contents", insts[0].view);
    ohUnpackDefaults(&origFacet, "r", def);
    ohUnpackFacetName(&origFacet, insts[0].name);
    VOVinputFacet( &origFacet );
    
    if(octCopyFacet(&facet, &origFacet) != OCT_OK) {
	errRaise("ROAD", 1, "Couldn't find the original facet %s", ohFormatName(&origFacet));
    }
    OH_ASSERT(octOpenFacet(&facet));
    
    /* Get or Create the CONNECTORS bag */
    OH_ASSERT(ohGetOrCreateBag(&facet,&CONNECTORSBag,"CONNECTORS"));
    
    /*  Delete all the connectors without a formal terminal on */
    OH_ASSERT(octInitGenContents(&CONNECTORSBag, OCT_INSTANCE_MASK, &instGener));
    while(octGenerate(&instGener, &Inst) == OCT_OK) {
	if(octGenFirstContent(&Inst, OCT_TERM_MASK, &aTerm) == OCT_OK)
	  if(octGenFirstContainer(&aTerm, OCT_TERM_MASK, &fTerm) == OCT_OK)
	    continue;
	octDelete(&Inst);
    }
    OH_ASSERT(octFreeGenerator(&instGener));
    
    VERBOSE ( printf("-> Delete pre-existing wires\n") );
    OH_ASSERT(octInitGenContents(&facet, OCT_NET_MASK, &netGener));
    while(octGenerate(&netGener, &Net) == OCT_OK) {
	octInitGenContents(&Net, OCT_PATH_MASK, &pathGener);
	while(octGenerate(&pathGener, &Path) == OCT_OK) {
	    octDelete(&Path);
	}
	octFreeGenerator(&pathGener);
    }
    octFreeGenerator(&netGener);

    if(RD_CHECK_FLAG(DEBUG_OPTION))  octFlushFacet( &facet );
    /*  Modify the position of all instances. */
    for(i=1 ; i<INST_NUM ; i++) {
	OH_ASSERT(ohGetByInstName(&facet, &Inst, insts[i].name));
	Inst.contents.instance.transform.translation.x = insts[i].x;
	Inst.contents.instance.transform.translation.y = insts[i].y;
	Inst.contents.instance.transform.transformType = RoadToTransf(insts[i].transf);
	octModify(&Inst);
    }

    if(RD_CHECK_FLAG(DEBUG_OPTION))  octFlushFacet( &facet );
    VERBOSE ( printf("-> Create new wires\n") );
    ForAllWires(i) {
	rtoCreateWire(&facet, i);
    }
    
    if(RD_CHECK_FLAG(DEBUG_OPTION))  octFlushFacet( &facet );
    if(RD_CHECK_FLAG(_CLEANUP)) {
	VERBOSE ( printf("-> Clean up all wires in the final facet\n") );
	OH_ASSERT(octInitGenContents(&facet, OCT_NET_MASK, &netGener));
	while(octGenerate(&netGener, &Net) == OCT_OK)
	  connReduce(&Net);
    }

    octCloseFacet(&facet);
}
