/*      rtoWire.c
 *
 *      Author:        Enrico Malavasi
 *      Ver.:          3.10
 *      Last revised:  January, 1991
 *
 *      Purpose: Transformation of `RoAD' wires into Oct paths
 *               for the Road-->Oct interface program.
 */
#include "port.h"
#include "oct.h"
#include "region.h"
#include "oh.h"
#include "symlib.h"
#include "roadGlobals.h"
#include  "transf.h"

extern octObject CONNECTORSBag;

typedef struct Segment {
    struct octPoint  Pts[2];
    int              NumPts;
    int              Lyr;
    octId            Id;
} Segment;

#define  MAX_CONNECT_NUM    10
typedef struct Connector {
    RoadCoord     x;
    RoadCoord     y;
    int           segment[MAX_CONNECT_NUM];
} Connector;

static   int        ConnNum=0;            /*  Connector number       */
static   Segment   *Arg;                  /*  Wire segments ........ */

/*  ============================================================
 *  This function returns YES if the two given connectors are
 *  perfectly overlapped.
 */
static RoadStatus
  sameConnector(c1, c2)
Connector *c1, *c2;
{
    int  i1, i2;

    if((c1->x == c2->x) &&  (c1->y == c2->y))
      for(i1=0 ; c1->segment[i1]!=NOTFOUND ; i1++)
	for(i2=0 ; c2->segment[i2]!=NOTFOUND ; i2++)
	  if(Arg[c1->segment[i1]].Lyr == Arg[c2->segment[i2]].Lyr)
	    return YES;
    return NO;
}

static void
  TermNameGenerate(NetNo, pinNo, TermNo, TermName)
int   NetNo;       /*  Net number       */
int   pinNo;       /*  Terminal number  */
int   TermNo;      /*  Terminal number  */
char *TermName;    /*  Terminal name  (output)  */
{
    if(pins[pinNo].TermName != NULL)
      if(pins[pinNo].TermName[0] != '\0') {
	  strcpy(TermName, pins[pinNo].TermName);
	  return;
      }
    if(rd_net[NetNo].name != NULL)
      if(rd_net[NetNo].name[0] != '\0') {
	  sprintf(TermName, "%s.%d", rd_net[NetNo].name, TermNo);
	  return;
      }
    sprintf(TermName, "term_%s.%d", rd_net[NetNo].name, TermNo);
}

/*  ============================================================
 *  This routine creates a new path on the given layer
 */
static void
  newPath(Layer, Path, Net, segment, PathWidth)
octObject       *Layer;       /* Path layer         */
octObject       *Path;        /* Return object      */
octObject       *Net;
struct octPoint *segment;     /* Segment extreme points */
RoadCoord        PathWidth;   /* Path width         */
{
    octCoord wdt;

    segment[0].x = segment[0].x;
    segment[0].y = segment[0].y;
    segment[1].x = segment[1].x;
    segment[1].y = segment[1].y;
    wdt = PathWidth;
    ohCreatePath(Layer, Path, wdt);
    octPutPoints(Path, 2, segment);
    OH_ASSERT(octAttachOnce(Net, Path));
}

static void
  unpack_name(name, view)
char *name,
     *view;
{
    char *s, *index();

    strcpy(view, "physical"); /* default */
    s = name;
    if((s=index(s,':')) != NULL) {
	*s++ = '\0';
	strcpy(view, s);
    }
}

static void
  formConn(Facet, Net, NetNo, wn)
octObject *Facet;          /* Main facet  */
octObject *Net;
int    NetNo;
int    wn;
{
    octObject     aTerm,
                  connInstance,
                  fTerm;
    octPoint      pnt;
    struct octBox BB;
    static int    TermCounter = 0;
    int           i, RoadOctSide;
    PathName      TermName;

    ForAllPins( i ) {
	if(pins[i].netNo == NetNo) {
	    if(pins[i].cellNo == NOTFOUND)  continue;
	    if(pins[i].cellNo == 0) {
		pnt.x = (pins[i].side == WEST)  ? cells[0].x1 : pin_x(pins+i);
		pnt.y = (pins[i].side == SOUTH) ? cells[0].y1 : pin_y(pins+i);
		pnt.x = (pnt.x);
		pnt.y = (pnt.y);
		if(regFindActual(Facet, &pnt, &aTerm, &BB) == REG_FAIL) {
		    warnMessage("No term found under pin n.%d", i);
		    continue;
		}
		octGenFirstContainer(&aTerm, OCT_INSTANCE_MASK, &connInstance);
		TermNameGenerate(NetNo, i, TermCounter++, TermName);
		ohCreateTerm(Facet, &fTerm, TermName);
		octAttach(&fTerm, &aTerm);
		OH_ASSERT(octAttachOnce(Net, &fTerm));
		if(pins[i].side == NORTH)  RoadOctSide = SYM_PUT_CONST_TOP;
		if(pins[i].side == SOUTH)  RoadOctSide = SYM_PUT_CONST_BOTTOM;
		if(pins[i].side == EAST)   RoadOctSide = SYM_PUT_CONST_RIGHT;
		if(pins[i].side == WEST)   RoadOctSide = SYM_PUT_CONST_LEFT;
		symSideConstraint(&connInstance, RoadOctSide, 0, 0);
		break;
	    }
	}
    }
}

/*  ============================================================
 *  This function creates a new connector instance, extracts its
 *  actual terminal and attaches it to the proper segments.
 */
static void
  newConnector(Facet, Net, conn, Arg, NetNo, wn)
octObject *Facet;          /* Main facet  */
octObject *Net;
struct Connector *conn;
struct Segment  *Arg;
int    NetNo;
int    wn;
{
    int          i, j,
                 ll[LAYER_MAX],
                 RoadOctSide;
    octCoord     widths[LAYER_MAX];
    octObject    ConnInstance,
                 actTerm,       /*  Connector actual terminal. */
                 forTerm,       /*  Formal terminal  */
                 inst,
                 Layers[LAYER_MAX],
                 Sgmnt;         /*  Segment.                   */
    PathName     name,
                 view,
                 TermName;      /* name of the extracted terminal.  */
    static int   TermCounter = 0;
    RoadCoord    xx, yy;
    RoadInstance II;

    II.name = II.master = II.view = NULL;
    if(conn->segment[0] == NOTFOUND) {
	fprintf(stderr, "\n*** Bad connector located in (x=%d,y=%d)\n",
		conn->x, conn->y);
	exit();
    }
    ll[0] = ll[1] = Arg[conn->segment[0]].Lyr;
    for(i=1 ; conn->segment[i]!=NOTFOUND ; i++)
      if(i == (MAX_CONNECT_NUM-1)) {
	  warnMessage("Wrong connector attachments in net %s\nSymbolic policy violation is possible", rd_net[NetNo].name);
	  conn->segment[MAX_CONNECT_NUM-1] = NOTFOUND;
	  break;
      } 
    for(i=1 ; conn->segment[i]!=NOTFOUND ; i++)
      if(Arg[conn->segment[i]].Lyr != ll[0]) {
	  ll[1] = Arg[conn->segment[i]].Lyr;
	  break;
      }
    
    for(i=0 ; i<2 ; i++) {
	ohGetOrCreateLayer(Facet, Layers+i, layers[ll[i]].name);
	widths[i] = (wireDesc(wn) == MINWWIDTH) ?
	  layers[ll[i]].wire_width : wSegmWidth(wn, 0);
    }
    
    ConnNameGenerate(2, Layers, widths, name);
    II.master = util_strsav( name );
    unpack_name(II.master, view);
    II.view = view;
    II.name = util_strsav( "*" );
    
    OH_ASSERT(ohCreateInstance(Facet, &ConnInstance,
			       II.name, II.master, II.view,
			       conn->x, conn->y, OCT_NO_TRANSFORM));
    OH_ASSERT(octGenFirstContent(&ConnInstance, OCT_TERM_MASK, &actTerm));
    OH_ASSERT(octAttach(&CONNECTORSBag, &ConnInstance));
    
    for(i=0 ; conn->segment[i]!=NOTFOUND ; i++) {
	ohGetById(&Sgmnt, Arg[conn->segment[i]].Id);
	OH_ASSERT(octAttachOnce(&Sgmnt, &actTerm));
    }
    
    /*  Attach all wires to the terminals coming from instances   */
    ForAllPins( j ) {
	if(pins[j].netNo == NetNo) {
	    if(pins[j].cellNo == NOTFOUND) continue;
	    if(pins[j].cellNo == 0) continue;
	    
	    xx = (pins[j].side == EAST)  ? cells[pins[j].cellNo].x2 : pin_x(pins+j);
	    yy = (pins[j].side == NORTH) ? cells[pins[j].cellNo].y2 : pin_y(pins+j);
	    if((xx == conn->x) && (yy == conn->y) &&
	       (pins[j].layer == Arg[conn->segment[0]].Lyr)) {
		octGenFirstContainer(&actTerm, OCT_GEO_MASK, &Sgmnt);
		if(conn->segment[1] == NOTFOUND)    octDelete(&ConnInstance);
		else                                octDetach(&Sgmnt, &actTerm);

		OH_ASSERT(ohGetByInstName(Facet, &inst, insts[cells[pins[j].cellNo].instanceNo].name));
		OH_ASSERT(ohGetByTermName(&inst, &actTerm, pins[j].TermName));

		OH_ASSERT(octAttachOnce(&Sgmnt,  &actTerm));
		OH_ASSERT(octAttachOnce(Net,     &actTerm));
		break;
	    }
	}
    }
}

static
  getConnSegment(conn, segmNo)
struct Connector *conn;
int               segmNo;
{
    int k;
    for(k=0 ; conn->segment[k]!= NOTFOUND ; k++)
      if(conn->segment[k] == segmNo)
	return segmNo;
    return NOTFOUND;
}
static void
addConnSegment(conn, segmNo)
struct Connector *conn;
int               segmNo;
{
  int k;
  for(k=0 ; conn->segment[k]!= NOTFOUND ; k++)
    if(conn->segment[k] == segmNo)
      return;
  conn->segment[k++] = segmNo;
  conn->segment[k] = NOTFOUND;
  return;
}
/*  ============================================================
 *  This function transforms a RoAD wire into an Oct wire
 */
rtoCreateWire(Facet, wn)
octObject *Facet;        /*  Main facet   */
int        wn;           /*  Wire index   */
{
    octObject       Net;                 /* Net of the wire  */
    octObject       Path;                /* Path .............. */
    octObject       LL[LAYER_MAX];       /* Layer pointers list */
    int             ws;                   /*  Wire segment number    */
    Connector      *connectors;
    octCoord        widths[LAYER_MAX];
    RoadCoord       xx, yy;              /* Temp. variables  */
    int             pos, i, j, k, l, jj, pntNo;   /* Loop counters.   */


    Arg        = ALLOC(struct Segment,     wireSize(wn)+1+PIN_NUM);
    connectors = ALLOC(struct Connector, 4*wireSize(wn)+1+PIN_NUM);
    WireDeleteDuplicates(wn);


    for(i=0 ; i<wireSize(wn) ; i++)
      if(wSegmSide(wn, i) == UP)
	for(j=i+1 ; j<wireSize(wn) ; j++)
	  if(wSegmSide(wn, j) == UP)
	    if((rRightX(wSegmRect(wn, i)) == rRightX(wSegmRect(wn, j))) &&
	       (rUpperY(wSegmRect(wn, i)) == rUpperY(wSegmRect(wn, j))))
	      warnMessage("Stacked vias in net %s", rd_net[wireNetNo(wn)].name);

    /*  Set the path as a trivial collection of small segments.  */
    for(i=ws=0 ; i<wireSize(wn) ; i++) {
	if(wSegmSide(wn, i) != UP) {
	    Arg[ws].Pts[1].x = wSegmRect(wn, i)->x2;
	    Arg[ws].Pts[1].y = wSegmRect(wn, i)->y2;
	    Arg[ws].Lyr      = wSegmLayer(wn, i);
	    
	    if(wSegmSide(wn, i) == NORTH) {
		Arg[ws].Pts[0].x   = wSegmRect(wn, i)->x1;
		Arg[ws++].Pts[0].y = wSegmRect(wn, i)->y2;
	    }
	    else {
		Arg[ws].Pts[0].x   = wSegmRect(wn, i)->x2;
		Arg[ws++].Pts[0].y = wSegmRect(wn, i)->y1;
	    }
	}
    }
    
    /*  ------------------------------------------------------------
     *  Now link the path segments.
     */
    for(i=0 ; i<ws ; i++) {
	xx = Arg[i].Pts[1].x;
	yy = Arg[i].Pts[1].y;
	
	for(pos=0 ; pos<wireSize(wn) ; pos++)
	  if(wSegmSide(wn, pos) == UP)
	    if((rRightX(wSegmRect(wn, pos)) == xx) && (rUpperY(wSegmRect(wn, pos)) == yy))
	      if((rLayer(wSegmRect(wn, pos)) == Arg[i].Lyr) ||
		 (rLayer(wSegmRect(wn, pos)) == (Arg[i].Lyr-1)))
		break;
	if(pos < wireSize(wn)) continue;
	
	if(Arg[i].Pts[0].y == Arg[i].Pts[1].y) {
	    for(j=0 ; j<ws ; j++)
	      if(Arg[j].Pts[0].y == Arg[j].Pts[1].y)
		if(Arg[j].Pts[0].y == yy)
		  if(Arg[j].Pts[0].x == xx)
		    if(Arg[j].Lyr == Arg[i].Lyr) {
			Arg[i--].Pts[1].x = Arg[j].Pts[1].x;
			Arg[j] = Arg[--ws];
			break;
		    }
	} else
	  if(Arg[i].Pts[0].x == Arg[i].Pts[1].x)
	    for(j=0 ; j<ws ; j++)
	      if(Arg[j].Pts[0].x == Arg[j].Pts[1].x)
		if(Arg[j].Pts[0].x == xx)
		  if(Arg[j].Pts[0].y == yy)
		    if(Arg[j].Lyr == Arg[i].Lyr) {
			Arg[i--].Pts[1].y = Arg[j].Pts[1].y;
			Arg[j] = Arg[--ws];
			break;
		    }
    }
    /*  ------------------------------------------------------------
     *  Cut the segments in the "tee" junctions.
     */
    for(i=0 ; i<ws ; i++) {
	xx = Arg[i].Pts[0].x;
	yy = Arg[i].Pts[0].y;
	for(j=0 ; j<ws ; j++) {
	    if(i == j) continue;
	    if((Arg[i].Pts[1].y == yy) &&                /* <i> is horizontal   */
	       (Arg[j].Pts[0].x == Arg[j].Pts[1].x) &&   /* <j> is vertical     */
	       (Arg[j].Pts[0].y <= yy) &&                /* They cross each other */
	       (Arg[j].Pts[1].y >= yy) &&
	       (Arg[j].Pts[0].x > xx) &&             /* There is a <tee> junction */
	       (Arg[j].Pts[0].x < Arg[i].Pts[1].x)) {
		Arg[ws] = Arg[i];
		Arg[ws].Pts[0].x = Arg[i].Pts[1].x  = Arg[j].Pts[0].x;
		ws++;
		continue;
	    }
	    if((Arg[i].Pts[1].x == xx) &&                /* <i> is vertical     */
	       (Arg[j].Pts[0].y == Arg[j].Pts[1].y) &&   /* <j> is horizontal   */
	       (Arg[j].Pts[0].x <= xx) &&                /* They cross each other */
	       (Arg[j].Pts[1].x >= xx) &&
	       (Arg[j].Pts[0].y > yy) &&             /* There is a <tee> junction */
	       (Arg[j].Pts[0].y < Arg[i].Pts[1].y)) {
		Arg[ws] = Arg[i];
		Arg[ws].Pts[0].y = Arg[i].Pts[1].y  = Arg[j].Pts[0].y;
		ws++;
	    }
	}
    }
    
    /*  ============================================================
     *  Now all path segments have been created.
     *  ============================================================ */
    /*  ------------------------------------------------------------
     *  Build up the connectors array.
     *  1.)  Vias. 2.) Same-layer connectors.
     */
    ConnNum = 0;
    for(pos=0 ; pos<wireSize(wn) ; pos++)
      if(wSegmSide(wn, pos) == UP) {
	  l                          = rLayer(wSegmRect(wn, pos));
	  connectors[ConnNum].x = xx = rRightX(wSegmRect(wn, pos));
	  connectors[ConnNum].y = yy = rUpperY(wSegmRect(wn, pos));
	  for(i=k=0 ; i<ws ; i++)
	    if(((Arg[i].Pts[0].x == xx) && (Arg[i].Pts[0].y == yy)) ||
	       ((Arg[i].Pts[1].x == xx) && (Arg[i].Pts[1].y == yy)))
	      if((Arg[i].Lyr == l) ||
		 (Arg[i].Lyr == l+1))
		connectors[ConnNum].segment[k++] = i;
	  connectors[ConnNum++].segment[k] = NOTFOUND;
      }
    
    for(i=0 ; i<ws ; i++)
      for(pntNo=0 ; pntNo<2 ; pntNo++) {
	  xx = Arg[i].Pts[pntNo].x;
	  yy = Arg[i].Pts[pntNo].y;
	  for(j=l=0 ; (j<ConnNum)&&(l==0) ; j++)
	    if((connectors[j].x == xx) && (connectors[j].y == yy))
	      for(k=0 ; connectors[j].segment[k]!=NOTFOUND ; k++)
		if(Arg[connectors[j].segment[k]].Lyr == Arg[i].Lyr) {
		    addConnSegment(connectors+j, i);
		    l = 1;
		    break;
		}
	  if(l == 0) {
	      connectors[ConnNum].x = xx;
	      connectors[ConnNum].y = yy;
	      connectors[ConnNum].segment[0] = i;
	      connectors[ConnNum++].segment[1] = NOTFOUND;
	  }
      }
    
    connectors[ConnNum].x = connectors[ConnNum].y = (-1);
    
    for(i=0 ; i<LAYER_NUM ; i++)
      ohGetOrCreateLayer(Facet, LL+i, layers[i].name);
    
    ohGetOrCreateNet(Facet, &Net, rd_net[wireNetNo(wn)].name);
    for(i=0 ; i<LAYER_NUM ; i++)
      widths[i] = (wireDesc(wn) == MINWWIDTH) ?
	layers[i].wire_width : wSegmWidth(wn, 0);
    for(i=0 ; i<ws ; i++) {
	newPath(LL+Arg[i].Lyr, &Path, &Net, Arg[i].Pts,
		(RoadCoord) widths[Arg[i].Lyr]);
	Arg[i].Id = Path.objectId;
    }

    /*  ------------------------------------------------------------
     *  Build up the set of connectors.
     */
    for(i=0 ; i<ConnNum ; i++)
      for(j=i+1 ; j<ConnNum ; j++)
	if(sameConnector(connectors+i, connectors+j) == YES) {
	    for(l=0 ; connectors[i].segment[l]!=NOTFOUND ; l++) ;
	    for(k=0 ; connectors[j].segment[k]!=NOTFOUND ; k++)
	      connectors[i].segment[l++] = connectors[j].segment[k];
	    for(jj=0 ; jj<l ; jj++)
	      for(k=jj+1 ; k<l ; k++)
		if(connectors[i].segment[k] == connectors[i].segment[jj])
		  connectors[i].segment[k--] = connectors[i].segment[--l];
	    connectors[i].segment[l] = NOTFOUND;
	    connectors[j--] = connectors[--ConnNum];
	}
    
    for(i=0 ; i<ConnNum ; i++)
      newConnector(Facet, &Net, connectors+i, Arg, wireNetNo(wn), wn);
/*    formConn(Facet, &Net, wireNetNo(wn), wn); */
    
    FREE(Arg);
    FREE(connectors);
}
