#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "octspice.h"

FILE *ofeDeckStream;	/* for writing out a spice deck */

static octObject groundObj;	/* XXX -- back door from ofeGround */
static analDesc *curAnalPtr = NIL(analDesc);	/* back door to ofePlot */

dary ofeInitSimulations()
{
    dary analDialogs;
    struct dialog dlg;
    int i;
    int fillAnalDialog();

    groundObj.objectId = oct_null_id;
    ofeOutInit();
    ofeInitSimulator();

    analDialogs = daAlloc(struct dialog, 0);
    INIT_DIALOG(&dlg);
    for (i = 0; i < ofeSimPtr->numAnalyses; i++) {
	if (fillAnalDialog(ofeSimPtr->analyses[i], i, &dlg)) {
	    *daSetLast(struct dialog, analDialogs) = dlg;
	    INIT_DIALOG(&dlg);
	}
    }

    return(analDialogs);
}

static int fillAnalDialog(simAnalPtr, index, dlgPtr)
IFanalysis *simAnalPtr;
int index;
struct dialog *dlgPtr;
{
    IFparm *parmPtr;
    int i;
    struct flagStruct *flagPtr;
    struct scalarStruct *scalarPtr;

    for (i = 0; i < simAnalPtr->numParms; i++) {
	parmPtr = &simAnalPtr->analysisParms[i];
	if (parmPtr->dataType & IF_SET) {
	    switch (parmPtr->dataType & IF_VARTYPES) {
		case IF_FLAG:
		    flagPtr = daSetLast(struct flagStruct, dlgPtr->flags);
		    INIT_FLAG(flagPtr);
		    flagPtr->name = parmPtr->keyword;
		    flagPtr->parmPtr = parmPtr;
		    break;
		case IF_INTEGER:
		case IF_REAL:
		    scalarPtr = daSetLast(struct scalarStruct, dlgPtr->scalars);
		    INIT_SCALAR(scalarPtr);
		    scalarPtr->name = parmPtr->keyword;
		    scalarPtr->parmPtr = parmPtr;
		    break;
		default:
		    daReset(dlgPtr->flags);
		    daReset(dlgPtr->scalars);
		    return(0);
	    }
	}
    }

    dlgPtr->analIndex = index;
    dlgPtr->simAnalPtr = simAnalPtr;
    dlgPtr->title = simAnalPtr->description;

    return(1);
}

void ofeChooseGround(objPtr)
octObject *objPtr;	/* the ground terminal or net */
{
    groundObj = *objPtr;
}

void ofeDeck(facetPtr, fileName)
octObject *facetPtr;
char *fileName;
{
    cktDesc *cktPtr;
    int checkGround();

    if ( ! checkGround(facetPtr, &groundObj)) return;

    if ((ofeDeckStream = fopen(fileName, "w")) == NULL) {
	ofeMessage(SEVERE, "can't open %s for writing\n", fileName);
	return;
    }

    ofeSimPtr = ofeDeckPtr;
    cktPtr = ofeCircuit(facetPtr, &groundObj);
    ofeFreeCircuit(cktPtr);
    ofeSimPtr = ofeSpicePtr;

    (void) fclose(ofeDeckStream);
}

void ofeAnalysis(facetPtr, dlgPtr)
octObject *facetPtr;
struct dialog *dlgPtr;
{
    static cktDesc *circuitPtr = NIL(cktDesc);
    analDesc *analPtr;
    IFuid uid;
    void setAnalysisParms();
    int checkGround();

    if ( ! checkGround(facetPtr, &groundObj)) return;

    if ( ! ofeDisplayDialog(dlgPtr)) {
	ofeMessage(INFO, "analysis aborted\n");
	return;
    }

    /* create the circuit */
    ofeMessage(INFO, "processing circuit...\n");
    if (circuitPtr) ofeFreeCircuit(circuitPtr);
    circuitPtr = ofeCircuit(facetPtr, &groundObj);

    /* create the task */
    ofeMessage(INFO, "beginning analysis...\n");
    OFE_NEW_UID(&uid, (IFuid) circuitPtr, progName, UID_TASK);
    circuitPtr->taskPtr = (taskDesc *) uid;
    IF_ASSERT(newTask, (circuitPtr->simEquiv, &circuitPtr->taskPtr->simEquiv, uid));

    /* create the analysis */
    OFE_NEW_UID(&uid, (IFuid) circuitPtr, dlgPtr->simAnalPtr->description,
					    UID_ANALYSIS);
    circuitPtr->taskPtr->analPtr = analPtr = (analDesc *) uid;
    IF_ASSERT(newAnalysis, (circuitPtr->simEquiv, dlgPtr->analIndex, uid,
			&analPtr->simEquiv, circuitPtr->taskPtr->simEquiv));
    curAnalPtr = analPtr;

    /* do the analysis */
    setAnalysisParms(analPtr, dlgPtr);
    IF_CHECK(doAnalyses,
		(circuitPtr->simEquiv, RESTART, circuitPtr->taskPtr->simEquiv),
		SEVERE,
		{ofeMessage(INFO, "analysis aborted\n"); return;} );
    IF_ASSERT(deleteTask, (circuitPtr->simEquiv, circuitPtr->taskPtr->simEquiv));

    ofeMessage(INFO, "analysis completed\n");
}

static int checkGround(facetPtr, groundPtr)
octObject *facetPtr;
octObject *groundPtr;
{
    octGenerator netGen;
    octObject net;
    octObject prop;
    int findCount;

    if (octGetById(groundPtr) == OCT_OK) return(1);

    /* try to find the right net by property */
    findCount = 0;
    octInitGenContents(facetPtr, OCT_NET_MASK, &netGen);
    while (octGenerate(&netGen, &net) == OCT_OK) {
	if (ohGetByPropName(&net, &prop, "NETTYPE") == OCT_OK &&
		prop.contents.prop.type == OCT_STRING &&
		strcmp(prop.contents.prop.value.string, "GROUND") == 0) {
	    *groundPtr = net;
	    findCount++;
	}
    }
    switch (findCount) {
	case 0:
	    ofeMessage(SEVERE,
		    "facet has no ground nets -- you must specify one\n");
	    groundPtr->objectId = oct_null_id;
	    return(0);
	case 1:
	    return(1);
	default:
	    ofeMessage(SEVERE,
		    "facet has %d ground nets -- you must specify one\n",
		    findCount);
	    groundPtr->objectId = oct_null_id;
	    return(0);
    }
}

static void setAnalysisParms(analPtr, dlgPtr)
analDesc *analPtr;
struct dialog *dlgPtr;
{
    int i;
    struct flagStruct *flagPtr;
    struct scalarStruct *scalarPtr;
    IFvalue value;

    /* set flags */
    for (i = 0; i < daLen(dlgPtr->flags); i++) {
	flagPtr = daGet(struct flagStruct, dlgPtr->flags, i);
	value.iValue = flagPtr->flag;
	IF_ASSERT(setAnalysisParm,
			    (analPtr->cktPtr->simEquiv, analPtr->simEquiv,
			     flagPtr->parmPtr->id, &value, NIL(IFvalue)));
    }

    /* set scalars */
    for (i = 0; i < daLen(dlgPtr->scalars); i++) {
	scalarPtr = daGet(struct scalarStruct, dlgPtr->scalars, i);
	switch (scalarPtr->parmPtr->dataType & IF_VARTYPES) {
	    case IF_INTEGER:
		value.iValue = scalarPtr->value;
		break;
	    case IF_REAL:
		value.rValue = scalarPtr->value;
		break;
	    default:
		ASSERT(0);
	}
	IF_ASSERT(setAnalysisParm,
			(analPtr->cktPtr->simEquiv, analPtr->simEquiv,
			 scalarPtr->parmPtr->id, &value, NIL(IFvalue)));
    }
}

/*
 *	ofeNodePlot
 *
 *	This is a simplified plotting function, that plots only node
 *	voltages against the independent variable.  It calls the more
 *	general ofePlot to do the real work.
 */
void ofeNodePlot(genPtr, fName)
octGenerator *genPtr;	/* generates nets to be plotted */
char *fName;		/* name of output file, or NIL(char) to fork xgraph */
{
    lsList nodeList;
    octObject net;
    char *val;
    int nSigs;
    void ofePlot();

    if (curAnalPtr == NIL(analDesc)) {
	ofeMessage(SEVERE,
		    "you must do an analysis before you can plot anything\n");
	return;
    }
    if (curAnalPtr->indepWave == NIL(waveStruct)) {
	ofeMessage(SEVERE, "no data for the independent variable (?)\n");
	return;
    }

    nodeList = lsCreate();
    nSigs = 0;
    while (octGenerate(genPtr, &net) == OCT_OK) {
	if (net.type != OCT_NET) {
	    ofeMessage(SEVERE, "can only plot voltages on nets\n");
	} else if (st_lookup(curAnalPtr->waveTable,
				(char *) net.objectId, &val)) {
	    lsNewEnd(nodeList, (lsGeneric) (waveStruct *) val, LS_NH);
	    nSigs++;
	} else if (net.objectId == curAnalPtr->indepWave->netPtr->netId) {
	    lsNewEnd(nodeList, (lsGeneric) curAnalPtr->indepWave, LS_NH);
	    nSigs++;
	} else if (net.objectId == groundObj.objectId) {
	    ofeMessage(INFO, "ground node (%s) omitted\n", ohFormatName(&net));
	} else {
	    ofeMessage(SEVERE, "no data available for %s\n",ohFormatName(&net));
	}
    }

    if (nSigs) ofePlot(curAnalPtr->indepWave, nodeList, fName);

    lsDestroy(nodeList, (void (*)()) 0);
}

void ofePlot(indepWave, waveList, fName)
waveStruct *indepWave;	/* the wave to use for the X axis */
lsList waveList;	/* list of waves to plot */
char *fName;		/* name of output file, or NIL(char) to fork xgraph */
{
    int dependFlags, indepFlags;

    if ( ! queryPlotFlags(&indepFlags, &dependFlags)) {
	ofeMessage(INFO, "plot aborted\n");
	return;
    }

    ofeGraph(curAnalPtr, indepFlags, indepWave, dependFlags, waveList, fName);
}

static int queryPlotFlags(indepPtr, dependPtr)
int *indepPtr;
int *dependPtr;
{
    static struct dialog plotFlags;
    static int initializedP = 0;
    struct flagStruct *flagPtr;
    int i;
    void fillPlotDialog();

    if ( ! initializedP) {
	initializedP = 1;
	INIT_DIALOG(&plotFlags);
	fillPlotDialog(&plotFlags);
    }

    if ( ! ofeDisplayDialog(&plotFlags)) return(0);

    *indepPtr = *dependPtr = 0;
    for (i = 0; i < daLen(plotFlags.flags); i++) {
	flagPtr = daGet(struct flagStruct, plotFlags.flags, i);
	if (flagPtr->flag) {
	    if (i < daLen(plotFlags.flags)/2) {
		*dependPtr |= flagPtr->weight;
	    } else {
		*indepPtr |= flagPtr->weight;
	    }
	}
    }

    return(1);
}

static void fillPlotDialog(dlgPtr)
struct dialog *dlgPtr;
{
    static struct {
	char *name;
	int initVal;
	int weight;
    } plotOptions[] = {
	{	"Real part",		1,	PFL_REAL	},
	{	"Imaginary part",	0,	PFL_IMAG	},
	{	"Magnitude",		0,	PFL_MAG		},
	{	"dB(magnitude)",	0,	PFL_DB		},
	{	"Phase",		0,	PFL_PHASE	},
	{	"Logarithmic axis",	0,	PFL_LOG		}
    };
    int nOpts = sizeof(plotOptions)/sizeof(plotOptions[0]);
    int i;
    struct flagStruct *flagPtr;

    INIT_DIALOG(dlgPtr);

    *daSetLast(char *, dlgPtr->flagColTitles) = "Y axis";
    for (i = 0; i < nOpts; i++) {
	flagPtr = daSetLast(struct flagStruct, dlgPtr->flags);
	INIT_FLAG(flagPtr);
	flagPtr->name = plotOptions[i].name;
	flagPtr->flag = plotOptions[i].initVal;
	flagPtr->weight = plotOptions[i].weight;
    }

    *daSetLast(char *, dlgPtr->flagColTitles) = "X axis";
    for (i = 0; i < nOpts; i++) {
	flagPtr = daSetLast(struct flagStruct, dlgPtr->flags);
	INIT_FLAG(flagPtr);
	flagPtr->name = plotOptions[i].name;
	flagPtr->flag = plotOptions[i].initVal;
	flagPtr->weight = plotOptions[i].weight;
    }
}
