#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "IFsim.h"
#include "IFerrmsgs.h"
#include "deck.h"

extern IFsimulator deckInfo;

static GENERIC *deckCurrentTask = NIL(GENERIC);
static GENERIC *deckCurrentAnal = NIL(GENERIC);

static GENERIC *deckCurrentCkt = NIL(GENERIC);
static IFnode deckGroundNode = NIL(GENERIC);

static struct nameSpace deckModelSpace, deckInstanceSpace, deckNodeSpace;

	/* Task-handling functions */



static int deckNewTask(cktPtr, taskPtr, uid)
GENERIC *cktPtr;
GENERIC **taskPtr;
IFuid uid;
{
    CHECK_CKT(cktPtr);
    if (deckCurrentTask) {
	errRtn = "can't define a new task while old one still around";
	return(E_EXISTS);
    }
    deckCurrentTask = (GENERIC *) uid;
    *taskPtr = deckCurrentTask;
    return(OK);
}

static int deckDeleteTask(cktPtr, taskPtr)
GENERIC *cktPtr;
GENERIC *taskPtr;
{
    CHECK_CKT(cktPtr);
    CHECK_SAME(taskPtr, deckCurrentTask, "not current task");
    deckCurrentAnal = NIL(GENERIC);
    deckCurrentTask = NIL(GENERIC);
    return(OK);
}

static void deckFlushTask()
{
    /* XXX */
}

	/* Analysis-handling functions */



/*ARGSUSED*/
static int deckNewAnalysis(cktPtr, analType, analUid, analPtr, taskPtr)
GENERIC *cktPtr;
int analType;
IFuid analUid;
GENERIC **analPtr;
GENERIC *taskPtr;
{
    CHECK_CKT(cktPtr);
    CHECK_SAME(taskPtr, deckCurrentTask, "not current task");
    if (deckCurrentAnal) {
	errRtn = "can't define a new analysis while old one still around";
	return(E_EXISTS);
    }
    deckCurrentAnal = (GENERIC *) analUid;
    *analPtr = deckCurrentAnal;
    return(OK);
}

/*ARGSUSED*/
static int deckSetAnalysisParm(cktPtr, analPtr, parmIndex, value, sel)
GENERIC *cktPtr;
GENERIC *analPtr;
int parmIndex;
IFvalue *value;
IFvalue *sel;
{
    errRtn = "deckSetAnalysisParm not implemented";
    return(E_UNSUPP);		/* XXX */
}

/*ARGSUSED*/
static int deckDoAnalyses(cktPtr, restart, taskPtr)
GENERIC *cktPtr;
int restart;
GENERIC *taskPtr;
{
    void deckFlushTask();

    CHECK_CKT(cktPtr);
    CHECK_SAME(taskPtr, deckCurrentTask, "not current task");
    deckFlushTask();
    return(OK);
}

	/* Circuit-handling functions */




static int deckNewCircuit(cktPtr)
GENERIC **cktPtr;
{
    void deckInitNameSpace();

    if (deckCurrentCkt) {
	errRtn = "can't define a new circuit while old one still around";
	return(E_EXISTS);
    }
    deckInitNameSpace(&deckModelSpace, "model", 1);
    deckInitNameSpace(&deckInstanceSpace, "i", 0);
    deckInitNameSpace(&deckNodeSpace, "n", 0);

    deckCurrentCkt = (GENERIC *) "";
    *cktPtr = deckCurrentCkt;
    return(OK);
}

static int deckDeleteCircuit(cktPtr)
GENERIC *cktPtr;
{
    void deckFreeNameSpace();

    CHECK_CKT(cktPtr);
    deckFlushInstances();
    deckFlushModels();

    deckFreeNameSpace(&deckModelSpace);
    deckFreeNameSpace(&deckInstanceSpace);
    deckFreeNameSpace(&deckNodeSpace);

    deckGroundNode = NIL(GENERIC);
    deckCurrentCkt = NIL(GENERIC);
    return(OK);
}

	/* Node-handling functions */




static int deckNewNode(cktPtr, nodePtr, uid)
GENERIC *cktPtr;
IFnode *nodePtr;
IFuid uid;
{
    char *deckUniqueName();

    CHECK_CKT(cktPtr);
    *nodePtr = (IFnode) deckUniqueName(&deckNodeSpace, uid);
    return(OK);
}

static int deckSetGroundNode(cktPtr, nodePtr, uid)
GENERIC *cktPtr;
IFnode *nodePtr;
IFuid uid;
{
    CHECK_CKT(cktPtr);
    if (deckGroundNode) {
	errRtn = "Ground node already defined";
	return(E_EXISTS);
    }
    deckGroundNode = (IFnode) "0";
    *nodePtr = deckGroundNode;

    /* XXX -- assumes this function is called first after circuit creation */
    (void) fprintf(DECK_STREAM, "* CIRCUIT %s\n", ofeCktNameOf(uid));

    return(OK);
}

	/* Model-handling functions */




static struct model *curDefModel = NIL(struct model);
static char bigString[5000];
static char *modTypeName = NIL(char);

static char *magicC = "AVIUEHGFRTCLKMMMDQJMWZS"; /* magic first letters */

static int deckNewModel(cktPtr, devType, modelPtr, uid)
GENERIC *cktPtr;
int devType;
GENERIC **modelPtr;
IFuid uid;
{
    char *deckUniqueName();

    CHECK_CKT(cktPtr);
    deckFlushModels();
    CHECK_DEV(devType);
    curDefModel = ALLOC(struct model, 1);
    curDefModel->devPtr = deckInfo.devices[devType];
    curDefModel->magicChar = magicC[devType];
    curDefModel->name = deckUniqueName(&deckModelSpace, uid);
    curDefModel->modelUid = uid;
    *modelPtr = (GENERIC *) curDefModel;
    return(OK);
}

/*ARGSUSED*/
static int deckSetModelParm(cktPtr, modelPtr, parm, value, sel)
GENERIC *cktPtr;
GENERIC *modelPtr;
int parm;
IFvalue *value;
IFvalue *sel;
{
    IFparm *parmPtr = NIL(IFparm);
    int parmIndex;

    CHECK_CKT(cktPtr);
    CHECK_SAME((struct model *) modelPtr, curDefModel, "not current model");

    for (parmIndex = 0; parmIndex < curDefModel->devPtr->numModelParms; parmIndex++) {
	if (curDefModel->devPtr->modelParms[parmIndex].id == parm) {
	    parmPtr = &curDefModel->devPtr->modelParms[parmIndex];
	    break;
	}
    }
    if (parmPtr == NIL(IFparm)) {
	errRtn = "bad model parameter";
	return(E_PARMVAL);
    }

    switch (parmPtr->dataType & IF_VARTYPES) {
	case IF_FLAG:
	    if (value->iValue) {
		modTypeName = parmPtr->keyword;		/* XXX what a hack */
	    }
	    break;
	case IF_INTEGER:
	    (void) sprintf(bigString + strlen(bigString), "%s=%d ",
				    parmPtr->keyword, value->iValue);
	    break;
	case IF_REAL:
	    (void) sprintf(bigString + strlen(bigString), "%s=%g ",
				    parmPtr->keyword, value->rValue);
	    break;
	default:
	    errRtn = "unsupported model parameter type";
	    return(E_UNSUPP);	/* XXX */
    }

    return(OK);
}

static deckFlushModels()
{
    if (curDefModel) {
	if (modTypeName == NIL(char) && bigString[0] == '\0') {
	    /* this omits the model if there really was none */
	    curDefModel->name = NIL(char);
	} else {
	    if (modTypeName == NIL(char)) {
		modTypeName = "bad-model";
	    }
	    (void) fprintf(DECK_STREAM, ".MODEL %s %s(%s)\n",
			    curDefModel->name, modTypeName, bigString);
	}

	curDefModel = NIL(struct model);
	bigString[0] ='\0';
	modTypeName = NIL(char);
    }
}

	/* Instance-handling functions */




#define MAX_TERMS	50
#define MAX_PARMS	100

static char *instName = NIL(char);
static struct model *curInstModel = NIL(struct model);
static char *termTable[MAX_TERMS];
static int numParms = 0;
static struct {
    IFparm *devParmPtr;
    IFvalue val;
} parmList[MAX_PARMS];

/*ARGSUSED*/
static int deckNewInstance(cktPtr, modelPtr, instPtr, uid)
GENERIC *cktPtr;
GENERIC *modelPtr;
GENERIC **instPtr;
IFuid uid;
{
    static char newPrefix[] = "i";
    char *name;
    char *deckUniqueName();

    deckFlushInstances();
    CHECK_CKT(cktPtr);
    curInstModel = (struct model *) modelPtr;

    newPrefix[0] = curInstModel->magicChar;
    deckInstanceSpace.prefix = newPrefix;
    name = NAME_OF_UID(uid);
    deckInstanceSpace.doPrefix =
		    (name == NIL(char) || name[0] != curInstModel->magicChar);
    *instPtr = (GENERIC *) instName = deckUniqueName(&deckInstanceSpace, uid);

    return(OK);
}

/*ARGSUSED*/
static int deckSetInstanceParm(cktPtr, instPtr, parm, value, sel)
GENERIC *cktPtr;
GENERIC *instPtr;
int parm;
IFvalue *value;
IFvalue *sel;
{
    int parmIdx;

    CHECK_CKT(cktPtr);
    CHECK_SAME((char *) instPtr, instName, "not current instance");
    if (numParms >= MAX_PARMS) {
	errRtn = "overflowed instance parameter table";
	return(E_NOMEM);
    }

    for (parmIdx = 0; parmIdx < curInstModel->devPtr->numInstanceParms; parmIdx++) {
	if (curInstModel->devPtr->instanceParms[parmIdx].id == parm) {
	    parmList[numParms].devParmPtr =
				&curInstModel->devPtr->instanceParms[parmIdx];
	    parmList[numParms++].val = *value;
	    return(OK);
	}
    }

    errRtn = "bad instance parameter";
    return(E_PARMVAL);
}

static int deckBindNode(cktPtr, instPtr, termIndex, nodePtr)
GENERIC *cktPtr;
GENERIC *instPtr;
int termIndex;
IFnode nodePtr;
{
    CHECK_CKT(cktPtr);
    CHECK_SAME((char *) instPtr, instName, "not current circuit");
    if (termIndex < 1 || termIndex > ABS(curInstModel->devPtr->terms)) {
	/* XXX should handle variable number of terminals */
	errRtn = "illegal terminal index";
	return(E_NOTERM);
    }
    if (termIndex >= MAX_TERMS) {
	errRtn = "overflowed instance terminal table";
	return(E_NOMEM);
    }
    termTable[termIndex] = (char *) nodePtr;
    return(OK);
}

static deckFlushInstances()
{
    static int ncCounter = 0;
    int i;
    void printParms();

    deckFlushModels();		/* the model should be defined, and we */
				/* need to have its name wiped if it */
				/* wasn't all there */
    if (instName) {
	(void) fprintf(DECK_STREAM, "%s ", instName);
	for (i = 1; i <= ABS(curInstModel->devPtr->terms); i++) {
	    if (termTable[i] == NIL(char)) {
		(void) fprintf(DECK_STREAM, "NC%d ", ++ncCounter);
	    } else {
		(void) fprintf(DECK_STREAM, "%s ", termTable[i]);
	    }
	    termTable[i] = NIL(char);
	}
	if (curInstModel->name) {
	    (void) fprintf(DECK_STREAM, "%s ", curInstModel->name);
	}
	printParms();
	(void) fprintf(DECK_STREAM, "\n");
	numParms = 0;
	curInstModel = NIL(struct model);
	instName = NIL(char);
    }
}

/*
*	These parameters are order-dependent, since they are printed
*   without keywords.
*/
static char *prefixParms[] = {
	"control",	"gain",				/* controlled sources */
	"inductor1",	"inductor2",	"coefficient",	/* coupled inductors */
	"resistance",	"capacitance",	"inductance",	/* simple components */
	"area",		"off",				/* junction devices */
	"dc",						/* sources */
	"acphase",	"acmag",			/* sources (special) */
	0
};

static void printParms()
{
    char **preParmPtr;
    int i, j;
    void printParmVal();

    for (preParmPtr = prefixParms; *preParmPtr; preParmPtr++) {
	for (i = 0; i < numParms; i++) {
	    if (parmList[i].devParmPtr &&
		    strcmp(parmList[i].devParmPtr->keyword, *preParmPtr) == 0) {
		if (strcmp(*preParmPtr, "acphase") == 0) {
		    /* blech -- look for magnitute, and print both */
		    (void) fprintf(DECK_STREAM, "ac ");
		    for (j = 0; j < numParms; j++) {
			if (parmList[j].devParmPtr &&
				    strcmp(parmList[j].devParmPtr->keyword,
					    "acmag") == 0) {
			    printParmVal(parmList[j].devParmPtr,
							    &parmList[j].val);
			    parmList[j].devParmPtr = NIL(IFparm);
			    break;
			}
		    }
		    if (j == numParms) (void) fprintf(DECK_STREAM, "1 ");
		} else if (strcmp(*preParmPtr, "acmag") == 0) {
		    /* double blech -- there is no phase, so just print mag */
		    (void) fprintf(DECK_STREAM, "ac ");
		}

		printParmVal(parmList[i].devParmPtr, &parmList[i].val);
		parmList[i].devParmPtr = NIL(IFparm);
		break;
	    }
	}
    }

    for (i = 0; i < numParms; i++) {
	if (parmList[i].devParmPtr) {
	    (void) fprintf(DECK_STREAM, "%s=", parmList[i].devParmPtr->keyword);
	    printParmVal(parmList[i].devParmPtr, &parmList[i].val);
	}
    }
}

static void printParmVal(parmPtr, valPtr)
IFparm *parmPtr;
IFvalue *valPtr;
{
    int i;

    switch (parmPtr->dataType & IF_VARTYPES) {
	case IF_FLAG:
	    if (valPtr->iValue) {
		(void) fprintf(DECK_STREAM, "%s ", parmPtr->keyword);
	    }
	    break;
	case IF_INTEGER:
	    (void) fprintf(DECK_STREAM, "%d ", valPtr->iValue);
	    break;
	case IF_REAL:
	    (void) fprintf(DECK_STREAM, "%g ", valPtr->rValue);
	    break;
	case IF_REALVEC:
	    (void) fprintf(DECK_STREAM, "(");
	    for (i = 0; i < valPtr->v.numValue; i++) {
		(void) fprintf(DECK_STREAM, "%g ", valPtr->v.vec.rVec[i]);
	    }
	    (void) fprintf(DECK_STREAM, ") ");
	    free((char *) valPtr->v.vec.rVec);
	    break;
	case IF_NODE:
	    /* XXX -- this assumes an IFuid is a name */
	    (void) fprintf(DECK_STREAM, "%s ", (char *) valPtr->nValue);
	    break;
	case IF_STRING:
	    (void) fprintf(DECK_STREAM, "%s ", valPtr->sValue);
	    break;
	case IF_INSTANCE:
	    /* XXX -- this assumes an IFuid is a name */
	    (void) fprintf(DECK_STREAM, "%s ", (char *) valPtr->uValue);
	    break;
	default:
	    (void) fprintf(DECK_STREAM, "***bad-value-type*** ");    /* oops */
    }
}

	/* Namespace management functions */




static void deckInitNameSpace(nmsp, prefix, doPrefix)
struct nameSpace *nmsp;
char *prefix;
int doPrefix;
{
    nmsp->nameTable = st_init_table(strcmp, st_strhash);
    nmsp->prefix = prefix;
    nmsp->counter = 0;
    nmsp->doPrefix = doPrefix;
}

static void deckFreeNameSpace(nmsp)
struct nameSpace *nmsp;
{
    st_generator *gen;
    char *name;

    gen = st_init_gen(nmsp->nameTable);
    while (st_gen(gen, &name, NIL(char *))) {
	FREE(name);
    }
    st_free_gen(gen);

    st_free_table(nmsp->nameTable);
}

static char goodFirstChars[] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
static char goodChars[] =
    "_ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789.#";

static char *deckUniqueName(nmsp, uid)
struct nameSpace *nmsp;
IFuid uid;
{
    char *cp1, *cp2;
    int demerits = 0;
    char testName[1000];

    cp1 = NAME_OF_UID(uid);
    if (cp1 == NIL(char) || cp1[0] == '\0') goto chooseName;

    cp2 = testName;
    if (nmsp->doPrefix) {
	(void) strcpy(testName, nmsp->prefix);
	cp2 += strlen(testName);
    } else {
	if (strchr(goodFirstChars, *cp1) == NIL(char)) {
	    *cp2++ = goodFirstChars[0];
	    demerits++;
	}
    }

    while (*cp1 != '\0') {
	if (strchr(goodChars, *cp1)) {
	    *cp2++ = *cp1++;
	    demerits--;
	} else {
	    *cp2++ = goodChars[0];
	    cp1++;
	    demerits++;
	}
    }
    *cp2 = '\0';

    if (demerits > 0) goto chooseName;

    while (st_is_member(nmsp->nameTable, testName)) {
chooseName:
	(void) sprintf(testName, "%s%d", nmsp->prefix, ++(nmsp->counter));
    }

    cp2 = util_strsav(testName);
    st_add_direct(nmsp->nameTable, cp2, (char *) uid);

    return(cp2);
}

static IFsimulator deckInfo = {
    "SPICE3",		/* my name (must be "SPICE3" since this name is
			 * the name of the model bag in oct) */
    "oct to SPICE3 input deck converter",  /* more about me */
    "1",		/* my version */

    deckNewCircuit,	/* newCircuit function */
    deckDeleteCircuit,	/* deleteCircuit function */

    deckNewNode,	/* newNode function */
    deckSetGroundNode,	/* groundNode function */
    deckBindNode,	/* bindNode function */
    0,			/* findNode function */
    0,			/* instToNode function */
    0,			/* setNodeParm function */
    0,			/* askNodeQuest function */
    0,			/* deleteNode function */

    deckNewInstance,	/* newInstance function */
    deckSetInstanceParm,/* setInstanceParm function */
    0,			/* askInstanceQuest function */
    0,			/* findInstance funciton */
    0,			/* deleteInstance function */

    deckNewModel,	/* newModel function */
    deckSetModelParm,	/* setModelParm function */
    0,			/* askModelQuest function */
    0,			/* findModel function */
    0,			/* deleteModel function */

    deckNewTask,	/* newTask function */
    deckNewAnalysis,	/* newAnalysis function */
    deckSetAnalysisParm,/* setAnalysisParm function */
    0,			/* askAnalysisQeust function */
    0,			/* findAnalysis function */
    0,			/* findTask function */
    deckDeleteTask,	/* deleteTask function */

    deckDoAnalyses,	/* doAnalyses function */

    0, NIL(IFdevice*),
    0, NIL(IFanalysis *),
    0, NIL(IFparm),
    0, NIL(char *),
};

IFsimulator *deckInit(spicePtr)
IFsimulator *spicePtr;
{

    deckInfo.numDevices = spicePtr->numDevices;
    deckInfo.devices = spicePtr->devices;

    deckInfo.numAnalyses = spicePtr->numAnalyses;
    deckInfo.analyses = spicePtr->analyses;

    deckInfo.numNodeParms = spicePtr->numNodeParms;
    deckInfo.nodeParms = spicePtr->nodeParms;

    deckInfo.numSpecSigs = spicePtr->numSpecSigs;
    deckInfo.specSigs = spicePtr->specSigs;

    return(&deckInfo);
}
