#include "copyright.h"
#include "port.h"
#include <sys/wait.h>
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "octspice.h"

void ofeGraph(analPtr, xFlags, xPtr, yFlags, yList, fileName)
analDesc *analPtr;
int xFlags;
waveStruct *xPtr;
int yFlags;
lsList yList;
char *fileName;
{
    FILE *fp;
    static int types[] = { PFL_DB, PFL_MAG, PFL_REAL, PFL_IMAG, PFL_PHASE };
    int i;
    int xType;
    lsGeneric dummy;
    waveStruct *yPtr;
    char *yLegend;
    FILE *startGraph();
    void oneCurve();
    char *makeLegend();

    /* trim X axis to exactly one aspect */
    xType = 0;
    for (i = 0; i < sizeof(types)/sizeof(types[0]); i++) {
	if (xFlags & types[i]) {
	    xType = types[i];
	    break;
	}
    }
    if (xType == 0) xType = PFL_REAL;
    xFlags = (xFlags & ~PFL_ASPECT) | xType;

    fp = startGraph(analPtr, xFlags, xPtr, yFlags, fileName);
    if (fp == NULL) return;

    /* plot each aspect of each value in yList */
    while (lsDelBegin(yList, &dummy) == LS_OK) {
	yPtr = (waveStruct *) dummy;
	for (i = 0; i < sizeof(types)/sizeof(types[0]); i++) {
	    if (yFlags & types[i]) {
		yLegend = makeLegend((yFlags&~PFL_ASPECT) | types[i], yPtr);
		(void) fprintf(fp, "\"%s\n", yLegend);
		FREE(yLegend);
		oneCurve(fp, xFlags, xPtr,
			    (yFlags&~PFL_ASPECT) | types[i], yPtr);
	    }
	}
    }

    (void) fclose(fp);
}

static void oneCurve(fp, xFlags, xPtr, yFlags, yPtr)
FILE *fp;
int xFlags;
waveStruct *xPtr;
int yFlags;
waveStruct *yPtr;
{
    int i;
    double x = 0, y = 0;
    int calcCoord();

    ASSERT(daLen(xPtr->val) == daLen(yPtr->val));
    for (i = 0; i < daLen(xPtr->val); i++) {
	if (calcCoord(i, xFlags, xPtr, &x) && calcCoord(i, yFlags, yPtr, &y)) {
	    (void) fprintf(fp, "%g %g\n", x, y);
	} else {
	    ofeMessage(WARNING, "can't take log of zero:  discarding point\n");
	}
    }
    (void) fprintf(fp, "\n");
}

static FILE *startGraph(analPtr, xFlags, xPtr, yFlags, fileName)
analDesc *analPtr;
int xFlags;
waveStruct *xPtr;
int yFlags;
char *fileName;		/* name of plot file, or NIL(char) to fork xgraph */
{
    char *argv[20];
    int i;
    char *title;
    char *xLegend;
    FILE *fp;
    FILE *openPipe(), *openFile();
    char *makeTitle(), *makeLegend();

    i = 0;
    argv[i++] = "xgraph";
    argv[i++] = "-t";
    argv[i++] = title = makeTitle(analPtr);

    if (xFlags & PFL_LOG) argv[i++] = "-lnx";
    if (yFlags & PFL_LOG) argv[i++] = "-lny";

    argv[i++] = "-x";
    argv[i++] = xLegend = makeLegend(xFlags, xPtr);

    argv[i++] = ofeDisplayName;
    argv[i] = NIL(char);

    if (fileName == NIL(char)) {
	fp = openPipe(argv);
    } else {
	fp = openFile(argv, fileName);
    }
    FREE(title);
    FREE(xLegend);
    return(fp);
}

static char *makeTitle(analPtr)
analDesc *analPtr;
{
    char *analName = analPtr->name;
    char *cktName = analPtr->cktPtr->name;
    char *title;

    title = ALLOC(char, (analName ? strlen(analName) : 0) +
			(cktName ? strlen(cktName) : 0) + 20);
    *title = '\0';
    if (cktName) (void) sprintf(title, "``%s''", cktName);
    if (analName) (void) sprintf(title + strlen(title), " (%s)", analName);

    return(title);
}

static char *makeLegend(flags, ptr)
int flags;
waveStruct *ptr;
{
    char *node;
    char *units = "";			/* XXX */
    char *fmtString;
    char *str;

    node = (ptr->netPtr && ptr->netPtr->name) ? ptr->netPtr->name : "";

    str = ALLOC(char, strlen(node) + 20);
    if (flags & PFL_REAL) {
	if (ptr->valType == IF_COMPLEX) {
	    fmtString = "Re(%s)%s";
	} else {
	    fmtString = "%s%s";
	}
    } else if (flags & PFL_IMAG) {
	fmtString = "Im(%s)%s";
    } else if (flags & PFL_DB) {
	fmtString = "dB(%s)";
    } else if (flags & PFL_MAG) {
	fmtString = "| %s |%s";
    } else if (flags & PFL_PHASE) {
	fmtString = "phase(%s) deg";
    } else {
	ASSERT(0);
    }

    (void) sprintf(str, fmtString, node, units);
    return(str);
}

static FILE *openPipe(argv)
char *argv[];
{
    FILE *fp;
    int fildes[2];
    int failed = 0;
    union wait status;
    extern errno;
    char *errString();

    if (pipe(fildes) < 0) {
	ofeMessage(SEVERE, "couldn't open pipe: %s\n", errString(errno));
	return(NULL);
    }

    switch (vfork()) {
	case 0:		/* child */
	    (void) dup2(fildes[0], fileno(stdin));
	    (void) dup2(fileno(stderr), fileno(stdout));
	    (void) close(fildes[0]);
	    (void) close(fildes[1]);

	    /* fork and exit to disconnect completely from the parent */
	    switch (vfork()) {
		case 0:		/* child */
		    (void) execvp(argv[0], argv);
		    failed = errno;
		    (void) _exit(1);
		    /*NOTREACHED*/
		case -1:	/* error */
		    (void) _exit(1);
		    /*NOTREACHED*/
		default:	/* parent */
		    (void) _exit(0);
		    /*NOTREACHED*/
	    }
	case -1:	/* error */
	    ofeMessage(SEVERE, "couldn't fork: %s\n", errString(errno));
	    return(NULL);
	default:	/* parent */
	    if (wait(&status) < 0 || status.w_status != 0) {
		ofeMessage(SEVERE, "couldn't fork\n");
		return(NULL);
	    }
	    if (failed) {
		ofeMessage(SEVERE, "couldn't exec %s: %s\n",
					argv[0], errString(failed));
		return(NULL);
	    }
	    (void) close(fildes[0]);
	    fp = fdopen(fildes[1], "w");
	    if (fp == NULL) {
		ofeMessage(SEVERE, "couldn't open pipe: %s\n",errString(errno));
	    }
	    return(fp);
    }
}

static FILE *openFile(argv, fileName)
char *argv[];
char *fileName;
{
    FILE *fp;
    char **argPtr;
    char *cPtr;
    char *errString();

    if ((fp = fopen(fileName, "w")) == NULL) {
	ofeMessage(SEVERE, "couldn't open file %s: %s\n",
				fileName, errString(errno));
	return(NULL);
    }

    (void) fputs("#! /bin/sh\nexec ", fp);
    for (argPtr = argv; *argPtr; argPtr++) {
	putc('\'', fp);
	for (cPtr = *argPtr; *cPtr != '\0'; cPtr++) {
	    if (*cPtr == '\'') {
		(void) fputs("'\\''", fp);
	    } else {
		putc(*cPtr, fp);
	    }
	}
	(void) fputs("' ", fp);
    }
    (void) fputs("<< 'EOF'\n", fp);

    return(fp);
}

static char *errString(err)
int err;
{
    extern int sys_nerr, errno;
    extern char *sys_errlist[];

    if (err < 0 || err >= sys_nerr) {
	return("unknown error");
    } else {
	return(sys_errlist[err]);
    }
}

static int calcCoord(i, flags, wavePtr, resultPtr)
int i;
int flags;
waveStruct *wavePtr;
double *resultPtr;
{
    IFcomplex val;
    double phase;
    double log10(), hypot(), atan2();

    if (wavePtr->valType == IF_COMPLEX) {
	val = *daGet(IFcomplex, wavePtr->val, i);
    } else {
	val.real = *daGet(double, wavePtr->val, i);
	val.imag = 0;
    }

    if ((flags & (PFL_LOG|PFL_DB)) && val.real == 0 && val.imag == 0) {
	return(0);
    }

    switch (flags & PFL_ASPECT) {
	case PFL_REAL:
	    *resultPtr = val.real;
	    break;
	case PFL_IMAG:
	    *resultPtr = val.imag;
	    break;
	case PFL_DB:
	    *resultPtr = 20 * log10(hypot(val.real, val.imag));
	    break;
	case PFL_MAG:
	    *resultPtr = hypot(val.real, val.imag);
	    break;
	case PFL_PHASE:
	    phase = (180/3.14159265358979323) * atan2(val.imag, val.real);
	    while (phase > *resultPtr+180) phase -= 360;
	    while (phase < *resultPtr-180) phase += 360;
	    *resultPtr = phase;
	    break;
	default:
	    ASSERT(0);
	    /*NOTREACHED*/
    }
    return(1);
}
