#include "copyright.h"
#include "st.h"
#include "list.h"
#include "da.h"
#include "IFsim.h"
#include "IFerrmsgs.h"

#ifdef RPC
#include "rpc.h"
#define fopen	RPCfopen
#endif /*RPC*/

#define ASSERT(exp)		\
    if ((exp) == 0) {							\
	ofeMessage(FATAL, "assertion failed at %s:%d\n", __FILE__, __LINE__);\
    }

#define IF_CHECK(fn,args,sev,action)	\
    if (ofeSimPtr->fn) {						\
	if (!ofeCheckIF( (*ofeSimPtr->fn)args, __FILE__, __LINE__, sev )) { \
	    action;							\
	}								\
    } else {								\
	ofeMessage(FATAL, "unimplemented simulator function at %s:%d\n",\
			__FILE__, __LINE__);				\
    }
#define IF_ASSERT(fn,args)	\
    IF_CHECK(fn, args, FATAL, exit(-1))
#define OFE_NEW_UID(old, new, name, type)	\
    if (!ofeCheckIF(ofeNewUid(NIL(GENERIC), (old), (new), (name),	\
			      (type), NIL(GENERIC)),			\
		    __FILE__, __LINE__, FATAL)) {			\
	exit(-1);							\
    }

#define LS_ASSERT(exp)		\
    if ((exp) != LS_OK) {						\
	ofeMessage(FATAL, "unexpected ls package error at %s:%d\n",	\
			__FILE__, __LINE__);				\
    }

#ifdef XDM_TRUE
#define XDM_ASSERT(exp)		\
    if ((exp) != XDM_TRUE) {						\
	ofeMessage(FATAL, "unexpected XDM error (%s) at %s:%d\n",	\
			XDMError(), __FILE__, __LINE__);		\
    }
#endif /*XDM_TRUE*/

typedef enum _messageType { INFO, WARNING, SEVERE, FATAL } messageType;

#define UID_FIELDS	\
    int uidType;		/* ANALYSIS,TASK,INSTANCE,MODEL,SIGNAL,OTHER */\
    struct _uidDesc *fChain;	/* doubly-linked circular list of */	\
    struct _uidDesc *bChain;	/*     all uid objects in facet */	\
    char *name;			/* printable name, or NIL */		\
    GENERIC *simEquiv;		/* simulator's id for this object, or NIL */\
    struct _cktDesc *cktPtr;	/* pointer to containing circuit */	\

typedef struct _uidDesc {
    UID_FIELDS
} uidDesc;
#define INIT_UID(uid, ckt)	{\
    (uid)->uidType = UID_OTHER;						\
    (uid)->fChain = (uidDesc *) (ckt);					\
    (uid)->bChain = (ckt)->bChain;					\
    (ckt)->bChain->fChain = (uidDesc *) (uid);				\
    (ckt)->bChain = (uidDesc *) (uid);					\
    (uid)->name = NIL(char);						\
    (uid)->simEquiv = NIL(GENERIC);					\
    (uid)->cktPtr = (ckt);						\
}

typedef struct _cktDesc {
    UID_FIELDS
    struct _taskDesc *taskPtr;	/* current task on this circuit */
    struct _cellDesc *rootPtr;	/* root cell of the hierarchy */
    st_table *modelTable;	/* master octId => modelDesc */
} cktDesc;
#define INIT_CKT(ckt)		{\
    (ckt)->bChain = (uidDesc *) (ckt);					\
    INIT_UID(ckt, ckt);							\
    (ckt)->taskPtr = NIL(taskDesc);					\
    (ckt)->rootPtr = NIL(cellDesc);					\
    (ckt)->modelTable = st_init_table(st_numcmp, st_numhash);		\
}

typedef struct _taskDesc {
    UID_FIELDS
    struct _analDesc *analPtr;	/* the (one) analysis in this task */
} taskDesc;
#define INIT_TASK(tsk, ckt)	{\
    INIT_UID(tsk, ckt);							\
    (tsk)->uidType = UID_TASK;						\
    (tsk)->analPtr = NIL(analDesc);					\
}

typedef struct _analDesc {
    UID_FIELDS
    struct _waveStruct *indepWave;/* independent variable waveform */
    int nDepWaves;		/* number of dependent variables */
    struct _waveStruct **depWaves;/* dependent variable waveforms */
    int depValType;		/* type of all dependent variables */
    st_table *waveTable;	/* net octId => waveStruct (not indep var) */
} analDesc;
#define INIT_ANAL(a, ckt)	{\
    INIT_UID(a, ckt);							\
    (a)->uidType = UID_ANALYSIS;					\
    (a)->indepWave = NIL(waveStruct);					\
    (a)->nDepWaves = 0;							\
    (a)->depWaves = NIL(waveStruct *);					\
    (a)->depValType = 0;						\
    (a)->waveTable = st_init_table(st_numcmp, st_numhash);		\
}

typedef struct _cellDesc {
    UID_FIELDS
    struct _cellDesc *parentPtr;/* the instantiating cell */
    octObject master;		/* the oct cell this structure describes */
    st_table *termTable;	/* term xid => netDesc */
    enum _cellClass { MODULE, LEAF } cellClass;
	/* for MODULE cells */
    st_table *netTable;		/* net octId => netDesc */
    st_table *childTable;	/* instance octId => cellDesc */
	/* LEAF cells */
    struct _modelDesc *modelPtr;/* the model that describes this instance */
} cellDesc;
#define INIT_CELL(c, ckt)	{\
    INIT_UID(c, ckt);							\
    (c)->uidType = UID_INSTANCE;					\
    (c)->parentPtr = NIL(cellDesc);					\
    (c)->master.objectId = oct_null_id;					\
    (c)->termTable = st_init_table(st_numcmp, st_numhash);		\
    (c)->cellClass = LEAF;						\
    (c)->netTable = st_init_table(st_numcmp, st_numhash);		\
    (c)->childTable = st_init_table(st_numcmp, st_numhash);		\
    (c)->modelPtr = NIL(modelDesc);					\
}

typedef struct _modelDesc {
    UID_FIELDS
    IFdevice *simDevPtr;/* the simulator device description structure */
} modelDesc;
#define INIT_MODEL(m, ckt)	{\
    INIT_UID(m, ckt);							\
    (m)->uidType = UID_MODEL;						\
    (m)->simDevPtr = NIL(IFdevice);					\
}

typedef struct _netDesc {
    UID_FIELDS
    struct _cellDesc *cellPtr;	/* the cell containing this net */
    octId netId;		/* octId of net */
} netDesc;
#define INIT_NET(net, ckt)	{\
    INIT_UID(net, ckt);							\
    (net)->uidType = UID_SIGNAL;					\
    (net)->cellPtr = NIL(cellDesc);					\
    (net)->netId = oct_null_id;						\
}

typedef struct _waveStruct {
    struct _netDesc *netPtr;
    int valType;
    dary val;
} waveStruct;

/*	plotting flags		*/

#define PFL_REAL	(1 << 0)
#define PFL_IMAG	(1 << 1)
#define PFL_DB		(1 << 2)
#define PFL_MAG		(1 << 3)
#define PFL_PHASE	(1 << 4)
#define PFL_LOG		(1 << 8)
#define PFL_ASPECT	(PFL_LOG - 1)



extern IFsimulator *ofeSpicePtr;/* SPICE's simulator functions */
extern IFsimulator *ofeDeckPtr;	/* deck-producing pseudo-simulator functions */
extern IFsimulator *ofeSimPtr;	/* current (pseudo-)simulator functions */
extern IFfrontEnd ofeFEstruct;	/* all my functions for the simulator */
extern char *progName;
extern char *ofeDisplayName;

extern int SPIinit();
extern IFsimulator *deckInit();

extern cktDesc *ofeCircuit();
extern void ofeDoTask(), ofeGraph(), ofeFreeCircuit();
extern void ofeModule(), ofeLeaf(), ofeFreeCell();

extern netDesc *ofeFindNet();
extern void ofeOutInit(), ofeInitSimulator(), ofeGenName(), ofeMessage();
extern int ofeNewUid();
extern void ofeFreeUid();
extern int ofeCheckIF();
extern void ofeNormLeafHandler();

extern void ofeGetModel(), ofeGetInstance();
extern void ofeSetInstParms(), ofeSetOneInstParm(), ofeWireLeaf();

/* commands */
extern dary ofeInitSimulations();
extern void ofeSetInstProps(), ofeFixLabels();
extern void ofeChooseGround(), ofeAnalysis(), ofeDeck();
extern void ofeNodePlot(), ofePlot();


	/* Dialog stuff */

struct flagStruct {
    char *name;
    int flag;
    int weight;
    int fieldId;
    IFparm *parmPtr;
};
#define INIT_FLAG(flg)	{						\
    (flg)->name = NIL(char);						\
    (flg)->flag = (flg)->weight = 0;					\
    (flg)->fieldId = -1;						\
    (flg)->parmPtr = NIL(IFparm);					\
}

struct scalarStruct {
    char *name;
    double value, dfltValue;
    int fieldId;
    IFparm *parmPtr;
};
#define INIT_SCALAR(val)	{					\
    (val)->name = NIL(char);						\
    (val)->value = (val)->dfltValue = 0;				\
    (val)->fieldId = -1;						\
    (val)->parmPtr = NIL(IFparm);					\
}

struct arrayStruct {
    char *name;
    dary values;
    int fieldId;
    IFparm *parmPtr;
};
#define INIT_ARRAY(ary)	{						\
    (ary)->name = NIL(char);						\
    (ary)->values = daAlloc(double, 0);					\
    (ary)->fieldId = -1;						\
    (ary)->parmPtr = NIL(IFparm);					\
}

struct dialog {
    int analIndex;
    IFanalysis *simAnalPtr;
    /*Window dialogWin;*/
    char *title;

    dary flagColTitles;
    dary flags;
    dary scalars;
    dary arrays;
    int OKId, abortId;
};
#define INIT_DIALOG(dlg)	{					\
    (dlg)->analIndex = 0;						\
    (dlg)->simAnalPtr = NIL(IFanalysis);				\
    /*(dlg)->dialogWin = 0;*/						\
    (dlg)->flagColTitles = daAlloc(char *, 0);				\
    (dlg)->flags = daAlloc(struct flagStruct, 0);			\
    (dlg)->scalars = daAlloc(struct scalarStruct, 0);			\
    (dlg)->arrays = daAlloc(struct arrayStruct, 0);			\
    (dlg)->OKId = (dlg)->abortId = -1;					\
}

#define BAD_FMT(msg)	{ VEMMSG(format); vemPrompt(); return(RPC_OK); }

#define COL_WIDTH	13
#define LABEL_BAG_NAME	"Silly-bitmask-dodge"	/* oct has no policy :-( */

extern int ofeDisplayDialog();
