#include "copyright.h"
#include "port.h"
#include "utility.h"
#include <varargs.h>
#include "oct.h"
#include "oh.h"
#include "options.h"
#include "octspice.h"

optionStruct optionList[] = {
    { "g",	"gnd_name",	"gnd_name is ground net or formal terminal" },
    { OPT_CONT,	"",		" (default is net with NETTYPE=GROUND)"	},
    { OPT_RARG,	"cell:view",	"cell to be simulated"			},
    { 0,	0,		0					}
};
char *progName;

struct commandStruct {
    char *command;
    char *args;
    char *summary;
    int (*function)();
};
static dary commandTable;	/* contains `struct commandStruct's */

#define WHITE_SPACE	" \t\n\r\f"

main(argc, argv)
int argc;
char *argv[];
{
    int option;
    dary analDialogs;
    char *groundName = NIL(char);
    octObject facet;
    extern char *optarg;
    extern int optind;
    void interactive();
    int ground();

    while ((option = optGetOption(argc, argv)) != EOF) {
	switch (option) {
	    case 'g':		/* ground terminal name */
		groundName = optarg;
		break;
	}
    }
    if (optind != argc-1) optUsage();
    progName = optProgName;

    octBegin();

    analDialogs = ofeInitSimulations();

    ohUnpackDefaults(&facet, "r", "::contents");
    if (ohUnpackFacetName(&facet, argv[optind++]) != OCT_OK) {
	optUsage();
    }

    if (octOpenFacet(&facet) < OCT_OK) {
	ofeMessage(SEVERE, "couldn't open %s (%s)\n",
			    ohFormatName(&facet), octErrorString());
	exit(1);
    }
    if (groundName) (void) ground(&facet, groundName, NIL(struct dialog));

    interactive(&facet, analDialogs);

    octEnd();

    exit(0);
}

static void interactive(facetPtr, analDialogs)
octObject *facetPtr;
dary analDialogs;
{
    char lineBuf[BUFSIZ];
    char *name, *args;
    int i;
    struct commandStruct *cmdPtr;
    struct dialog *dlgPtr;
    void buildCommandTable();
    void badCommand(), oneCmdHelp();

    buildCommandTable(analDialogs);
    while ((void) printf(":%s: ", progName),
		    fgets(lineBuf, sizeof(lineBuf)/sizeof(char), stdin)) {
	/* separate the command name and args */
	for (name = lineBuf; *name != '\0' && isspace(*name); name++) ;
	if (*name == '\0') continue;
	for (args = name; *args != '\0' && ! isspace(*args); args++) ;
	if (*args != '\0') {
	    *args++ = '\0';
	    for ( ; *args != '\0' && isspace(*args); args++) ;
	}

	for (i = 0; i < daLen(commandTable); i++) {
	    cmdPtr = daGet(struct commandStruct, commandTable, i);
	    if (strcmp(name, cmdPtr->command) == 0) {
		/* the analysis commands come first */
		if (i < daLen(analDialogs)) {
		    dlgPtr = daGet(struct dialog, analDialogs, i);
		} else{
		    dlgPtr = NIL(struct dialog);
		}
		if ( ! (*cmdPtr->function)(facetPtr, args, dlgPtr)) {
		    oneCmdHelp(cmdPtr, "usage: ");
		}
		break;
	    }
	}
	if (i == daLen(commandTable)) {
	    badCommand(name);
	}
    }
}

/*ARGSUSED*/
static int iProps(facetPtr, args, dlgPtr)
octObject *facetPtr;
char *args;
struct dialog *dlgPtr;
{
    char name[BUFSIZ], dummy[2];
    octObject inst;

    if (sscanf(args, "%s %1s", name, dummy) != 1) return(0);

    if (ohGetByInstName(facetPtr, &inst, name) == OCT_OK) {
	ofeSetInstProps(&inst);
    } else {
	ofeMessage(SEVERE, "can't find instance named `%s'\n", name);
    }

    return(1);
}

#ifdef LABEL_HACK
/*ARGSUSED*/
static int labels(facetPtr, args, dlgPtr)
octObject *facetPtr;
char *args;
struct dialog *dlgPtr;
{
    ofeFixLabels(facetPtr);
    return(1);
}
#endif /*LABEL_HACK*/

/*ARGSUSED*/
static int wFacet(facetPtr, args, dlgPtr)
octObject *facetPtr;
char *args;
struct dialog *dlgPtr;
{
    char name[BUFSIZ], dummy[2];
    octObject newFacet;

    switch (sscanf(args, "%s %1s", name, dummy)) {
	case 0:		/* write back to same facet */
	    facetPtr->contents.facet.mode = "a";
	    if (octOpenFacet(facetPtr) < OCT_OK) {
		ofeMessage(SEVERE, "couldn't reopen facet for append:  %s\n",
			    octErrorString());
	    }
	    OH_ASSERT(octCloseFacet(facetPtr));
	    return(1);
	case 1:		/* write to other facet */
	    newFacet = *facetPtr;
	    newFacet.contents.facet.mode = "a";
	    OH_ASSERT(ohUnpackFacetName(&newFacet, name));
	    if (octWriteFacet(&newFacet, facetPtr) < OCT_OK) {
		ofeMessage(SEVERE, "couldn't write to facet:  %s\n",
			    octErrorString());
	    }
	    return(1);
	default:	/* bozo */
	    return(0);
    }
}

/*ARGSUSED*/
static int ground(facetPtr, args, dlgPtr)
octObject *facetPtr;
char *args;
struct dialog *dlgPtr;
{
    char name[BUFSIZ], dummy[2];
    octObject netOrTerm;

    if (sscanf(args, "%s %1s", name, dummy) != 1) return(0);

    if (ohGetByNetName(facetPtr, &netOrTerm, name) == OCT_OK ||
		    ohGetByTermName(facetPtr, &netOrTerm, name) == OCT_OK) {
	ofeChooseGround(&netOrTerm);
    } else {
	ofeMessage(SEVERE, "can't find net or terminal named `%s'\n", name);
    }

    return(1);
}

static int doSavePlot(facetPtr, fileName, netName)
octObject *facetPtr;
char *fileName;
char *netName;
{
    octObject tempBag;
    octObject net;
    octGenerator gen;

    if (strcmp(netName, "*") == 0) {
	octInitGenContents(facetPtr, OCT_NET_MASK, &gen);
	ofeNodePlot(&gen, fileName);
    } else {
	OH_ASSERT(ohCreateBag(facetPtr, &tempBag, "### plot bag ###"));
	octMarkTemporary(&tempBag);

	do {
	    if (ohGetByNetName(facetPtr, &net, netName) == OCT_OK) {
		OH_ASSERT(octAttach(&tempBag, &net));
	    } else {
		ofeMessage(SEVERE, "can't find the net named %s\n", netName);
	    }
	} while (netName = strtok(NIL(char), WHITE_SPACE));

	octInitGenContents(&tempBag, OCT_NET_MASK, &gen);
	ofeNodePlot(&gen, fileName);

	OH_ASSERT(octDelete(&tempBag));
    }

    return(1);
}

/*ARGSUSED*/
static int plot(facetPtr, args, dlgPtr)
octObject *facetPtr;
char *args;
struct dialog *dlgPtr;
{
    char *firstNetName;

    if (firstNetName = strtok(args, WHITE_SPACE)) {
	return(doSavePlot(facetPtr, NIL(char), firstNetName));
    } else {
	return(0);
    }
}

/*ARGSUSED*/
static int save(facetPtr, args, dlgPtr)
octObject *facetPtr;
char *args;
struct dialog *dlgPtr;
{
    char *fileName;
    char *firstNetName;

    if ((fileName = strtok(args, WHITE_SPACE)) &&
		(firstNetName = strtok(NIL(char), WHITE_SPACE))) {
	return(doSavePlot(facetPtr, fileName, firstNetName));
    } else {
	return(0);
    }
}

/*ARGSUSED*/
static int deck(facetPtr, args, dlgPtr)
octObject *facetPtr;
char *args;
struct dialog *dlgPtr;
{
    char fileName[BUFSIZ], dummy[2];

    if (sscanf(args, "%s %1s", fileName, dummy) != 1) return(0);

    ofeDeck(facetPtr, fileName);

    return(1);
}

/*ARGSUSED*/
static int analyze(facetPtr, args, dlgPtr)
octObject *facetPtr;
char *args;
struct dialog *dlgPtr;
{
    ofeAnalysis(facetPtr, dlgPtr);
    return(1);
}

/*ARGSUSED*/
static int quit(facetPtr, args, dlgPtr)
octObject *facetPtr;
char *args;
struct dialog *dlgPtr;
{
    exit(0);
    /*NOTREACHED*/
}

/*ARGSUSED*/
static int help(facetPtr, args, dlgPtr)
octObject *facetPtr;
char *args;
struct dialog *dlgPtr;
{
    char command[25], dummy[2];
    int i;
    void oneCmdHelp(), badCommand();

    switch (sscanf(args, "%20s %1s", command, dummy)) {
	case -1:
	case 0:
	    for (i = 0; i < daLen(commandTable); i++) {
		oneCmdHelp(daGet(struct commandStruct, commandTable, i), "");
	    }
	    return(1);
	case 1:
	    for (i = 0; i < daLen(commandTable); i++) {
		if (strcmp(daGet(struct commandStruct, commandTable, i)->command,
							    command) == 0) {
		    oneCmdHelp(daGet(struct commandStruct, commandTable, i),"");
		    return(1);
		}
	    }
	    badCommand(command);
	    return(1);
	default:
	    return(0);
    }
}

static void oneCmdHelp(cmdPtr, prefix)
struct commandStruct *cmdPtr;
char *prefix;
{
    char cmdAndArgs[50];

    (void) sprintf(cmdAndArgs, "%.20s %.20s", cmdPtr->command, cmdPtr->args);
    (void) printf("%s%-30s%s\n", prefix, cmdAndArgs, cmdPtr->summary);
}

static void badCommand(name)
char *name;
{
    (void) printf("unknown command `%s' -- type `help' for command list\n",
									name);
}

void buildCommandTable(analDialogs)
dary analDialogs;
{
    static struct commandStruct fixedCommandTable[] = {
	"inst-props",	"inst-name",	"edit instance properties",	iProps,
#ifdef LABEL_HACK
	"fix-labels",	"",		"clean up all instance labels", labels,
#endif /*LABEL_HACK*/
	"write",	"[new-facet]",	"write updated facet",		wFacet,
	"ground",	"net-name",	"identify ground net",		ground,
	"plot",		"net-name... (or *)",	"plot net voltage(s)",	plot,
	"save",		"file net-name... (or *)",
					"save net voltage(s) in file",	save,
	"deck",		"file",		"write a spice deck into file",	deck,
	"quit",		"",		"exit from program",		quit,
	"help",		"[command]",	"print help message",		help,
	"?",		"[command]",	"another name for help",	help,
	NIL(char),	NIL(char),	NIL(char),		(int (*)()) 0
    };
    struct commandStruct *cmdPtr;
    int i;

    /* create the command table */
    commandTable = daAlloc(struct commandStruct, 0);

    /* tabulate the analyses */
    for (i = 0; i < daLen(analDialogs); i++) {
	cmdPtr = daSetLast(struct commandStruct, commandTable);
	cmdPtr->command =
		daGet(struct dialog, analDialogs, i)->simAnalPtr->name;
	cmdPtr->args = "";
	cmdPtr->summary =
		daGet(struct dialog, analDialogs, i)->simAnalPtr->description;
	cmdPtr->function = analyze;
    }

    /* copy the fixed commands */
    for (cmdPtr = fixedCommandTable; cmdPtr->command; cmdPtr++) {
	*daSetLast(struct commandStruct, commandTable) = *cmdPtr;
    }
}

/*VARARGS2*/
void ofeMessage(type, format, va_alist)
messageType type;
char *format;
va_dcl
{
    char *typeName;
    va_list ap;

    switch (type) {
	case INFO:
	    typeName = "";
	    break;
	case WARNING:
	    typeName = "Warning: ";
	    break;
	case SEVERE:
	    typeName = "Error: ";
	    break;
	case FATAL:
	    typeName = "Fatal error: ";
	    break;
	default:
	    typeName = "Unknown error code: ";
	    type = FATAL;
	    break;
    }

    va_start(ap);		/*SUPPRESS67*/
    (void) fprintf(stderr, "%s: %s", progName, typeName);
    (void) vfprintf(stderr, format, ap);
    if (type == FATAL) exit(-1);
    va_end(ap);
}
