#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "octspice.h"

static st_table *ofeCircuitTable = NIL(st_table);

int ofeNewUid(simCkt, newUid, oldUid, suffix, type, pointer)
GENERIC *simCkt;
IFuid *newUid;
IFuid oldUid;
char *suffix;
int type;
GENERIC *pointer;
{
    char *val;
    cktDesc *cktPtr;
    uidDesc *old, *new;

    if (ofeCircuitTable == NIL(st_table)) {
	ofeCircuitTable = st_init_table(st_ptrcmp, st_ptrhash);
    }

    old = (uidDesc *) oldUid;
    if (old) {
	cktPtr = old->cktPtr;
    } else if (simCkt) {
	if ( ! st_lookup(ofeCircuitTable, (char *) simCkt, &val)) {
	    errRtn = "ofeNewUid:  attempt to create a uid in unknown circuit";
	    return(E_NOTFOUND);
	}
	cktPtr = (cktDesc *) val;
    } else {
	/* create a new circuit */
	cktPtr = ALLOC(cktDesc, 1);
	INIT_CKT(cktPtr);
	if (st_insert(ofeCircuitTable, (char *) pointer, (char *) cktPtr)) {
	    errRtn = "ofeNewUid:  attempt to create duplicate circuit";
	    return(E_EXISTS);
	}
	new = (uidDesc *) cktPtr;

	goto finish;
    }

    switch (type) {
	case UID_TASK:
	    new = (uidDesc *) ALLOC(taskDesc, 1);
	    INIT_TASK((taskDesc *) new, cktPtr);
	    break;
	case UID_ANALYSIS:
	    new = (uidDesc *) ALLOC(analDesc, 1);
	    INIT_ANAL((analDesc *) new, cktPtr);
	    break;
	case UID_INSTANCE:
	    new = (uidDesc *) ALLOC(cellDesc, 1);
	    INIT_CELL((cellDesc *) new, cktPtr);
	    break;
	case UID_MODEL:
	    new = (uidDesc *) ALLOC(modelDesc, 1);
	    INIT_MODEL((modelDesc *) new, cktPtr);
	    break;
	case UID_SIGNAL:
	    new = (uidDesc *) ALLOC(netDesc, 1);
	    INIT_NET((netDesc *) new, cktPtr);
	    break;
	case UID_OTHER:
	    new = ALLOC(uidDesc, 1);
	    INIT_UID(new, cktPtr);
	    break;
    }

finish:
    if (suffix && *suffix != '\0') {
	if (old && old->uidType == UID_INSTANCE && old->name) {
	    new->name = ALLOC(char, strlen(old->name) + strlen(suffix) + 2);
	    (void) sprintf(new->name, "%s%s%s", old->name,
			(old->uidType == new->uidType) ? "." : "#", suffix);
	} else {
	    new->name = util_strsav(suffix);
	}
    }
    new->simEquiv = pointer;
    *newUid = (IFuid) new;

    return(OK);
}

char *ofeNameOf(uid)
IFuid uid;
{
    return(((uidDesc *) uid)->name);
}

char *ofeCktNameOf(uid)
IFuid uid;
{
    return(((uidDesc *) uid)->cktPtr->name);
}

void ofeFreeUid(uidPtr)
uidDesc *uidPtr;
{
    char *key, *value;
    int i;
    void freeWave();

    uidPtr->fChain->bChain = uidPtr->bChain;
    uidPtr->bChain->fChain = uidPtr->fChain;

    FREE(uidPtr->name);
    switch (uidPtr->uidType) {
	case UID_TASK:
	    break;
	case UID_ANALYSIS:
	    freeWave(((analDesc *) uidPtr)->indepWave);
	    for (i = 0; i < ((analDesc *) uidPtr)->nDepWaves; i++) {
		freeWave(((analDesc *) uidPtr)->depWaves[i]);
	    }
	    FREE(((analDesc *) uidPtr)->depWaves);
	    st_free_table(((analDesc *) uidPtr)->waveTable);
	    break;
	case UID_INSTANCE:
	    st_free_table(((cellDesc *) uidPtr)->termTable);
	    st_free_table(((cellDesc *) uidPtr)->netTable);
	    st_free_table(((cellDesc *) uidPtr)->childTable);
	    break;
	case UID_MODEL:
	    break;
	case UID_SIGNAL:
	    break;
	case UID_OTHER:
	    key = (char *) uidPtr->simEquiv;
	    (void) st_delete(ofeCircuitTable, &key, &value);
	    break;
    }
    FREE(uidPtr);
}

static void freeWave(wavePtr)
waveStruct *wavePtr;
{
    void freeVal();

    if (wavePtr) {
	daFree(wavePtr->val, (void (*)()) 0);
	FREE(wavePtr);
    }
}

netDesc *ofeFindNet(cellPtr, netName, netId)
cellDesc *cellPtr;
char *netName;
octId netId;
{
    char **valPtr;
    netDesc *netPtr;
    IFuid uid;

    if (st_find_or_add(cellPtr->netTable, (char *) netId, &valPtr)) {
	return((netDesc *) *valPtr);
    } else {
	/* make local net */
	OFE_NEW_UID(&uid, (IFuid) cellPtr, netName, UID_SIGNAL);
	netPtr = (netDesc *) uid;
	netPtr->cellPtr = cellPtr;
	netPtr->netId = netId;
	IF_ASSERT(newNode, (cellPtr->cktPtr->simEquiv, &netPtr->simEquiv, uid));
	*valPtr = (char *) netPtr;
	return(netPtr);
    }
}

/* some basic front-end functions */
static int ofePauseTest()
{
    return(0);
}
static double ofeSeconds()
{
    return(0.0);
}
/*ARGSUSED*/
static int ofeError(flags, format, names)
int flags;
char *format;
IFuid *names;
{
    /* XXX */
}
static int ofeNoPointOut()
{
    errRtn = "pointwise output not supported";
    return(E_UNSUPP);
}
static int ofeNoWindowOut()
{
    errRtn = "windowed output not supported";
    return(E_UNSUPP);
}
static int ofeNoDomainOut()
{
    errRtn = "multi-dimensional output not supported";
    return(E_UNSUPP);
}
static int ofeNotSupported()
{
    errRtn = "some output function not supported";
    return(E_UNSUPP);
}

IFfrontEnd ofeFEstruct = {
    ofeNewUid,		/* create a new UID in the circuit */
    ofePauseTest,	/* should we stop now? */
    ofeSeconds,		/* what time is it? */
    ofeError,		/* output an error or warning message */
    ofeNoPointOut,	/* start pointwise output plot */
    ofeNoPointOut,	/* data for pointwise plot */
    ofeNoWindowOut,	/* start windowed output plot */
    ofeNoWindowOut,	/* independent vector for windowed plot */
    ofeNoWindowOut,	/* data for windowed plot */
    ofeNoWindowOut,	/* signal end of windows */
    ofeNotSupported,	/* end of plot */
    ofeNoDomainOut,	/* start nested domain */
    ofeNoDomainOut,	/* end nested domain */
    ofeNotSupported,	/* supply plot or signal attributes */
};

void ofeInitSimulator()
{
    if (ofeSpicePtr == NIL(IFsimulator)) {
	if (SPIinit(&ofeFEstruct, &ofeSpicePtr) != OK) {
	    ofeMessage(FATAL, "couldn't start up simulator\n");
	}
	ofeDeckPtr = deckInit(ofeSpicePtr);
    }
    ofeSimPtr = ofeSpicePtr;
}

int ofeCheckIF(retCode, file, line, severity)
int retCode;
char *file;
int line;
messageType severity;
{
    char *basicError = "";

    switch (retCode) {
	case OK:						return(1);
	case E_PAUSE:		basicError = "pause";			break;
	case E_PANIC:		basicError = "PANIC";			break;
	case E_EXISTS:		basicError = "already exists";		break;
	case E_NODEV:		basicError = "no such instance";	break;
	case E_NOMOD:		basicError = "no such model";		break;
	case E_NOANAL:		basicError = "no such analysis";	break;
	case E_NOTERM:		basicError = "no such terminal";	break;
	case E_BADPARM:		basicError = "unknown parameter";	break;
	case E_NOMEM:		basicError = "out of memory";		break;
	case E_NODECON:		basicError = "node already connected";	break;
	case E_UNSUPP:		basicError = "unsupported request";	break;
	case E_PARMVAL:		basicError = "bad parameter value";	break;
	case E_NOTEMPTY:	basicError = "item still referenced";	break;
	case E_NOCHANGE:	basicError = "can't change ckt";	break;
	case E_NOTFOUND:	basicError = "not found";		break;
	case E_BAD_DOMAIN:	basicError = "unmatched output domains";break;
    }

    ofeMessage(severity, "unexpected simulator error (%s%s%s%s%s) at %s:%d\n",
		    basicError,
		    (errRtn ? ":" : ""), (errRtn ? errRtn : ""),
		    (errMsg ? ":" : ""), (errMsg ? errMsg : ""),
		    file, line);
    FREE(errMsg);
    errMsg = errRtn = NIL(char);

    return(0);
}
